#ifndef _ANALYZER_H_
#define _ANALYZER_H_

#include "Grid.h"

#include "PassAnalyzer.h"
#include "ScoreGoalAnalyzer.h"
#include "DribbleAnalyzer.h"
#include "PositionAnalyzer.h"
#include "GoalKickAfterCatchPositionAnalyzer.h"
#include "InterceptAnalyzer.h"
#include "../intentions/targets/Target.h"
#include "../intentions/conditions/KickConditions.h"
#include "../intentions/conditions/AttackConditions.h"

class KickPossibleCondition;
class GoalReachableCondition;


class Analyzer {
  public:
    typedef unsigned long SituationClass;

    /**@name Variablen f\"ur die Klassifikation der aktuellen Situation. */
    //@{
    static const SituationClass NOTHING = 0; 					// 0
    ///
    static const SituationClass NEW_NORMAL_SITUATION = 1;		// 1
    ///
    static const SituationClass INITIAL_SITUATION_OUR_TEAM =
      NEW_NORMAL_SITUATION<<1;									// 2
    ///
    static const SituationClass INITIAL_SITUATION_THEIR_TEAM =
      INITIAL_SITUATION_OUR_TEAM<<1;							// 4
    ///
    static const SituationClass INITIAL_SITUATION_ACTIVE =
      INITIAL_SITUATION_THEIR_TEAM<<1;							// 8
    ///
    static const SituationClass INITIAL_SITUATION_PASSIVE =
      INITIAL_SITUATION_ACTIVE<<1;								// 16
     /// 
     static const SituationClass FREEKICK_SITUATION = 
 	INITIAL_SITUATION_PASSIVE<<1; 								// 32
    ///
    static const SituationClass FREEKICK_SITUATION_OUR_TEAM =
      INITIAL_SITUATION_PASSIVE<<1;								// 64
    ///
    static const SituationClass FREEKICK_SITUATION_THEIR_TEAM =
      FREEKICK_SITUATION_OUR_TEAM<<1;							// 128
    ///
    static const SituationClass FREEKICK_SITUATION_ACTIVE =
      FREEKICK_SITUATION_THEIR_TEAM<<1;							// 256
    ///
    static const SituationClass GAME_STATE_CHANGE =
      FREEKICK_SITUATION_ACTIVE<<1;								// 512
    ///
    static const SituationClass DEFENSE_SITUATION =
      GAME_STATE_CHANGE<<1;										//1024							
    ///
    static const SituationClass ATTACK_SITUATION =
      DEFENSE_SITUATION<<1;										//2048
    //@}

    Analyzer();
    ~Analyzer();

    void analyze();
    bool isPassPossible();
    bool isRiskPassPossible();
    const Vector& getPassTarget();
    int getPassReceiverId();
    bool isScoreGoalPossible();
    bool isRiskScoreGoalPossible();
    bool isGoalKickAfterCatchPossible();

    const Vector& getDribbleTarget();
    const Vector& getScoreGoalTarget();
    const Vector& getPositionTarget();
    const Vector& getGoaliePositionTarget();
    const Vector& getGoaliePositionTurnVector();
    const Vector& getGoalieMoveTarget();
    const Vector& getGoalKickAfterCatchPassTarget();
		int getGoalKickAfterCatchPassReceiverId();

    /** Liefert den Klassifikationswert der aktuellen Situation,
     wie er von \Ref{situationClassifcation} berechnet wurde, zur\"uck. */
    const SituationClass classifySituation();

    /** Liefert den Id-String. */
    const string& idString() const;

    int getMyStepsToBall();
    int getOurGoalieStepsToBall();
    int getNearestEnemyStepsToBall();
    int getNearestTeammateStepsToBall();
    int get2ndNearestTeammateStepsToBall();

    const Vector& getMyInterceptPosition();
    const Vector& getOurGoalieInterceptPosition();
    const Vector& getNearestEnemyInterceptPosition();
    const Vector& getNearestTeammateInterceptPosition();
    const Vector& get2ndNearestTeammateInterceptPosition();

    const bool getGoalieInterceptRunBehaviour();
    
    // soll von auen gesetzt werden wenn neue seeInfos da sind
    bool haveToAnalyze;

  protected:
    PassAnalyzer                        *thePassAnalyzer;
    ScoreGoalAnalyzer                   *theScoreGoalAnalyzer;
    DribbleAnalyzer                     *theDribbleAnalyzer;
    PositionAnalyzer                    *thePositionAnalyzer;
    GoalKickAfterCatchPositionAnalyzer  *theGoalKickAfterCatchPositionAnalyzer;
    InterceptAnalyzer                   *theInterceptAnalyzer;

  private:
    /** Klassifizierung einer Situation. Es wird die Situation
	klassifiziert und eine Interger-Konstante des Typs SituationClass
	zur\"uckgeleifert. Diese Methode wird von PlaySoccerIntention
	benutzt, um eine Grobanalyse der Situation zu machen. */    
    const SituationClass situationClassifcation();
    
    KickPossibleCondition *theIsKickPossible;
    GoalReachableCondition *theIsGoalReachable;
    int thePositionTimeStamp;

    /** Der Id-String. */
    string theIdString;

    int simStepOfLastAnalyze;
};

inline const string& Analyzer::idString() const{
  return theIdString;
}

inline bool Analyzer::isPassPossible()
{
  return thePassAnalyzer->isPassPossible();
}

inline bool Analyzer::isRiskPassPossible()
{
  return thePassAnalyzer->isRiskPassPossible();
}

inline const Vector& Analyzer::getPassTarget()
{
   return thePassAnalyzer->getPassTarget();
}

inline int Analyzer::getPassReceiverId() {
	return thePassAnalyzer->getPassReceiverId();
}

inline const Vector& Analyzer::getDribbleTarget()
{
   return theDribbleAnalyzer->getDribbleTarget();
}

inline bool Analyzer::isScoreGoalPossible()
{
  return theScoreGoalAnalyzer->isScoreGoalPossible();
}

inline bool Analyzer::isRiskScoreGoalPossible()
{
  return theScoreGoalAnalyzer->isRiskScoreGoalPossible();
}

inline const Vector& Analyzer::getScoreGoalTarget()
{
   return theScoreGoalAnalyzer->getScoreGoalTarget();
}


inline const Vector& Analyzer::getPositionTarget()
{
   return thePositionAnalyzer->getPositionTarget();
}

inline const Vector& Analyzer::getGoaliePositionTarget()
{
   return thePositionAnalyzer->getGoaliePositionTarget();
}

inline const Vector& Analyzer::getGoaliePositionTurnVector()
{
   return thePositionAnalyzer->getGoaliePositionTurnVector();
}

inline const Vector& Analyzer::getGoalieMoveTarget()
{
   return theGoalKickAfterCatchPositionAnalyzer->getGoalKickAfterCatchPositionTarget();
}

inline const Vector& Analyzer::getGoalKickAfterCatchPassTarget()
{
   return theGoalKickAfterCatchPositionAnalyzer->getGoalKickAfterCatchPassTarget();
}

inline bool Analyzer::isGoalKickAfterCatchPossible()
{
   return theGoalKickAfterCatchPositionAnalyzer->isGoalKickAfterCatchPossible();
}

inline int Analyzer::getGoalKickAfterCatchPassReceiverId()
{
   return theGoalKickAfterCatchPositionAnalyzer->getGoalKickAfterCatchPassReceiverId();
}

inline int Analyzer::getMyStepsToBall(){
  return theInterceptAnalyzer->getMyStepsToBall();
}

inline int Analyzer::getOurGoalieStepsToBall(){
  return theInterceptAnalyzer->getOurGoalieStepsToBall();
}

inline int Analyzer::getNearestEnemyStepsToBall() {
  return theInterceptAnalyzer->getNearestEnemyStepsToBall();
}

inline int Analyzer::getNearestTeammateStepsToBall() {
  return theInterceptAnalyzer->getNearestTeammateStepsToBall();
}

inline int Analyzer::get2ndNearestTeammateStepsToBall() {
  return theInterceptAnalyzer->get2ndNearestTeammateStepsToBall();
}

inline const Vector& Analyzer::getMyInterceptPosition() {
  return theInterceptAnalyzer->getMyInterceptPosition();
}

inline const Vector& Analyzer::getOurGoalieInterceptPosition() {
  return theInterceptAnalyzer->getOurGoalieInterceptPosition();
}

inline const Vector& Analyzer::getNearestEnemyInterceptPosition() {
  return theInterceptAnalyzer->getNearestEnemyInterceptPosition();
}

inline const Vector& Analyzer::getNearestTeammateInterceptPosition() {
  return theInterceptAnalyzer->getNearestTeammateInterceptPosition();
}

inline const Vector& Analyzer::get2ndNearestTeammateInterceptPosition() {
  return theInterceptAnalyzer->get2ndNearestTeammateInterceptPosition();
}

inline const bool Analyzer::getGoalieInterceptRunBehaviour() {
  return theInterceptAnalyzer->getGoalieInterceptRunBehaviour();
}


extern Analyzer *globalAnalyzer;


#endif

