#include "GoalKickAfterCatchPositionAnalyzer.h"

#include "../basic/Timer.h"
#include "../worldmodel/WorldModelInterface.h"
#include "../worldmodel/Situation.h"
//#include "../serverInterface/BodyInfos.h"

GoalKickAfterCatchPositionAnalyzer::GoalKickAfterCatchPositionAnalyzer()
{
	catchCount = 0;
	moveCount = 0;
}

const Vector& GoalKickAfterCatchPositionAnalyzer::getGoalKickAfterCatchPositionTarget()
{
	return goalKickAfterCatchPositionTarget;
}

const Vector& GoalKickAfterCatchPositionAnalyzer::getGoalKickAfterCatchPassTarget()
{
	return goalKickAfterCatchPassTarget;
}

const int GoalKickAfterCatchPositionAnalyzer::getGoalKickAfterCatchPassReceiverId()
{
	if( passReceiverId >= MIN_TEAMMATE_ID && passReceiverId < MAX_EXISTING_TEAMMATE_ID) {
		return passReceiverId;
	}
	else {
		return UNDEF_TEAMMATE_ID;
	}	
}

void GoalKickAfterCatchPositionAnalyzer::calculateValues() {
//int startTime = globalTimer->getAgentTime();

  INFO( Log::ANY,  10,  idString() , globalLog->infoStream
    << "GoalKickPositionAnalyzer::calculateValues started at " << globalTimer->getAgentTime() << endl);
	
	// haben wir neu gecatched? - dann duerfen wir jetzt rechnen.
	if( catchCount <  BodyInfos::staticCatches ) {
		moveCount = BodyInfos::staticMoves+1;
		//setGoalKickAfterCatchPossible( NO );
		//cout << "GoalKickAfterCatchPositionAnalyzer: gecatched - first calculate. "  << catchCount << " " << moveCount << endl;
	}	
	// wenn wir nach dem Catch einmal gemoved haben, dann wird kein neues 
	// passTarget berechnet... damit wir dann auch in die richtige Richtung
	// kicken. 
  if( moveCount == BodyInfos::staticMoves ) {
    	//cout << "GoalKickAfterCatchPositionAnalyzer: schon gemoved - return. " << catchCount << " " << moveCount << endl;
		return;	
  }
    
	setGoalKickAfterCatchPossible( NO );
//	cout << "SIMSTEP: " << globalWorldModel->simStep() <<" NEUER AUFRUF VON CALCULATE_VALUES "  << catchCount << " " << moveCount << endl;
//	cout << "Analyse auf Situation :" << globalWorldModel->situation() << endl;

	
	// Beampositionen, die durchgetestet werden 
	vector <Vector> ballPositions;
//	ballPositions.push_back( Vector(-37.0,  19.0) );
	ballPositions.push_back( Vector(-37.0,  15.0) );
	ballPositions.push_back( Vector(-37.0,   9.0) );
	ballPositions.push_back( Vector(-37.0,   0.0) );
	ballPositions.push_back( Vector(-37.0,  -9.0) );
	ballPositions.push_back( Vector(-37.0, -15.0) );
//	ballPositions.push_back( Vector(-37.0, -19.0) );



	double maxAngle =  radFromDegree(75) ; // Winkel links und rechts von preferredKickDir, die auf Passmoeglichkeiten abgetestet wird
	double testAngle = 0.0;
	double step = radFromDegree(3.0); // Schritte in denen auf Passmoeglichkeiten getestet wird
	
  double ballSpeed = 2.7;

	int safetySteps = 3; // Steps, die unser Spiele naeher am Ball sein muss, damit gepasst wird
	double posConfMin = 0.2;

	bool addTestAngle = YES;
 
	int theirSteps = 51;
	int oldTheirSteps = 51;
	int ourSteps = 51;
	
	// Distanz, die der Spieler maximal vom Abstosspunkt entfernt sein darf
 	// damit noch geprueft wird, ob er angepasst werden kann 
 	double maxDistForGoalKick = 30.0;

	Situation theSituation = globalWorldModel->situation();
    
  ModelledPlayer myself = theSituation.playerWithId( globalWorldModel->id() );
  Vector ballPos = Vector();
  
  vector <ModelledPlayer> ourPlayers; 
  vector <double> distToGoal;
  vector <double> value;
  
  double distToBall = 31.0;
  double wayOfBall = 0.0;
  double maxValue = -23;
  double aktValue = 0;
    
  Vector interceptPoint = Vector( 100.0, 100.0 );
    
	ModelledBall theBall = globalWorldModel->ball();
     
     
   	// eigenen Player rausnehmen
	theSituation.removePlayerWithId( globalWorldModel->id() );

	// alle eigenen Spieler mit zu geringer posConfidence aus der Situation rauswerfen 
	// - zu denen passen wir eh nicht -
	for( int i=0; i < theSituation.playerCount(); i++ ){
	  const ModelledPlayer& p = theSituation.playerWithIndex(i);
		if(	p.team() == OUR_TEAM_ID ) {
			if(   ( p.posConfidence().reliability() < posConfMin )
			   || ( ( p.pos().x() < -32.0 ) && ( fabs(p.pos().y()) < 25.0 ) ) ) {
				theSituation.removePlayerWithId( theSituation.index2Id(i) );
			}
			else{
				// gedeckelt auf 51 - also hoechstens bis zur Mittellinie				 
				distToGoal.push_back( ((p.distTo( globalWorldModel->field().ourGoal() ) > 50) ? 51 : p.distTo( globalWorldModel->field().ourGoal() )) );
				ourPlayers.push_back( p );
			}	
		} 			
	}//endfor 

//cout << " bereinigte Situation : " << theSituation << endl;

	// alle Beampositionen durchgehen
	for( int i = 0; i < (int) ballPositions.size(); i++ ) {
		ballPos = ballPositions[i]; 
		theBall.setPos( ballPos );
		theSituation.setBall( theBall );
		
		testAngle = 0.0;
		
//		cout << "test Beamposition: " << ballPos << endl; 
 					
		// Schusswinkel ( zwischen prefferedKickDir +/- maxAngle ) durchtesten	
		while( testAngle < maxAngle ) {
	
			// Abwechselnd wird der testAngle zum/vom 0.0 addiert/abgezogen - und anschliessend um step erhoeht 
			if( addTestAngle == YES ) {
				theBall.setSpeed( Vector( ballSpeed, 0.0+testAngle, POLAR, RADIAN)); // (globalWorldModel->preferredKickDirection( ballPos )).angle()+testAngle, POLAR, RADIAN));
		 		addTestAngle = NO;
			}	
			else {
				theBall.setSpeed( Vector( ballSpeed, 0.0-testAngle, POLAR, RADIAN)); //(globalWorldModel->preferredKickDirection( ballPos )).angle()-testAngle, POLAR, RADIAN));
				addTestAngle = YES;
				testAngle += step;
			} 
			theSituation.setBall( theBall );

			oldTheirSteps = theirSteps;
			theirSteps = theSituation.stepsOfNearestPlayer( THEIR_TEAM_ID );
			
			if( theirSteps > 51 ) { 
				theirSteps = 51;
			}	
			
			ourSteps = theSituation.stepsOfNearestPlayer( OUR_TEAM_ID );
			if( ourSteps > 51 ) {
				ourSteps = 51;
			}

			// DEBUG	
//			cout << "GoalKickAfterCatchPositionAnalyzer: teste Winkel: " << degFromRadiant( theBall.speed().angle() ) << endl;
			if( (ourSteps + safetySteps) > theirSteps ) {
//				cout << "Gegner scheller am Ball. ourSteps: " << ourSteps << " + safety: " << safetySteps << " theirSteps: " << theirSteps << endl;	
			}
			// wir schneller am Ball ? 
			else  {		
				// Strecke, die der Ball zuruecklegt, bis wir ihn haben
				/*wayOfBall = 5;
				for( int i = 0; i <= ourSteps; i++ ) {  
					ballSpeed = ballSpeed * globalConf->ballDecay;
					wayOfBall = wayOfBall + ballSpeed;
				}
				
				interceptPoint = theBall.pos() + Vector( wayOfBall,						
			   				   ( (ballPos+theBall.speed()) - theBall.pos() ).angle(),
			   				   POLAR, RADIAN);*/
			  interceptPoint = theBall.pos() 
			  								 + Vector( theBall.speed().length() * globalSkills->getNewBallDistance(ourSteps),
			  								 					 theBall.speed().angle(), POLAR, RADIAN);

				// ist der Ball noch im Feld, oder schon im aus?												
				if( globalWorldModel->field().isInsideField( interceptPoint ) ) {

				  for( int k=0; k < (int) ourPlayers.size(); k++) {
					
					  // welcher unserer Spieler kriegt den Ball ?
					  if( theSituation.nearestPlayer( OUR_TEAM_ID ).id() == ourPlayers[ k ].id() )  {

							// Strecke, die der Spieler vom Abstosspunkt entfernt ist
							distToBall = ourPlayers[ k ].distTo( ballPos );
							
							// Bewertung des Passes...
							// Abstand zum gegnerischen Tor -> mglichst weit weg, max  
							// Anzahl der gegnerischen-eigenen Schritte zum Ball -> mglichst grosse Sicherheit. max
							aktValue = distToGoal[ k ]-15 + (theirSteps - ourSteps);  
/*	
							cout << "teste Spieler: " << ourPlayers[ k ].id() << endl;
							cout << "maxValue: " << maxValue << " ourSteps: " << ourSteps << " theirSteps: " << theirSteps << endl;
							cout << "aktValue: " << aktValue << " distToGoal: " << distToGoal[ k ]-15  << " stepDiff: " << (theirSteps-ourSteps) << endl;   
							cout << "interceptPoint: " << interceptPoint << "wayofball: " << wayOfBall << " distToBall: "<< distToBall << endl; 
*/									
							if( maxValue > aktValue ) {
//								cout << "maxValue war groesser: " << maxValue << " > " << aktValue << endl;
							} 
							if( distToBall > maxDistForGoalKick ) {
//								cout << "zu weit weg." << endl;
							}//endif 
							
							if( maxValue < aktValue && 
								distToBall < maxDistForGoalKick ) {
	
//								cout <<"das Beamtarget ist: " << ballPos << endl;
//								cout <<"das PassTarget ist: " << ballPos + theBall.speed() << endl; 
	
								maxValue = aktValue;
								goalKickAfterCatchPossible = YES;		
					   			setGoalKickAfterCatchPassTarget( ballPos + theBall.speed() );
					   			setGoalKickAfterCatchPassReceiverId( ourPlayers[ k ].id() );
						   		setGoalKickAfterCatchPositionTarget( ballPos );
								//cout << "GoalKickAfterCatchPositionAnalyzer: setzte move und pass target " << endl;	
							}// endif
							break;
						}//endif	
					}// endfor
				}//endif
//				else cout << "Ball im aus. " << interceptPoint <<  endl;				
			}//endelse 
  	}// endwhile
	}// endfor  (Beampositionen-Schleife)

	// es konnte kein Passtarget gefunden werden ! Notfall-Abstoss !	
	if ( goalKickAfterCatchPossible == NO ) {
//		cout <<"GoalKickAfterCatchPositionAnalyzer: pass NOT Possible - notfallAbstoss:" << endl;
		setGoalKickAfterCatchPositionTarget(Vector(-15.0, 10.0));
		setGoalKickAfterCatchPassTarget(Vector( 25.0, 34.0 )); 
		setGoalKickAfterCatchPassReceiverId( UNDEF_TEAMMATE_ID );		
	}
	else {
//		cout <<"GoalKickAfterCatchPositionAnalyzer: passPossible ! nach Schleife" << endl;
	}	

  catchCount = BodyInfos::staticCatches;

  INFO( Log::ANY,  10,  idString() , globalLog->infoStream
    << "GoalKickPositionAnalyzer::calculateValues ended at " << globalTimer->getAgentTime() << endl );
//int duration = globalTimer->getAgentTime() - startTime;
//cerr << "GoalKickPositionAnalyzer::calculateValues in " << globalWorldModel->simStep() << "  in " << duration << " ms\n";
}

