/*
 * Aktueller Stand
 *
 *   $RCSfile: Grid.cc,v $
 *   $Revision: 1.1.1.1 $
 *   $Date: 2001/07/31 22:23:08 $
 *   $Author: rami $
 *
 */

#include <math.h>
#include <malloc.h>

#include "Grid.h"

#include "../worldmodel/WorldModelInterface.h"

Grid::Grid(){
  // Funktionsfelder setzen
  for( int i = -GRID_RANGE_X; i <= GRID_RANGE_X; ++i ) {
    for( int j = -GRID_RANGE_Y; j <= GRID_RANGE_Y; ++j ) {
      f[0][i + GRID_RANGE_X][j + GRID_RANGE_Y] = f0(i,j);
      f[1][i + GRID_RANGE_X][j + GRID_RANGE_Y] = f1(i,j);
      f[2][i + GRID_RANGE_X][j + GRID_RANGE_Y] = f2(i,j);
      f[3][i + GRID_RANGE_X][j + GRID_RANGE_Y] = f3(i,j);
    }
  }

  // Datenfeld leeren
  for( int gridType = 0; gridType < GRID_TYPE_COUNT; ++gridType )
    for( int i = 0; i <= GRID_RANGE_X; ++i )
      for( int j = 0; j <= GRID_RANGE_Y; ++j )
        theGrid[gridType][i][j] = 0.0;

  theTimeStamp = 0;
  thePause = 1;
}

Grid::~Grid()
{

}

// liefert Feld mit den Zeilensummen
double* Grid::getLineSum (int gridType ) {
  double* pdouble;

  pdouble = ( double* ) malloc( ( GRID_RANGE_Y + 1 ) * sizeof( double ) );
  for( int y = 0; y <= GRID_RANGE_Y; ++y ) pdouble[y] = 0.0;

  for( int y = 0; y <= GRID_RANGE_Y; ++y )
    for( int x = 0; x <= GRID_RANGE_X; ++x )
      pdouble[y] += theGrid[gridType][x][y];

  return (pdouble);
}

// liefert Feld mit den Spaltensummen
double* Grid::getColumnSum ( int gridType ) {
  double* pdouble;

  pdouble = (double*) malloc ((GRID_RANGE_X + 1) * sizeof(double));
  for( int x = 0; x <= GRID_RANGE_X; ++x ) pdouble[x] = 0.0;

  for( int x = 0; x <= GRID_RANGE_X; ++x )
    for( int y = 0; y <= GRID_RANGE_Y; ++y )
      pdouble[x] += theGrid[gridType][x][y];

  return (pdouble);
}

/** Ausgabe des gesamten Grids fuer Gnuplot
      die Ausgabe mu in eine Datei umgeleitet werden */
void Grid::getGridPoints( int gridType ) const {
  cout << "Anfang\n";
  for( int x = 0; x <= GRID_RANGE_X; ++x ) {
    for( int y = 0; y <= GRID_RANGE_Y; ++y ) {
      cout << (double) x - 52.5 << " " << (double) y - 34.0
           << " " << theGrid[gridType][x][y] << "\n";
    }
  }
  cout << "\nEnde\n";
}

// setzt Zahl, die angibt nach wie vielen Takten eine neue Berechnung
// des Grids erfolgt
void Grid::setPause (int paused) {
  thePause = paused;
  theTimeStamp = globalWorldModel->simStep() - paused; 	
}
// setzt Spieler auf Kantenkreuzung
Point Grid::rePos(Point thePos) {
	Point pos;

	pos.x = round(thePos.x);
	pos.y = round(thePos.y);
	return (pos);
}


// lscht Grid
void Grid::resetValues(int gridType) {
  /** Implementation mit Pointern (wegen der Geschwindigkeit)
      --> die alte Variante ist noch auskommentiert vorhanden
      Zugriff folgendermassen :
      Adresse von theGrid[a][b][c] => *( *( theGrid + a ) + b ) + c
      wenn man den Pointer auf diese Adresse dann um 1 hochzaehlt,
      dann durchlaeuft er erst alle c, dann wird b um 1 erhoeht,
      dann durchlaeuft er wieder alle c, ...
  */

  double *gridPointer = **(theGrid + gridType);

  for( int x = 0; x <= GRID_RANGE_X; ++x ) {
    for( int y = 0; y <= GRID_RANGE_Y; ++y ) {
// Alte Variante :  theGrid[gridType][x][y] = 0.0;
      *gridPointer = 0.0;
      ++gridPointer;
    }
  }
}

// addiert Funktion auf das Grid 
// *Vector: Position des Spielers (dieser wurde mit rePos bereits korrekt verschoben
// *weight: Gewichtung der Funktion
// *fType: welche Funtkion soll benutzt werden
void Grid::addValues( Vector pos, double weight, int fType, int gridType ) {
  /** Implementation mit Pointern (wegen der Geschwindigkeit)
      --> die alte Variante ist noch auskommentiert vorhanden
      Zugriff folgendermassen :
      Adresse von theGrid[a][b][c] => *( *( theGrid + a ) + b ) + c
      wenn man den Pointer auf diese Adresse dann um 1 hochzaehlt,
      dann durchlaeuft er erst alle c, dann wird b um 1 erhoeht,
      dann durchlaeuft er wieder alle c, ...
  */

  if( globalWorldModel->field().isInsideField( pos ) ) {
    // wir bewerten nur Spieler die sich auf dem Spielfeld befinden
    int x = -round(pos.x() + 52.5) + GRID_RANGE_X;
    int y = -round(pos.y() + 34.0) + GRID_RANGE_Y;

    // Pointer aufs Grid und die Funktion
    double *gridPointer = **(theGrid + gridType);
    double *fPointer;

    for( int i = 0; i <= GRID_RANGE_X; ++i ) {
      /** bei dem Pointer auf die Funktion sind noch die x und y Offsets
          zu beachten, da wir aus dem Feld (Groesse von 4 Fussballfeldern)
          nur ein Rechteck ausschneiden, sodass die Mitte von dem Funktionsfeld
          auf Pos liegt
      */
      fPointer = *(*(f + fType) + i + x) + y;
      for( int j = 0; j <= GRID_RANGE_Y; ++j ) {
        // alte Variante :  theGrid[gridType][i][j] += weight * f[fType][i + x][j + y];
        *gridPointer += *fPointer * weight;
        ++gridPointer;
        ++fPointer;
      }
    }
  }
}

/** diese Funktion hebt das durch die uebergebenen Koordinaten
    definierte Rechteck um den Wert value an
*/
void Grid::liftArea( double lowerXPos, double lowerYPos,
                     double upperXPos, double upperYPos,
                     double value, int gridType ){
  // die Grenzen muessen getestet werden, sonst gibts Fehler
  // --> werden ggf. getauscht
  // das sollte aber schon vorher passieren (wegen der Semantik)
  if( lowerXPos > upperXPos ) {
    double h = lowerXPos;
    lowerXPos = upperXPos;
    upperXPos = h;
  }
  if( lowerYPos > upperYPos ) {
    double h = lowerYPos;
    lowerYPos = upperYPos;
    upperYPos = h;
  }

  // Positionen ausserhalb des Spielfeldes werden zurueckgesetzt
  if( lowerXPos < -52.5 ) lowerXPos = -52.5;
  if( lowerXPos >  52.5 ) lowerXPos =  52.5;
  if( upperXPos < -52.5 ) upperXPos = -52.5;
  if( upperXPos >  52.5 ) upperXPos =  52.5;

  if( lowerYPos < -34.0 ) lowerYPos = -34.0;
  if( lowerYPos >  34.0 ) lowerYPos =  34.0;
  if( upperYPos < -34.0 ) upperYPos = -34.0;
  if( upperYPos >  34.0 ) upperYPos =  34.0;

  int xOffSet = round(lowerXPos + 52.5);
  int yOffSet = round(lowerYPos + 34.0);

  int lBoundX = round( lowerXPos );
  int uBoundX = round( upperXPos );
  int lBoundY = round( lowerYPos );
  int uBoundY = round( upperYPos );

  // Pointer aufs Grid
  double *gridPointer = *(*(theGrid + gridType) + xOffSet);

  for( int x = lBoundX; x <= uBoundX; ++x ) {
    gridPointer += yOffSet;
    for( int y = lBoundY; y <= uBoundY; ++y ) {
      *gridPointer += value;
      ++gridPointer;
    }
  }
}

/** addiert Werte zum Spielfeldrand
    ACHTUNG: das gegnerische Tor wird ausgelassen
    Parameter 1: Wert der auf die Seitenauslinien und Grundlinien addiert wird
    Parameter 2: Wert der auf die Linien (1m zur Spielfeldmitte) neben den Seitenauslinien
                 und Grundlinien addiert wird
*/
void Grid::addValuesToFieldSides( double onLine, double nextToLine, int gridType ) {
   // die beiden Seitenauslinien
   for( int x = 0; x <= GRID_RANGE_X; ++x ) {
     theGrid[gridType][x][0] += onLine;
   }
   for( int x = 0; x <= GRID_RANGE_X; ++x ) {
     theGrid[gridType][x][GRID_RANGE_Y] += onLine;
   }

   // die Linien 1m neben den beiden Seitenauslinien
   for( int x = 0; x <= GRID_RANGE_X; ++x ) {
     theGrid[gridType][x][1] += nextToLine;
   }
   for( int x = 0; x <= GRID_RANGE_X; ++x ) {
     theGrid[gridType][x][GRID_RANGE_Y - 1] += nextToLine;
   }

   // die eigene Grundlinie und die Linie 1m daneben
   for( int y = 2; y <= GRID_RANGE_Y - 2; ++y ) {
     theGrid[gridType][0][y] += onLine;
   }
   for( int y = 2; y <= GRID_RANGE_Y - 2; ++y ) {
     theGrid[gridType][1][y] += nextToLine;
   }

   // die gegnerische Grundlinie und die Linie 1m daneben
   // das Tor wird ausgelassen
   for( int y = 2; y <= 26; ++y ) {
     theGrid[gridType][GRID_RANGE_X][y] += onLine;
   }
   for( int y = 42; y <= GRID_RANGE_Y - 2; ++y ) {
     theGrid[gridType][GRID_RANGE_X][y] += onLine;
   }
   for( int y = 2; y <= 26; ++y ) {
     theGrid[gridType][GRID_RANGE_X - 1][y] += nextToLine;
   }
   for( int y = 42; y <= GRID_RANGE_Y - 2; ++y ) {
     theGrid[gridType][GRID_RANGE_X - 1][y] += nextToLine;
   }
}

// fasst Grid zu einem kleineren Grid zusammen
// Bsp.: standardmig wird ein 3x3 Feld zu einem Punkt zusammengefasst
//       (dabei wird der Mittelwert der 9 Zahlen berechnet
double* Grid::getLessDiscreteGrid( int newX = GRID_RANGE_X/3, int newY = GRID_RANGE_Y/3, int gridType ) {
  double *grid = (double*)malloc( (newX+1) * (newY+1) * sizeof( double ) );
  int *num = (int*)malloc( (newX+1) * (newY+1) * sizeof( int ) );
  int i, j;

  // reset
  for( i = 0; i <= newY; i++ )
    for( j = 0; j <= newX; j++ ) {
      grid[i*(newX+1)+j] = 0.0;
      num[i*(newX+1)+j] = 0;
    }

  // sum
  for( i = 0; i <= GRID_RANGE_Y; i++ ) {
    for( j = 0; j <= GRID_RANGE_X; j++ ) {
      grid[((int)(i*(newY+1)/(GRID_RANGE_Y+1)))*(newX+1) + j*(newX+1)/(GRID_RANGE_X+1)] += theGrid[gridType][i][j];
      ++num[((int)(i*(newY+1)/(GRID_RANGE_Y+1)))*(newX+1) + j*(newX+1)/(GRID_RANGE_X+1)];
    }
  }

  // norm
  for( i = 0; i <= newY; i++ )
    for( j = 0; j <= newX; j++ )
      grid[i*(newX+1)+j] /= num[i*(newX+1)+j];
  free(num);
  return grid;
}


// liefert Wert von einem Grid - Punkt
// *Vector: Grid - Punkt
// *Rckgabe: Wert des Grid - Punktes
double Grid::getValue(Vector pos, int gridType ) {
  if( globalWorldModel->field().isInsideField( pos ) ) {
    return theGrid[gridType][round( pos.x() + 52.5 )][round( pos.y() + 34.0 )];
  }
  else {
    return -20.0;
  }
}

Vector* Grid::getLocalMaximum(int radius = 1, int gridType ) {
  ModelledPlayer thisPlayer = globalWorldModel->player();
  int xLBound = max((int) (thisPlayer.pos().x() - radius + 52.5), 0);
  int xUBound = min((int) (thisPlayer.pos().x() + radius + 52.5), GRID_RANGE_X);
  int yLBound = max((int) (thisPlayer.pos().y() - radius + 34.0), 0);
  int yUBound = min((int) (thisPlayer.pos().y() + radius + 34.0), GRID_RANGE_Y);

  int xMax = xUBound;
  int yMax = ( yLBound + yUBound ) / 2;
  double maxValue = theGrid[gridType][xMax][yMax];

  for( int x = xLBound; x <= xUBound; ++x ) {
    for( int y = yLBound; y <= yUBound; ++y ) {
      if ( maxValue < theGrid[gridType][x][y] ) {
        maxValue = theGrid[gridType][x][y];
        xMax = x;
        yMax = y;
      }
    }
  }
  return (new Vector((double)(xMax - 52.5), (double)(yMax - 34.0)));
}

Vector* Grid::getMaximumAroundPos(Vector pos, int radius = 15, int gridType ) {
  int xLBound = max((int) (pos.x() - radius + 52.5), 0);
  int xUBound = min((int) (pos.x() + radius + 52.5), GRID_RANGE_X);
  int yLBound = max((int) (pos.y() + 34.0 - radius), 0);
  int yUBound = min((int) (pos.y() + 34.0 + radius), GRID_RANGE_Y);

  int xMax = xUBound;
  int yMax = ( yLBound + yUBound ) / 2;
  double maxValue = theGrid[gridType][xMax][yMax];

  for( int x = xLBound; x <= xUBound; ++x ) {
    for( int y = yLBound; y <= yUBound; ++y ) {
      if ( maxValue < theGrid[gridType][x][y] ) {
        maxValue = theGrid[gridType][x][y];
        xMax = x; yMax = y;
      }
    }
  }
  return (new Vector((double)(xMax - 52.5), (double)(yMax - 34.0)));
}

// sucht den Punkt mit dem gobalen Maximum
// *Rckgabe: Punkt des globalen Maximums
Vector* Grid::getGlobalMaximum(int gridType ) {
  Point maxPos;
  double max;

  maxPos.x = 0;
  maxPos.y = 0;
  max = theGrid[gridType][0][0];

  for( int x = 0; x <= GRID_RANGE_X; ++x ) {
    for( int y = 0; y <= GRID_RANGE_Y; ++y ) {
      if ( max < theGrid[gridType][x][y] ) {
        maxPos.x = x;
        maxPos.y = y;
        max = theGrid[gridType][x][y];
      }
    }
  }
  return (new Vector((double)(maxPos.x - 52.0), (double)(-maxPos.y + 33.0)));
}

///////////////////

// berechnet RGB - Werte des Feldes, wobei die Werte auf 255 bzw. -255 normiert werden
// (ein Grnwert von 255 heisst also nicht, dasss der Wert dort 255 betrgt, sondern dass
// dieser insgesamt ber das Feld betrachtet den hchsten Wert darstellt
// *Rckgabe: Zeiger auf das Feld
unsigned long* Grid::getGridRGB(int gridType ) {
  int i, j, valint;
  double max, valdouble;
  unsigned long *pGrid, *pFirstGrid;

  pGrid = (unsigned long*) malloc ((GRID_RANGE_X + 1) * (GRID_RANGE_Y + 1)
                * sizeof(unsigned long));
  pFirstGrid = pGrid;

  max = fabs(theGrid[gridType][0][0]);
  for( i = 0; i <= GRID_RANGE_X; ++i ) {  // Maximum finden
		for( j = 0; j <= GRID_RANGE_Y; ++j ) {
      valdouble = fabs(theGrid[gridType][i][j]);
      if (valdouble > max) max = valdouble;
    }
  }

  max = 255.0 / max;
  for( i = 0; i <= GRID_RANGE_X; i++ ) {  // BGR - Werte berechnen
		for( j = 0; j <= GRID_RANGE_Y; ++j ) {
			valint = round(theGrid[gridType][i][j] * max);
      if (valint < 0) {    // BGR
        *pGrid= ((255 + valint) << 16) + ((255 + valint) << 8) + (255) ;
      } else {
        *pGrid = ((255 - valint) << 16) + (255 << 8) + (255 - valint);
      }
      pGrid++;
		}
	}

  return (pFirstGrid);
}

///////////////////
// hier sind die Funktionen definiert (diese sollten einen Funktionsbereich von 1.0 bis -1.0
// besitzten (eine Gewichtung erfolgt ber das "weight" - Argument bei "addValues

double Grid::f0(double x, double y) {
	return 1.0/((x*x + y*y)/50 + 1);
}

double Grid::f1(double x, double y) {
	return exp(-(x*x + y*y)/(100));
}

double Grid::f2(double x, double y) {
	if (sqrt(x * x + y * y) < 20) { return 0.35 * exp(-(x * x + y * y) / 100); }
	else { return ( 0.35 * (-0.5 * (sqrt(x * x + y * y) / 20 - 1))); };
}

double Grid::f3(double x, double y) {
	return exp(-(x*x + y*y)/(500));
}

////////////////////
// Logger Methoden

// hier wird nur ein Zeiger auf die Agenten gesetzt
// *players: Zeiger auf ein "ToGrid" - struct
void Grid::init (ToGrid *players) {
  playersLogger = players;
}

// addiert Funktion auf das Grid 
// *Vector: Position des Spielers (dieser wurde mit rePos bereits korrekt verschoben
// *weight: Gewichtung der Funktion
// *fType: welche Funtkion soll benutzt werden
void Grid::addValuesLogger(Point pos, double weight, int fType, int gridType ) {
  int x, y;

  x = round(pos.x) - GRID_RANGE_X; // Optimierung
  y = round(pos.y) - GRID_RANGE_Y;
  for( int i = 0; i <= GRID_RANGE_X; i++ ) {
    for( int j = 0; j <= GRID_RANGE_Y; j++ ) {
      theGrid[gridType][i][j] += weight * f[fType][i - x][j - y];
    }
  }
}

// geht alle 23 Spieler durch und legt entsprechend Funktionen auf des Grid
// * WorldModel: struct mit allen Agenten 
// * number: Nummer des Spielers
void Grid::calculateValuesLogger(ToGrid *WorldModel, int number, int gridType ) {
  double weight = 1;
  Point p;

  resetValues(gridType);
  for( int i = 0; i < 23; i++ ) {
    p = makePoint(WorldModel->thePlayers[i].theX, WorldModel->thePlayers[i].theY);
    if(i != number)	 { // nicht diesen Spieler betrachten
      if (WorldModel->thePlayers[i].theId == WorldModel->thePlayers[number].theId) { // Mitspieler
				addValuesLogger(rePos(p), -1*weight, 1, gridType);
      }
      else { // Gegenspieler
	if (WorldModel->thePlayers[i].theId != 0) {
 	  switch( WorldModel->thePlayers[number].theRole) {
	    case 1: //Angreifer
     	      addValuesLogger(rePos(p), -1*weight, 1, gridType);
	      break;
	    case 2: //Verteidiger
	      addValuesLogger(rePos(p), 1*weight, 0, gridType);
              break;
	    default:;
	  }
	}
        else {	// Ball
 	  addValuesLogger(rePos(p), 2*weight, 3, gridType);
 	}
      }
    }
    else { // HomePosition
//    addValues(rePos(makePoint(WorldModel->thePlayers[i].theX,
//    WorldModel->thePlayers[i].theY)), 1*weight, 2);
    }
  }
}


// berechnet RGB - Werte des Feldes, wobei die Werte auf 255 bzw. -255 normiert werden
// (ein Grnwert von 255 heisst also nicht, dasss der Wert dort 255 betrgt, sondern dass
// dieser insgesamt ber das Feld betrachtet den hchsten Wert darstellt
// *Rckgabe: Zeiger auf das Feld
unsigned long* Grid::calculateLoggerRGB(int gridType ) {
  int i, j, valint;
  double max, valdouble;
  unsigned long *pGrid, *pFirstGrid;

  calculateValuesLogger(playersLogger, playersLogger->indexOfPlayerToShow, gridType);

  pGrid = (unsigned long*) malloc ((GRID_RANGE_X + 1) * (GRID_RANGE_Y + 1)
              * sizeof(unsigned long));
  pFirstGrid = pGrid;

  max = fabs(theGrid[gridType][0][0]);
  for( i = 0; i <= GRID_RANGE_X; ++i ) {  // Maximum finden
    for( j = 0; j <= GRID_RANGE_Y; ++j ) {
      valdouble = fabs(theGrid[gridType][i][j]);
      if (valdouble > max) max = valdouble;
    }
  }
  max = 255.0 / max;
  for( i = 0; i <= GRID_RANGE_X; ++i ) {  // BGR - Werte berechnen
    for( j = 0; j <= GRID_RANGE_Y; ++j ) {
      valint = round(theGrid[gridType][i][j] * max);
      if (valint < 0) {    // BGR
        *pGrid= ((255 + valint) << 16) + ((255 + valint) << 8) + (255) ;
      }
      else {
        *pGrid = ((255 - valint) << 16) + (255 << 8) + (255 - valint);
      }
      pGrid++;
    }
  }

  return (pFirstGrid);
}
