#include <stdlib.h>
#include <string>
#include <stream.h>

#include "Agent.h"
#include "Common.h"
#include "Configuration.h"
#include "Formation.h"
#include "Formation2.h"
#include "Log.h"
#include "Say.h"
#include "../worldmodel/Neck.h"
#include "../worldmodel/WorldModelInterface.h"

/*
 * Aktueller Stand
 *
 *   $RCSfile: Agent.cc,v $
 *   $Revision: 1.1.1.1 $
 *   $Date: 2001/07/31 22:23:08 $
 *   $Author: rami $
 *
 */

int main(int argc, char* argv[])
{
  try {
    Agent* theAgent = new Agent (argc, argv);
    theAgent->run();
  }
  catch (Exception e) {
    cerr << "Error: " << e.toString() << endl;
    ERROR( e.toString() );
    globalLog->flushStampStream();
    globalLog->flushInfoStream();
    exit( -1 );
  }
  return 0;
}

Agent::Agent(int argCount, char* argv [])
{

  teamName = "AT_Humboldt";
  playerNumber = UNDEF_INT;
  isGoalie = NO;
  serverName = "localhost";
  serverConfigurationFile = "server.conf";
  agentConfigurationFile = "agent.conf";
  extendedConfigurationFiles = vector<Ath98String>();
  isHelpWanted = NO;

  readParameter(argCount, argv);
  if( isHelpWanted ) {
    showHelp( Ath98String(argv[0]) );
    exit(0);
  }
  // der Log-Konstruktor wird erst hier gerufen,
  // damit er den Teamnamen mitnehmen kann,
  // um daraus einen Logdateinamen zu erzeugen
  globalLog = new Log(  1 );
  globalConf = new Configuration();

  globalConf->readConfigurationFile( serverConfigurationFile );
  globalConf->readConfigurationFile( agentConfigurationFile );
  if( extendedConfigurationFiles.size() != 0 ) {
    for( unsigned int i = 0;  i < extendedConfigurationFiles.size();  i++ ) {
      globalConf->readConfigurationFile( extendedConfigurationFiles[i] );
    }
  }
  globalLog->init( globalConf );

  globalConf->logCurrentParameterValues();
  // TMINFO GLOBALLOG BRAUCHT OPERATOR
  // INFO( Log::CONFIGURATION  ,  2  ,  idString() , globalLog->infoStream <<
  // globalLog ) ;

  globalFormation = new Formation();
  globalFormation2 = new Formation2();

  serverPort = globalConf->portNumber;
  startComponents();
}

void Agent::run() {
  theDeliberator->deliberate();
}


void Agent::readParameter( int argCount, char* argv [] ) {
  int argIndex;
  Ath98String parameter;

  for(argIndex = 1;  argIndex < argCount;  argIndex++) {
    parameter = Ath98String( argv[argIndex] );
    // TMINFO
    // Absturz in Array-Template, wenn Kommandozeilenparameter bergeben werden
    // Ursache : LogObjekt existiert noch nicht
    // INFO( Log::CONFIGURATION  ,  5  ,  "Agent" , globalLog->infoStream << 
    // " Parameter: " <<    parameter  <<  " \n"  ) ; 
  
    if( parameter == "-goalie" ) {  // einstellige Parameter 
      isGoalie = YES;  
    } 
    else if( parameter == "-help" || parameter == "-h" ) { 
      isHelpWanted = YES;
      return;  
    } 
    else {  // zweistellige Parameter 
      argIndex++;  
      if( !(argIndex < argCount) ) {
	throwException(InitFailedException( "Missing parameter after option " + 
	    Ath98String( argv[argIndex-1] ) + ". Aborting initialization."));  
      }
       
      if( parameter == "-team" ) { 
	teamName = Ath98String( argv[argIndex] );  
      } 
      else if( parameter == "-number" ) { 
	playerNumber = atoi( argv[argIndex] );  
      } 
      else if( parameter == "-host" ) { 
	serverName = Ath98String( argv[argIndex] );  
      } 
      else if( parameter == "-port" ) { 
	serverPort = atoi( argv[argIndex] );
      } 
      else if( parameter == "-sfile" ) {
	serverConfigurationFile = Ath98String( argv[argIndex] );
      } 
      else if( parameter == "-afile" ) { 
	agentConfigurationFile = Ath98String( argv[argIndex] );
      } 
      else if( parameter == "-efile" ) { 
	extendedConfigurationFiles.push_back( Ath98String( argv[argIndex] ) );
      } 
      else { 
	argIndex--;
        cerr << "Unknown option in Commandline : " << 
	  Ath98String( argv[argIndex] ) << " is ignored." << endl;  
      }  
    } 
  }  // Ende for-Schleife 
} 
 
void Agent::initRandomNumberGenerator() 
{
  static long probabilityState[32] = { 
    3, 
    0x9a319039, 0x32d9c024, 0x9b663182, 0x5da1f342, 
    0x7449e56b, 0xbeb1dbb0, 0xab5c5918, 0x946554fd, 
    0x8c2e680f, 0xeb3d799f, 0xb11ee0b7, 0x2d436b86, 
    0xda672e2a, 0x1588ca88, 0xe369735d, 0x904f35f7, 
    0xd7158fd6, 0x6fa6f051, 0x616e6b96, 0xac94efdc, 
    0xde3b81e0, 0xdf0a6fb5, 0xf103bc02, 0x48f340fb,
    0x36413f93, 0xc622c298, 0xf5a42ab8, 0x8a88d77b, 
    0xf5ad9d0e, 0x8999220b, 0x27fb47b9 
  };  
  unsigned int seed = 1;  
  int n = 128;  
   
  initstate(seed, (char *)probabilityState, n);  
  setstate( (char *)probabilityState );
  // setstate( (const char *)probabilityState );
  // Cast wegen altmodischer Interfacedeklaration von initstate und setstate. 
} 


void Agent::startComponents() {
  bool isFirstHalf = (playerNumber == UNDEF_INT);
  initRandomNumberGenerator();

  // heterogene Spieler
  globalPlayerTypes = new PlayerTypes;
  Message::setTeamname( teamName );

  // Sensoren und Effektoren
  theCommunicator = new Communicator( serverName, serverPort );
  theSensors = new Sensors( theCommunicator );
  theEffectors = new Effectors( theCommunicator );

  // globales Weltmodell, Planen und Ueberlegen
  globalWorldModel = new WorldModelInterface( theSensors );

  // starten der Analyzerkomponenten
//   globalDelib = new Delib( dynamic_formation_442 );
//   globalGridPosition = new GridPosition;
//   globalAnalyzer = new Analyzer;

  globalSkills = new Skills();
  globalSay = new Say();
  theDeliberator = new Deliberator();
  globalNeck = new Neck ();

  globalTimer = new Timer( theSensors, theEffectors,
			   theDeliberator, globalNeck, globalSay );

  // Feld betreten
  if (isFirstHalf) {
    theEffectors->enterField( teamName, isGoalie );
  }
  else {
    theEffectors->reEnterField( teamName, playerNumber, isGoalie );
  }

  // Logfile anlegen (benoetigt die Spielernummer)
  // muss bereits hier erfolgen, da eine der folgenden Klassen etwas ins
  // Log schreibt
  if ( !globalLog->isFileReady() ) {
    globalLog->createFiles( globalWorldModel->playerNumber(), teamName );
  }

  // der Deliberator darf erst initialisiert werden, wenn es bereits
  // einen Timer gibt, weil im Konstruktor von PlaySoccerIntention
  // der Timer aufgerufen wird.
  theDeliberator->initialize();
}

void Agent::showHelp( Ath98String programName ) {
  cout << endl << "*******  AT Humboldt 2000  *******" << endl << endl;
  cout << "Humboldt University Berlin" << endl;
  cout << "Department of Computer Science" << endl;
  cout << "AI Domain" << endl;
  cout << endl << endl;
  cout << "Usage: " << programName << " [options]" << endl;
  cout << "Options:" << endl;

  cout << "  -h, -help          " << "\t" <<
    "this help info" << endl;
  cout << "  -host HOSTNAME     " << "\t" <<
    "the host where the server runs (localhost)" << endl;
  cout << "  -team TEAMNAME     " << "\t" <<
    "the name of the team (AT_Humboldt)" << endl;
  cout << "  -number NUMBER     " << "\t" <<
    "the number of the player (only needed for the second half)" << endl;
  cout << "  -goalie            " << "\t" <<
    "if the player should act as an goalie" << endl;
  cout << "  -port PORT_NUMBER  " << "\t" <<
    "the port at which the server waits for the agents (6000)" << endl;
  cout << "  -sfile SERVER_FILE " << "\t" <<
    "the configuration file of the soccerserver (server.conf)" << endl;
  cout << "  -afile AGENT_FILE  " << "\t" <<
    "the configuration file of the agent (agent.conf)" << endl;
  cout << "  -efile AGENT_FILE  " << "\t" <<
    "the extended configuration file of the agent (extended.conf) " <<
    "- multiple use allowed" << endl;
}
