#include <strstream.h> 
#include <string> 
#include <sys/param.h>
#include <netinet/in.h>
#include <time.h> 
 
#include "Log.h" 
#include "Ath98String.h" 
#include "Configuration.h" 

#include "../coach/serverInterface/ServerTypes.h"

#include "../worldmodel/Situation.h"
#include "../worldmodel/WorldModelInterface.h"

 
Log *globalLog;  
 

Log::Log( int debugLevel = 0 ) 
{ 
  infoStream.clear();
  stampStream.clear();
  theFileStreamsExists = NO;
  theModeActivation = vector<int>(MODE_COUNT);
  for( int i = 0;  i < MODE_COUNT;  i++ ) { 
    theModeActivation.push_back( debugLevel );
  } 
} 


 
 Log::~Log( ) 
 { 
   delete theConstantTargetStream;  
   delete theVariableTargetStream;  
   fclose(rec_fp);
 } 
 
void Log::init( Configuration *aConfigurationObject ) 
{ 
  int maxActivation = 0;  
  for( int i = 1;  i < MODE_COUNT;  i++ ) { 
    theModeActivation[i] = aConfigurationObject->debugModes[i-1];  
    maxActivation = max( maxActivation, theModeActivation[i] );  
  } 
  theModeActivation[0] = maxActivation;  
} 
	 
bool Log::isActive( int mode, int debugLevel )  
{ 
  if( debugLevel <= theModeActivation[mode] ) { 
    return YES;  
  } 
  else { 
    return NO;  
  } 
} 
   
void Log::createFiles( const int playerNumber , const  Ath98String& teamName )
{
  // falls der Agent ohne Soccerserver gestartet wird, ist playerNumber == 5 
  if (playerNumber != UNDEF_INT) {
    theFileStreamsExists = YES;
    theConstantTargetStream = new ofstream ( generateFileName( CONSTANT_LOG, teamName,  playerNumber ) );   
    if( theConstantTargetStream == NULL ){ 
      cerr << "Fehler beim Oeffnen des ConstantStream \n"; 
      theFileStreamsExists = NO;
    }

    theVariableTargetStream = new ofstream ( generateFileName( VARIABLE_LOG, teamName,  playerNumber ) );   
    if( theVariableTargetStream == NULL ) { 
      cerr << "Fehler beim Oeffnen des VariableStream \n"; 
      theFileStreamsExists = NO;
    }
/*
    if( globalConf->situationLog == YES ){
      openSituationLogFile( playerNumber, teamName );
    }
*/
  }
}


void Log::openSituationLogFile( const int playerNumber , const  Ath98String& teamName )
{
  char buffer[5];
  
  //ffnen des Files
  if( (rec_fp = fopen( generateFileName( SITUATION_LOG, teamName, playerNumber ), "w" )) == NULL ){
    cerr << "Can't open record File for LogSituation!\n";
  }
/*
  // Header fr das gamerec
  buffer[0] = 'U';
  buffer[1] = 'L';
  buffer[2] = 'G';
  buffer[3] = 1;  // version

  // Header schreiben
  fwrite( buffer, 4, 1, rec_fp );
  fflush(rec_fp);
*/
}


char* Log::generateFileName(int LogType, const Ath98String& teamName, const int playerNumber ) const  
{ 
  ostrstream os; 
  // String zusammensetzen 
  os << teamName << "_" << playerNumber;
  if (LogType==CONSTANT_LOG) { 
    os << ".const.log"; 
  }   
  if (LogType==VARIABLE_LOG) { 
    os << ".var.log"; 
  }  
  if (LogType==SITUATION_LOG) { 
    os << ".situation.log"; 
  }  
  os << ends; 
  char* s = os.str() ; 
  return s; 
} 
 
 
void Log::flushInfoStream()  
{    
  // nach N.Josuttis "C++ STL" Seite 493 
  // String im os mit "ends" abschliessen 
  infoStream << ends; 
  // aus dem Stream auf den filestream schreiben 
  *theVariableTargetStream << infoStream.str(); 
  // Stream entsperren 
  infoStream.rdbuf()->freeze(false); 
  // Schreibpostion auf Anfang setzen 
  infoStream.seekp(0, ios::beg); 
} 


void Log::flushStampStream()  
{    
  // nach N.Josuttis "C++ STL" Seite 493 
  // String im os mit "ends" abschliessen 
  stampStream << ends; 
  // aus dem Stream auf den filestream schreiben 
  *theConstantTargetStream << stampStream.str(); 
  // Stream entsperren 
  stampStream.rdbuf()->freeze(false); 
  // Schreibpostion auf Anfang setzen 
  stampStream.seekp(0, ios::beg); 
} 


void Log::writeSituationLog( const Situation& aSituation )
{
  // Joschas neue "grafische" Ausgabe

  const int width = 150;
  const int height = width * 68 / 105 /2; // Ratio von Hhe zu Breite sollte fest sein

  char field[width+2][height+2];

  for( int i = 0; i <= width; ++i ){
    for( int j = 0; j <= height; ++j ){
      field[i][j] = ' ';
    }
  }

//  const ModelledBall& cball = globalWorldModel->correctSituation().ball();
  const ModelledBall& oball = aSituation.ball();

   
//  field[(int) ((cball.pos().x() + 52.5)*width/105.0)]
//       [(int) ((cball.pos().y() + 34.0)*height/68.0)] = 'O';

//  field[(int) ((oball.pos().x() + cball.speed().x()*10.0 + 52.5)*width/105.0)]
//       [(int) ((oball.pos().y() + cball.speed().y()*10.0 + 34.0)*height/68.0)] = '.';
         
  if( oball.isValid() == YES ){
    field[(int) ((oball.pos().x() + 52.5)*width/105.0)]
         [(int) ((oball.pos().y() + 34.0)*height/68.0)] = 'o';
    if( oball.speed().length() == UNDEF_DOUBLE ) {
    field[(int) ((oball.pos().x() + 52.5)*width/105.0)]
         [(int) ((oball.pos().y() + 34.0)*height/68.0)] = '~';
    }
    else{
      field[(int) ((oball.pos().x() + oball.speed().x()*10.0 + 52.5)*width/105.0)]
           [(int) ((oball.pos().y() + oball.speed().y()*10.0 + 34.0)*height/68.0)] = ',';
    }
  }

  int x, y;
  
  for( int i = 0; i < aSituation.playerCount(); ++i ){
    const ModelledPlayer& player = aSituation.playerWithIndex(i);
    Vector bD = player.pos() + Vector(5.0,player.bodyDir(),POLAR,RADIAN);
    Vector fD = player.pos() + Vector(5.0,player.faceDir(),POLAR,RADIAN);

    x = (int) ((player.pos().x() + 52.5)*width/105.0);
    y = (int) ((player.pos().y() + 34.0)*height/68.0);

    int bx = (int) ((bD.x() + 52.5)*width/105.0);
    int by = (int) ((bD.y() + 34.0)*height/68.0);
    int fx = (int) ((fD.x() + 52.5)*width/105.0);
    int fy = (int) ((fD.y() + 34.0)*height/68.0);

    if( x < 0 ){
      x = 0;
    }
    if( x > width ){
      x = width;
    }
    if( y < 0 ){
      y = 0;
    }
    if( y > height ){
      y = height;
    }

    if( field[x][y] == 'o' ){
      if( player.team() == OUR_TEAM_ID ){
        field[x][y] = '';
      }
      else if( player.team() == THEIR_TEAM_ID ){
        field[x][y] = '';
      }
      else{
        field[x][y] = '@';
      }      
    }
    else if( field[x][y] == ' ' ){
      if( player.team() == OUR_TEAM_ID ){
        if( player.isGoalie() == YES ){
          field[x][y] = '{';
        }
        else if( player.id() != UNDEF_TEAMMATE_ID ){
          field[x][y] = 'a' + player.id() - MIN_TEAMMATE_ID;
          if( player.id() == globalWorldModel->id() ){
            field[bx][by] = '';
            field[fx][fy] = '^';
          }
        }
        else{
          field[x][y] = 'x';
        }
      }
      else if( player.team() == THEIR_TEAM_ID ){
        if( player.isGoalie() == YES ){
          field[x][y] = '}';
        }
        else if( player.id() != UNDEF_OPPONENT_ID ){
          field[x][y] = 'A' + player.id() - MIN_OPPONENT_ID;
        }
        else{
          field[x][y] = 'X';
        }
      }
      else{
        field[x][y] = '*';
      }  
    }
    else{
      field[x][y] = '#';
    }
  }

  cout << "\x1b[fSimstep : " << aSituation.simStep() << endl;

  for( int i = 0; i <= width + 2; ++i ){
    cout << '-';
  }
  cout << endl;

  for( int i = height; i >= 0; --i ){
    cout << '|';
    for( int j = 0; j <= width; ++j ){
      cout << field[j][i];
    }
    cout << '|' << endl;
  }  

  for( int i = 0; i <= width + 2; ++i ){
    cout << '-';
  }
  cout << endl;

  
cout << "##################################################\n";
cout << aSituation.simStep() << endl;


  
/* FUNKTIONIERT NOCH NICHT !!!!!!!!!!!!!!
  dispinfo_t dispinfo;

  dispinfo.mode = htons(SHOW_MODE);

  dispinfo.body.show.pmode = htons(PM_Null);

  char teamNameLeft[16] = "LEFT";
  char teamNameRight[16] = "RIGHT";
  
  dispinfo.body.show.team[0].name = teamNameLeft;
  dispinfo.body.show.team[0].score = htons(0);   // Achtung Score nicht aktuell
  dispinfo.body.show.team[1].name = teamNameRight;
  dispinfo.body.show.team[1].score = htons(0);   // Achtung Score nicht aktuell

  dispinfo.body.show.pos[0].enable = htons(0x00); // DISABLE
  dispinfo.body.show.pos[0].side = htons(0);      // NEUTRAL
  dispinfo.body.show.pos[0].unum = htons(0);
  dispinfo.body.show.pos[0].angle = htons((short) ( 0.0 * 16.0));
  dispinfo.body.show.pos[0].x = htons((short) ( aCorrectSituation.ball().pos().x() * 16.0 ));
  dispinfo.body.show.pos[0].y = htons((short) ( aCorrectSituation.ball().pos().y() * 16.0 ));

  for( int i = 0; i < aCorrectSituation.playerCount(); ++i ){
    dispinfo.body.show.pos[i+1].enable = htons((aCorrectSituation.playerWithIndex( i ).isGoalie() == YES) ? 
                                               (0x08) : (0x01));
    dispinfo.body.show.pos[i+1].side = htons(( aCorrectSituation.playerWithIndex( i ).team() == OUR_TEAM_ID ) ? 
                                             ( (globalWorldModel->side() == LEFT_SIDE_ID) ? (1) : (-1) ) :   // LEFT = 1
                                             ( (globalWorldModel->side() == LEFT_SIDE_ID) ? (-1) : (1) ));    // RIGHT = -1
    dispinfo.body.show.pos[i+1].unum = htons(aCorrectSituation.playerWithIndex( i ).id() - 
                                             (aCorrectSituation.playerWithIndex( i ).team() == OUR_TEAM_ID) ?
                                             (MIN_TEAMMATE_ID) : (MIN_OPPONENT_ID));
    dispinfo.body.show.pos[i+1].angle = htons((short) ( 0.0 * 16.0));
    dispinfo.body.show.pos[i+1].x = htons((short) ( aCorrectSituation.playerWithIndex( i ).pos().x() * 16.0 ));
    dispinfo.body.show.pos[i+1].y = htons((short) ( aCorrectSituation.playerWithIndex( i ).pos().y() * 16.0 ));
  }
  
  dispinfo.body.show.time = htons((short) aCorrectSituation.simStep());

  fwrite( (char*) &dispinfo.mode, sizeof(dispinfo.mode), 1, rec_fp );
  fwrite( (char*) &dispinfo.body.show, sizeof(dispinfo.body.show), 1, rec_fp );
  fflush(rec_fp);
*/
}

 
Ath98String Log::toString() const 
{ 
  return Ath98String( Ath98String( "Log Information: \tModi: (" ) + 
                 "\nModus: Communicator: " +  
                 Ath98String( theModeActivation[COMMUNICATOR] ) + 
                 "\nModus: Configuration: " + 
                 Ath98String( theModeActivation[CONFIGURATION] ) + 
                 "\nModus: Sensors: " + 
                 Ath98String( theModeActivation[SENSORS] ) + 
                 "\nModus: Effectors: " + 
                 Ath98String( theModeActivation[EFFECTORS] ) + 
                 "\nModus: Situation: " + 
                 Ath98String( theModeActivation[SITUATION] ) + 
                 "\nModus: WorldModel: " + 
                 Ath98String( theModeActivation[WORLD_MODEL] ) + 
                 "\nModus: Timer: " + 
                 Ath98String( theModeActivation[TIMER] ) + 
                 "\nModus: Deliberator: " + 
                 Ath98String( theModeActivation[DELIBERATOR] ) + 
                 "\nModus: Constraints: " + 
                 Ath98String( theModeActivation[CONSTRAINTS] ) + 
                 "\nModus: Analyzer: " + 
                 Ath98String( theModeActivation[ANALYZER] ) + 
                 "\nModus: ActiveOptions: " + 
                 Ath98String( theModeActivation[ACTIVE_OPTIONS] ) + 
                 "\nModus: ModelledObjects: " + 
                 Ath98String( theModeActivation[MODELLED_OBJECTS] ) + 
                 "\nModus: Skills: " + 
                 Ath98String( theModeActivation[SKILLS] ) + 
                 "\nModus: Role: " +
                 Ath98String( theModeActivation[ROLE] ) + 
                 " )\n" )
                 ;  
} 
  
