#ifndef _LOG_H_
#define _LOG_H_

#include <fstream.h>
#include <stdio.h>
#include <string>
#include <strstream>
#include <vector>

#include "Ath98String.h" // TM
class Situation;

#define ConfigurationTest
//#define SensorTest
 
class Ath98String;
class Configuration;

/**
  Diese Klasse ffnet beim Start des Agenten zwei Filestreams. 
  Der STAMPSTREAM bekommt Informationen in einem festgelegten Format. Das 
  Makro STAMP ruft Memberfunktionen der Klassen toStampStream(), die 
  unkommentierte Werte ausgeben, diese werden vom Logbrowser verarbeitet. 
  Der INFOSTREAM bekommt die Informationen, die in agent.conf eingeschaltet
  wurden. Das Makro INFO ruft die Ausgabeoperatoren << der Klassen, die 
  kommentiert Werte ausgeben. In den Optionen rufen diese Operatoren virtuelle
  Funktionen toInfoStream(), damit die richtigen Ausgaben der speziellen 
  Option auf dem Stream ankommen.
  Im Makefile kann man beide STREAMS abschalten.
                      INFO                STAMP

   Schalter           DEBUG_ON            STAMP_ON
   Datei              *.var.log           *.const.log
   Ausgabe-operator   <<                  toStampStream()
   Ausgabeart	      kommentierte W.     unkommentierte Werte
   Verwendung         Fehlersuche         Logbrowser

  Abh\"angig vom eingestellten DebugModus und dazu geh\"origen  DebugLevel
  werden Zeilen ausgegeben oder nicht. \\
  Der DebugModus setzt sich aus mehreren Modi zusammen, deren DebugLevel
  im Konfigurationsfile des Agenten festgelegt werden.
  Warnungen und Fehler (Debug-Informationen der Level 0 und 1) werden immer
  ausgegeben.

  Modi:
  \begin{tabular}{ll}
  Communicator & Debug-Informationen des Communicators \\
  Sensor & Debug-Informationen der Sensoren \\
  ... & ...
  \end{tabular}

  DebugLevel:
  \begin{tabular}{ll}
  0     & Fehler ausgeben \\
  1     & zus\"atzlich Warnungen ausgeben \\
  2-4   & zus\"atzlich Funktionseintritte und Funktionsbeendigungen \\
  5-7   & zus\"atzlich Ergebnisse von Funktionen (wichtige Informationen,
            wie see, hear, utility-Werte) \\
  8-10  & zus\"atzlich Informationen zu Zwischenschritten \\
  11-15 & zus\"atzlich weitere Informationen
  \end{tabular} \\
  Wichtige Funktionseintritte sollten beispielsweise mit einem 
  Level von 2, unwichtige Funktionseintritte dagegen mit einen Level von 4
  aktiviert werden. \\

  Beispiel zur Benutzung:
  \begin{verbatim}
  INFO( Log::WORLD_MODEL, 7, "Hallo" )
  \end{verbatim}
  In diesem Fall wird "Hallo" ausgegeben, wenn der DebugLevel f\"ur das
  Weltmodell auf 7 oder gr\"o\3er gesetzt wurde.

  Wenn man viele Debug-Ausschriften ausgeben m\"ochte, dann bietet sich 
  diese Benutzung an:
  \begin{verbatim}
  if ( IS_ACTIVE( Log::WORLD_MODEL, 7 ) ) {
    globalLog->info( "Hallo" );
    globalLog->info( "nochmal Hallo" );
  }
  \end{verbatim}

  Die Makros schreiben immer ins globalLog.

  {\small {\bf Quelltext}

  [\URL[Header]{src/Log.h} | [\URL[Implementierung]{src/Log.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: Log.h,v $
  *   $Revision: 1.1.1.1 $
  *   $Date: 2001/07/31 22:23:08 $
  *   $Author: rami $ 
  *
  **************************************************
  \end{verbatim} }*/
class Log
{
public:

  /** @name M\"ogliche Modi f\"urs Logging. */
  //@{
    /// beliebiger Modus
    static const int ANY = 0;
    /// Modus f\"ur den Communicator
    static const int COMMUNICATOR = 1;
    /// Modus f\"ur die Configuration
    static const int CONFIGURATION = 2;
    /// Modus f\"ur die Sensoren
    static const int SENSORS = 3;
    /// Modus f\"ur die Effektoren
    static const int EFFECTORS = 4;
    /// Modus f\"ur die Situation
    static const int SITUATION = 5;
    /// Modus f\"ur das Weltmodel
    static const int WORLD_MODEL = 6;
    /// Modus f\"ur den Timer
    static const int TIMER = 7;
    /// Modus f\"ur den Deliberator
    static const int DELIBERATOR = 8;
    /// Modus f\"ur die Static Constraints
    static const int CONSTRAINTS = 9;
    /// Modus f\"ur die PassiveOptions
    static const int ANALYZER = 10;
    /// Modus f\"ur die ActiveOptions
    static const int ACTIVE_OPTIONS = 11;
    /// Modus f\"ur alle ModelledObjects
    static const int MODELLED_OBJECTS = 12;
    /// Modus f\"ur die Skills
    static const int SKILLS = 13;    
    /// Modus f\"ur die Intentions
    static const int INTENTIONS = 14;    
    /// Modus f\"ur die Conditions
    static const int CONDITIONS = 15;
    /// Modus f\"ur den Neck
    static const int NECK = 16;
    /// Modus f\"ur die Targets
    static const int TARGET = 17;
    /// Modus f\"ur die Roles
    static const int ROLE = 18;

    /// Anzahl aller Modi
    static const int MODE_COUNT = 19;

    // Konstante fr generateFileName 
    static const int CONSTANT_LOG = 0;

    // Konstante fr generateFileName 
    static const int VARIABLE_LOG = 1;

    static const int SITUATION_LOG = 2;

    // StringStream als Puffer fr INFO
    // dynamisch oder statisch anlegen
     ostrstream infoStream;
     
    // StringStream als Puffer fr STAMP
    // dynamisch oder statisch anlegen
     ostrstream stampStream;

     FILE* rec_fp;
  //@}

  /** @name Log anlegen und initialisieren */
  //@{
    /** Konstruktor. Setzt das DebugLevel und den Ausgabe-Stream. Weiterhin
      aktiviert er alle Modi f\"ur die Ausgabe mit dem angegebenen
      DebugLevel. */
    Log( int debugLevel = 0 );

     /** Destruktor */
     ~Log( );
  
    /** Initialisiert das Log-Objekt neu. Initialisiert, mit dem als Parameter
      angegebenen Configurationsobjekt, die Debug-Level f\"ur alle Ausgabemodi.
      @param aConfigurationObject Pointer auf das Configuration-Objekt. */
    void init( Configuration *aConfigurationObject );
  //@}

  /** @name Anfragen ob Modus mit DebugLevel aktiv */
  //@{
    /** Fragt ob der als Paramter angegebene Modus mit dem angebenen 
      DebugLevel aktiv ist. Der Modus ist aktiv, wenn sein eingestellter
      DebugLevel gr\"o\3er oder gleich dem angegebenen DebugLevel ist.
      Diese Methode sollte in einer if-Anweisung, die \Ref{info} Anweisungen
      kapseln.
      @param mode Modus f\"ur den die Aktivit\"at erfragt wird
      @param debugLevel Debug-Level f\"ur den die Aktivit\"at erfragt wird */
    bool isActive( int mode, int debugLevel );
  //@}

  /** @name Ins Log schreiben. */
  //@{
     /** Der infoStream wird auf den variablen FileStream geschrieben */
     void flushInfoStream(); 
  
     /** Der stampStream wird auf den konstanten FileStream geschrieben */
     void flushStampStream(); 

     /** Schreibt das SituationLog */
     void writeSituationLog( const Situation& aSituation );
  //@}

  /** @name Log-Makros */
  // TMINFO fuer MAKROTEST g++ -E FreekickOption.cc > FKO.s.cc
  //@{
    #if DEBUG_ON==1
    
    /** Fragt, ob ein Debug-Mode aktiviert ist.
      Wenn ja, wird die der folgende Block ausgefuehrt, andernfalls nicht. 
      @param debugMode Modus f\"ur den die Aktivit\"at erfragt wird
      @param debugLevel Debug-Level f\"ur den die Aktivit\"at erfragt wird */
    #define IS_ACTIVE( debugMode, debugLevel )  \
      ( globalLog->isActive( (debugMode), (debugLevel) ) )

    /** Meldet Fehler. Eine Ausgabe erfolgt immer im \Ref{globalLog}. */
    #define ERROR( message ) \
    { \
      globalLog->infoStream << "ERROR : " << message << "\n"; \
    }

    /** Meldet Warnung. Eine Ausgabe erfolgt im \Ref{globalLog}, wenn der
      angegebene Modus debugMode
      im Log-Objekt mit einen debugLevel von 1 oder h\"oher eingestellt ist.
      @param debugMode Der Modus bei dem ausgegeben werden soll.
      @param message Die Nachricht die ausgegeben werden soll. */
    #define WARNING( debugMode, message ) \
    { \
      if( globalLog->isActive( debugMode, 1 ) ) { \
        globalLog->infoStream << "WARNING : " << message ; \
      } \
    }

    /** Debug-Info. Eine Ausgabe erfolgt im \Ref{globalLog}, wenn der
      angegebene Modus debugMode
      im Log-Objekt mit den angegebenen debugLevel aktiv ist.
      Der Modus ist aktiv, wenn sein eingestellter
      DebugLevel gr\"o\3er oder gleich dem angegebenen DebugLevel ist.
      @param debugMode Der Modus bei dem ausgegeben werden soll.
      @param debugLevel Der Debug-Level ab dem ausgeben werden soll.
      @param call Der Aufruf des StreamOperators mit Nachrichten. */

     // TMINFO
     // der Aufruf von globalLog->flushInfoStream() erfolgt
     // nicht mehr hier im Makro INFO sondern
     // in Agent::run() vor dem Aufruf von pause()
     // Vorteil : es wird also seltener geflusht
     // Nachteil : der streambuffer  wird grer
     // fr zuknftigen VarLogBrowser vor call-Zeile einfgen :
     // globalLog->infoStream << idString << " " << debugLevel << " ";
     // flush nur fuers Debugging : (hinter call einfuegen)
     // if (globalLog->isFileReady() ) { globalLog->flushInfoStream(); }

    #define INFO( debugMode, debugLevel, idString, call ) \
    { \
      if( globalLog->isActive( debugMode, debugLevel ) ) { \
       call; \
      } \
    }

     #else
     #define ERROR( message )
     #define WARNING( debugMode, message )
     #define INFO( debugMode, debugLevel, idString, call )
     #define IS_ACTIVE( debugMode, debugLevel ) (NO)
     #endif

    #if STAMP_ON==1
     /** Log-Info mit festem Format */
    #define STAMP(call) \
    { \
       call; \
    }
    #else
    #define STAMP(call) 
    #endif 
  //@}

     /** liefert Dateinamen fr die Logfiles  */  
     char* generateFileName(const int, const Ath98String&, const int ) const;
     
     /** liefert Status der Logfiles */
     bool isFileReady() const;
     
     /** Erzeugt die Logfiles */
     void createFiles(const int, const Ath98String&);


     /** Beschreibung als Ath98String fuer die Ausgabe. */
     Ath98String toString() const;
     
 private:
     /** Array, welches f\"ur jeden Modus enth\"alt, wie stark er aktiviert
	 wurde. */
     vector<int> theModeActivation;
     
     /** Der FileStream auf dem die Ausgabe des variablen Logs erfolgt. */
     ofstream *theConstantTargetStream;
     
     /** Der FileStream auf dem die Ausgabe des konstanten Logs erfolgt. */
     ofstream *theVariableTargetStream;

     /** Flag das die Existenz der FileStreams kennt  */
     bool theFileStreamsExists;

     void openSituationLogFile( const int playerNumber , const  Ath98String& teamName );
};

extern Log *globalLog; 

inline bool Log::isFileReady() const
{
  return theFileStreamsExists;
}

#endif


