#include <sys/time.h>

#include "Timer.h"
#include "Deliberator.h"

Timer *globalTimer;

/* 
 * Aktueller Stand 
 * 
 *   $RCSfile: Timer.cc,v $ 
 *   $Revision: 1.1.1.1 $ 
 *   $Date: 2001/07/31 22:23:08 $ 
 *   $Author: rami $  
 * 
 */ 

Timer::Timer( Sensors *sensors, Effectors *effectors, Deliberator *deliberator,
              Neck *neck, Say *say ) :
  theSensors( sensors ), theEffectors( effectors ),
  theDeliberator( deliberator ), theNeck( neck ), theSay( say )
{
  theIdString = "Timer";

  /// Parameter, auch fuer Configfile
  controllerTimeStep = 3; 

  // diese Werte existieren in der ServerConfig ...=> evtl nur Verzgerung 
  timeBetweenTwoActions = 2;//globalConf->receiveStep + 10;
  timeBetweenTwoSayActions = 
      globalConf->hearCapacityDecay * globalConf->simulatorStep + 10; // 210

  systemDependMinSleepTime = 20 ;
  
  /// Die "20" sollte mit ins Configfile
  /// Kay: geandert auf 25
  actionDeadline = globalConf->simulatorStep - 20;

  theLastMainAction = Action( EMPTY_ACTION , 0 );
  
  
  

  setConstructionTime();
  
  now = getAgentTime();
  timeOfLastControllerCall = now;
  timeOfLastSense = now;
  timeOfLastButOneSense = now;
  timeOfLastSeeInfo = now;
  timeOfLastSimStep = now;
  simStep = 0;
  timeOfLastSentAction = now;
  timeOfLastSayAction = now;
  timeOfLastStampFlush = now;
  timeOfLastInfoFlush = now;
  
  theIsNewInfo = NO;
  haveToSendMainAction = NO;
  haveToSendViewAction = NO;
  haveToSendNeckAction = NO;
  didSentMainAction = NO;
  didSentViewAction = NO;
  didSentNeckAction = NO;
  didSentSayAction = NO;
  bewareNextMainAction = NO;


  INFO( Log::TIMER  ,  2  ,  idString() ,
        globalLog->infoStream << " TIMER STARTED\n"  ); 

}
 

Timer::~Timer() {
}


bool Timer::controller() {
  now = getAgentTime();
 
  // DEBUG Infos -> sollten auch nur dann mit kompiliert werden
  if ( now - timeOfLastControllerCall < 1 ) 
    {
      WARNING( Log::TIMER , "Controller zu OFT aufgerufen\n"  );  
    }
  else 
    if ( now - timeOfLastControllerCall > 20 )
      {
        WARNING( Log::TIMER , "Controller zu SELTEN aufgerufen\n"  ); 
      }
  // wird nur fuer die Debuginformation benutzt
  timeOfLastControllerCall = now;


  if ( ( (now-timeOfLastSense) <= controllerTimeStep) &&
       ( (now-timeOfLastSimStep) <= actionDeadline ) ) 
    {   
      return NO;
    }

  theSensors->initSensors(); 
  INFO( Log::TIMER ,  10  , idString() , globalLog->infoStream << 
        "    Sensors started at time: " << now <<  " \n"  ) ; 
  theSensors->sense();
  now = getAgentTime();
  INFO( Log::TIMER  ,  10  ,  idString() , globalLog->infoStream << 
	"    Sensors ended at time: " << now <<  " \n"  ) ; 
  timeOfLastButOneSense = timeOfLastSense;
  timeOfLastSense = now;

  theIsNewInfo = theSensors->isNewInfo();
  if ( theSensors->isNewFullstateInfo() == YES) {
    globalWorldModel->integrateCorrectSensorInfo();
  }
  if ( theIsNewInfo == YES )  
    {  
      INFO( Log::TIMER  ,  5 ,  idString() , globalLog->infoStream << 
            "    New Sensor Info  at time: " << now <<  " \n"  ) ; 
      //Weltmodell updaten
      INFO( Log::TIMER  ,  10  ,  idString() , globalLog->infoStream << 
	          "    WorldModel started at time: " << now <<  " \n"  ) ;
      globalWorldModel->integrateSensorInfo();
      now = getAgentTime();
      INFO( Log::TIMER  ,  8 ,  idString() , globalLog->infoStream << 
	          "    WorldModel ended at time: " << now <<  " \n"  ) ; 

      //wenn neuer Takt, dann Zeit merken, MainAction und NeckAction freigeben
      // neuer Takt: nach 50 ms entweder bei neuem Simstep oder neuer Bodyinfo
      if ( ( (now-timeOfLastSimStep) >= (globalConf->simulatorStep / 2) ) && 
           ( ( theSensors->isNewBodyInfo() == YES ) ||
             ( simStep != theSensors->simStepOfSoccerServer() ) ) )
        {
          simStep = theSensors->simStepOfSoccerServer();
	        // TM
	        INFO( Log::TIMER ,  1, idString() , globalLog->infoStream << 
		      "\n\n NEW SIMSTEP : "<< simStep << " at time: " << now <<  "\n"  ) ;
          INFO( Log::TIMER  ,  5  ,  idString() , globalLog->infoStream << 
          "    new SoccerServer simStep: " << simStep   <<  " \n"  ) ; 
          timeOfLastSimStep = timeOfLastButOneSense;
          // schlimmster Fall: Info kam direkt nach dem vorletzten Sense        
          
         STAMP( globalLog->stampStream << "\n" <<
                 STEP_MARK << simStep);
         STAMP( theSensors->
                lastBodyInfos().toStampStream( globalLog->stampStream ) );
         STAMP( globalWorldModel->toStampStream( globalLog->stampStream ) );
      

          didSentMainAction = NO;
          didSentNeckAction = NO;          
          didSentSayAction = NO;
          haveToSendMainAction = NO;
        }
      if ( theSensors->isNewBodyInfo() == YES )
        {
          theLastBodyInfos =  theSensors->lastBodyInfos();
        }
    
    // MainAction, ViewAction, NeckAction anfordern, ViewAction freigeben
    if ( theSensors->isNewSeeInfo() ) {
      INFO( Log::TIMER  , 6  ,  idString() , globalLog->infoStream << 
	          "   new ViewInfo at time: " << now <<  " \n"  ) ;
      timeOfLastSeeInfo = timeOfLastButOneSense;
      haveToSendViewAction = YES;   
      haveToSendNeckAction = YES;
      didSentViewAction = NO;
    }
  
  }
  // wenn >actionDeadline (75ms) im Takt vergangen sind, MainAction senden
  if ( (now-timeOfLastSimStep) >= actionDeadline ) {
    haveToSendMainAction = YES;
  }


  // hier: Abschicken der Aktion
  // neue Aktion kann geschickt werden
  if ( (now - timeOfLastSentAction) >= timeBetweenTwoActions ) {
    if ( ( haveToSendMainAction == YES ) &&
         ( didSentMainAction == NO ) )
      {
        // was ist, wenn es keine Aktion gibt?
        theLastMainAction = theDeliberator->execute();
        theEffectors->performAction( theLastMainAction, theSensors->lastBodyInfos() );
        INFO( Log::TIMER  ,  5  ,  idString() , globalLog->infoStream << 
        "    MainAction sent started at time: " << now <<  " \n"  ) ;
        // cerr << "MainAction send at: " << now - timeOfLastSimStep << endl;
        haveToSendMainAction = NO;
        didSentMainAction = YES;
        timeOfLastSentAction = getAgentTime();
        return YES;
      }
    
    
    // es gab neue SichtInfos
    if ( ( haveToSendViewAction == YES ) &&   
         // Aktion wurde noch nicht geschickt  
         ( didSentViewAction == NO ) &&
         ( theNeck->needToSendViewCommand() == YES ) )
      {    
      
        theEffectors->performAction( theNeck->getChangeViewCommand() );
        INFO( Log::TIMER  ,  5  ,  idString() , globalLog->infoStream << 
        "    ViewAction sent started at time: " << now <<  " \n"  ) ;
        didSentViewAction = YES;
        haveToSendViewAction = NO;
        timeOfLastSentAction = getAgentTime();
        return YES;
      } 
/*
    // muss bald MainAction geschickt werden?
    if  ( ( didSentMainAction == NO ) || // YES NO ?
          ( now - timeOfLastSimStep) <= (actionDeadline - timeBetweenTwoActions) ) {
      bewareNextMainAction = YES;
    }
    else {
      bewareNextMainAction = NO;
    }
*/
    if ( /*( haveToSendNeckAction == YES ) && */
         // Aktion wurde noch nicht geschickt          
         ( didSentNeckAction == NO ) &&   
         // es gibt ueberhaupt eine Nackenaktion        
         ( theNeck->needToSendNeckCommand() == YES )
      && ( didSentMainAction == YES )
//       && ( bewareNextMainAction == NO )
// 	 &&
         
//          // das Kommando wird im Time-Slot vor der naechsten Sichtinfo 
//          // geschickt ( macht nur Sinn, falls Neck so spaet wie moeglich
//          // vor der naechsten Sichtinfo gesendet werden soll)
//          //spaet genug:
//          ( (now - timeOfLastSeeInfo) >= 
//           (viewTimeStep() - 3 * timeBetweenTwoActions) ) &&   
// //          //frueh genug (noch nicht zu spaet zum abschicken):
//          ( (now - timeOfLastSeeInfo) <=
//          (viewTimeStep() - 1 * timeBetweenTwoActions) )
 ) {    
      theEffectors->performAction( theNeck->getTurnNeckCommand(),
                                   theSensors->lastBodyInfos() );
      INFO( Log::TIMER  ,  5  ,  idString() , globalLog->infoStream << 
            "    NeckAction sent started at time: " << now <<  " \n"  );
      didSentNeckAction = YES;
      haveToSendNeckAction = NO;
      timeOfLastSentAction = getAgentTime();
      return YES;
    }

    if( (didSentSayAction == NO) && (globalWorldModel->simStep() > 0)
// && (didSentMainAction == YES)
) {
      if( ( (globalWorldModel->simStep() % 6 == 0) && (globalWorldModel->id() == 420) ) ||
          ( (globalWorldModel->simStep() % 6 == 2) && (globalWorldModel->id() == 421) ) ||
          ( (globalWorldModel->simStep() % 6 == 4) && (globalWorldModel->id() == 426) ) ) {    
        theEffectors->performAction( globalSkills->saySituationAction(), theSensors->lastBodyInfos() );
        // fuer die Statistik:
        didSentSayAction = YES;
        timeOfLastSayAction = getAgentTime();   
        timeOfLastSentAction = getAgentTime();
        return YES;
      }
    }
/*
    if( didSentMainAction || didSentViewAction || didSentNeckAction || didSentSayAction ){
      return YES;
    }
*/
  }
  // hier erfolgen die Aufrufe von flush() fr die beiden Streams der Klasse Log
  // Intervall fuer Flush im ms :
  // fuer ein ruckfreies Spiel bei sehr umfangreichen Ausgaben empfiehlt sich 100

   if ( (now - timeOfLastInfoFlush) >= 100 ){
     if( globalConf->situationLog ) {
       globalLog->writeSituationLog( globalWorldModel->situation() );
     }
     globalLog->flushInfoStream();
     timeOfLastInfoFlush=getAgentTime();
   }
   if ( (now - timeOfLastStampFlush) >= 5000 ){
     // cerr << " FLUSH_STAMP " << globalWorldModel->playerNumber() << " " << globalWorldModel->teamName() << " " << " AT " << globalTimer->getAgentTime() << endl;
     globalLog->flushStampStream();
     timeOfLastStampFlush=getAgentTime();
   }

  return theIsNewInfo;
}





void Timer::agentSleep()
{
  if ( ( didSentMainAction == NO ) &&
       ( ( getAgentTime() - timeOfLastSimStep ) >=
         ( actionDeadline - systemDependMinSleepTime) ) ) {
    haveToSendMainAction = YES;
  }
  while ( controller() == NO )
    {
      now = getAgentTime();
      usleep(1000);
      INFO( Log::TIMER  ,  7  ,  idString() , globalLog->infoStream <<
	    "    agent slept for " << (getAgentTime() - now) <<  " ms \n"  ) ;

      if ( ( didSentMainAction == NO ) &&
           ( ( getAgentTime() - timeOfLastSimStep ) >=
             ( actionDeadline - systemDependMinSleepTime) ) )
        {
          haveToSendMainAction = YES;
        }
    }
}


