#include "TeamInfo.h"
#include "../../basic/Log.h"
#include "../../worldmodel/Situation.h"
#include "../../worldmodel/WorldModelInterface.h"
#include "stdlib.h"

#define PI 3.14159

#define DEBUGOUTPUT

TeamInfo *globalTeamInfo;

TeamInfo::TeamInfo()
{
  playerInfos=0;
  playerCount=0;
  for( int i =0; i < 12; i++ ) {
    changedOpponentPlayerTypeByNumber[i] = NO;
  }
}

TeamInfo::~TeamInfo()
{
  if (playerInfos) {delete [] playerInfos;};
}

void TeamInfo::PlayerInfo::Initialize(int iIndex)
{
  INFO( Log::SENSORS, 7, idString(), globalLog->infoStream <<
        "Initialising player with index " << iIndex << endl );
  playerIndex=iIndex;
  maxSpeed=0.0f;
  avgDecay=0.0f;
  maxDecay=-1000.0f;
  minDecay=1000.0f;
  DecayCount=0;
  maxKickRange=0.0f;
  lastRotation=-1234.0f;
  dashValue=0.0f;
  lastSpeed=Vector();
  lastPos=Vector();
  playerType=-1;
}

void TeamInfo::PlayerInfo::CalcPlayerMaxSpeed()
{
  const Situation& situation = globalWorldModel->situation();

  float currentSpeed=situation.playerWithIndex(playerIndex).speed().length();

  if (currentSpeed > maxSpeed) {
    maxSpeed = currentSpeed;
  }
}

void TeamInfo::PlayerInfo::CalcPlayerDecay()
{
  if ((lastRotation!=-1234.0f) &&
      (lastSpeed.length()>0.1f))
  {
    const Situation& situation = globalWorldModel->situation();

    float currentRotation=situation.playerWithIndex(playerIndex).bodyDir();
    float currentSpeed=situation.playerWithIndex(playerIndex).speed().length();

    float rotationDelta=lastRotation-currentRotation;
    while (rotationDelta<-PI) {rotationDelta+=2*PI;};
    while (rotationDelta>PI)  {rotationDelta-=2*PI;};
    if (fabs(rotationDelta)>0.1)
    {
      float currentDecay=currentSpeed/lastSpeed.length();

      if (currentDecay>=0.3f) {
        avgDecay=((avgDecay*DecayCount)+currentDecay)/float(++DecayCount);
        minDecay=min(minDecay,currentDecay);
        maxDecay=max(maxDecay,currentDecay);
      };
    };
  };
}

void TeamInfo::PlayerInfo::CalcPlayerDashValue()
{
  if (!lastSpeed.isValid()) {return;};

  const Situation& situation = globalWorldModel->situation();

  // bei Abbremsen -> nicht werten
  if (situation.playerWithIndex(playerIndex).speed().length()<lastSpeed.length()) {
    return;
  };

  float acceleration=(situation.playerWithIndex(playerIndex).speed()-lastSpeed).length();
  if (acceleration>dashValue) {
    dashValue=acceleration;
  };
}

void TeamInfo::PlayerInfo::SafeState()
{
  const Situation& situation = globalWorldModel->situation();

  lastRotation=situation.playerWithIndex(playerIndex).bodyDir();
  lastSpeed=situation.playerWithIndex(playerIndex).speed();
  lastPos=situation.playerWithIndex(playerIndex).pos();
}

void TeamInfo::PlayerInfo::Output()
{
  cout << "player "        << playerIndex   << '\t';
  cout << "maxSpeed="      << maxSpeed      << '\t';
  cout << "avgDecay="      << avgDecay      << '\t';
  cout << "minDecay="      << minDecay      << '\t';
  cout << "maxDecay="      << maxDecay      << '\t';
  cout << "maxKickRange="  << maxKickRange  << '\t';
  cout << "dashValue="     << dashValue     << '\t';
  cout << endl;
}

bool TeamInfo::CouldHaveKicked(int iPlayerIndex)
{
  const Situation& situation = globalWorldModel->situation();

  // bei Spielerrotation -> kein Kick mglich gewesen
  float rotationDelta=playerInfos[iPlayerIndex].lastRotation-situation.playerWithIndex(iPlayerIndex).bodyDir();
  while (rotationDelta<-PI) {rotationDelta+=2*PI;};
  while (rotationDelta>PI)  {rotationDelta-=2*PI;};
  if (fabs(rotationDelta)>0.1f) {
    return false;
  };

  // bei Spielerbeschleunigung -> kein Kick mglich gewesen
  if (situation.playerWithIndex(iPlayerIndex).speed().length()>playerInfos[iPlayerIndex].lastSpeed.length()) {
    return false;
  };

  return true;
}

void TeamInfo::CalcPlayerMaxKickRange(int playerIndex)
{
  // alle checks mit Daten vom letzten Tick machen !!!

  const float playerTestRange=2.5f; // max: player_size+kickable_margin - 0.9+0.3<2.0f

  float playerBallDist=
           (playerInfos[playerIndex].lastPos-lastBallPos).length();


  if ((playerBallDist<playerInfos[playerIndex].maxKickRange) ||       // checken, ob Nhe zum Ball berhaupt grer als die alten Infos
      (playerBallDist>=playerTestRange) ||                            // checken, ob Spieler nahe genug am Ball
      (!CouldHaveKicked(playerIndex)))                                // checken, Spieler kein anderes Komando abgegeben hat
  {
    return;
  };

  // checken, ob auch andere Spieler nah am Ball sind
  for (int i=0;i<playerCount;i++)
  {
    if ((i!=playerIndex) &&
        ((playerInfos[i].lastPos-lastBallPos).length()<playerTestRange) &&
        (CouldHaveKicked(i)))
    {
      return;
    };
  };

  // maxKickRange updaten
  playerInfos[playerIndex].maxKickRange=playerBallDist;
}

void TeamInfo::Calculate()
{
  const Situation& situation = globalWorldModel->situation();

  // cecken, ob sich playercount gendert hat
  if (situation.playerCount()!=playerCount)
  {
    Initialize();
  };

  bool ballDirChange=false;
  Vector currentBallSpeed=situation.ball().speed();
  if ((lastBallSpeed.length()>0.2f) && (currentBallSpeed.length()>0.2f) &&
      ((lastBallSpeed/lastBallSpeed.length()-currentBallSpeed/currentBallSpeed.length()).length()>0.2f))
  {
    ballDirChange=true;
  };

  for (int i=0;i<situation.playerCount();i++)
  {
    // maximalspeed messen
    playerInfos[i].CalcPlayerMaxSpeed();

    // playerdecay messen
    playerInfos[i].CalcPlayerDecay();

    // playerdashValue messen
    playerInfos[i].CalcPlayerDashValue();

    // bei Ball-Richtunsnderung
    if (ballDirChange)
    {
      // Kickrangemessung
      CalcPlayerMaxKickRange(i);
    };
  }

  lastBallSpeed=situation.ball().speed();
  lastBallPos=situation.ball().pos();
  for (int i=0;i<situation.playerCount();i++)
  {
    playerInfos[i].SafeState();
  };
}

void TeamInfo::Initialize()
{
  //cout << "TeamInfo::Initialize()";

  const Situation& situation = globalWorldModel->situation();

  if (playerInfos) {delete [] playerInfos;};
  playerCount=situation.playerCount();
  playerInfos=new PlayerInfo[situation.playerCount()];

  playerTypeCount=7;

  for (int i=0;i<situation.playerCount();i++ ) {
    playerInfos[i].Initialize(i);
  };

  lastBallSpeed=Vector(0,0);
  lastBallPos=Vector();
}

void TeamInfo::CalcPlayerTypes()
{
  // Initialisieren -> alle gleichwahrscheinlich
  for (int i=0;i<playerCount;i++) {
    for (int j=0;j<playerTypeCount;j++) {
      possibleTypes[i][j].playerType=j;
      possibleTypes[i][j].confidence=1.0f;
    };
  };

  // "Messdaten" auswerten

  // maxSpeed
  for (int j=0;j<playerTypeCount;j++) {
    // maxreacheble Speed = playerMaxSpeed*playerDecay*1.1
    float playerTypeMaxSpeed=globalPlayerTypes->playerMaxSpeed(j)*globalPlayerTypes->playerDecay(j)*1.14142136f;
    for (int i=0;i<playerCount;i++) {
      if (playerTypeMaxSpeed<playerInfos[i].maxSpeed) {
        // Wahrscheinlichkeit abwerten
        float f=max(0.1,floor(playerTypeMaxSpeed/playerInfos[i].maxSpeed));
        possibleTypes[i][j].confidence*=f;
      };
    };
  };

  // maxDecay
  for (int j=0;j<playerTypeCount;j++) {
    // maxDecay auslesem
    float playerTypeDecay=globalPlayerTypes->playerDecay(j);
    for (int i=0;i<playerCount;i++) {
        // Wahrscheinlichkeit abwerten
        float f=max(0.1,1-5*fabs(playerTypeDecay-playerInfos[i].avgDecay));
        possibleTypes[i][j].confidence*=f;
    };
  };

  // maxKickRange
  const float ballSize=0.085f;
  for (int j=0;j<playerTypeCount;j++) {
    // kickableRange auslesen
    float playerTypeKickRange=globalPlayerTypes->kickableMargin(j)+globalPlayerTypes->playerSize(j)+ballSize;
    for (int i=0;i<playerCount;i++) {
      if (playerTypeKickRange<playerInfos[i].maxKickRange) {
        // Wahrscheinlichkeit abwerten
        float f=max(0.1,floor(playerTypeKickRange/playerInfos[i].maxKickRange));
        possibleTypes[i][j].confidence*=f;
      };
    };
  };

  // DashPowerRate
  const float maxPower=100.0f;
  for (int j=0;j<playerTypeCount;j++) {
    // kickableRange auslesen
    float playerTypeDashValue=maxPower*globalPlayerTypes->dashPowerRate(j)*(1.14142136*1.14142136)*globalPlayerTypes->playerDecay(j);
    for (int i=0;i<playerCount;i++) {
      if (playerTypeDashValue<playerInfos[i].dashValue) {
        // Wahrscheinlichkeit abwerten
        float f=max(0.1f,playerTypeDashValue/playerInfos[i].dashValue);
        possibleTypes[i][j].confidence*=pow(f,2);
      };
    };
  };

  // Wahrscheinlichkeiten berechnen
  for (int i=0;i<playerCount;i++) {
    float sumConfidence=0.0f;
    for (int j=0;j<playerTypeCount;j++) {
      possibleTypes[i][j].confidence=pow(possibleTypes[i][j].confidence,6);
      sumConfidence+=possibleTypes[i][j].confidence;
    };
    for (int j=0;j<playerTypeCount;j++) {
      possibleTypes[i][j].confidence/=sumConfidence;
    };
  };
};

void TeamInfo::Output()
{
  // Ausgabe
  for (int i=0;i<playerCount;i++) {
    for (int j=0;j<playerTypeCount;j++) {
      cout << floor(possibleTypes[i][j].confidence*100.0f) << "%\t";
    };

    int playerType=globalPlayerTypes->typeOfPlayerWithId(globalWorldModel->situation().playerWithIndex(i).id());
    float fMax=0.0f;
    int   iIdx=-1;
	for (int j=0;j<playerTypeCount;j++) {
      if (((playerType!=j) && (possibleTypes[i][j].confidence>fMax)) ||
          ((playerType==j) && (possibleTypes[i][j].confidence>=fMax))) {
        fMax=possibleTypes[i][j].confidence;
        iIdx=j;
      };
    };
    // bei bereinstimmung check, ob es nur ein maximum gibt
    if (iIdx==playerType) {
      for (int j=0;j<playerTypeCount;j++) {
        if ((fMax==possibleTypes[i][j].confidence) &&
            (j!=iIdx)) {
          iIdx=-1;
          break;
        };
      };
      if (iIdx==-1) {
        cout << " ok";		// mehrere Maxima
      } else {
        cout << " OK";      // nurnoch ein Maximum
      };
    };
    cout << endl;

    for (int j=0;j<playerType;j++) {cout << '\t';};
    cout << "^^^^\n";
  };
}

const TeamInfo::TeamPlayerTypes& TeamInfo::getPlayerTypes()
{
  CalcPlayerTypes();
  return *(TeamPlayerTypes*)(&possibleTypes);
};

int TeamInfo::typeOfPlayerWithIndex( int index ) 
{
  int i = index;
  int bestType = -1;
  double bestConfidence = 0.0;
  double secondConfidence = 0.0;

  for (int j=0;j<playerTypeCount;j++) {
    if( possibleTypes[i][j].confidence > bestConfidence ) {
      secondConfidence = bestConfidence;
      bestConfidence = possibleTypes[i][j].confidence;
      bestType = j;
    }
    else {
      if( possibleTypes[i][j].confidence == bestConfidence ) {
        bestType = -1;
      }
    }
  }
  double limit = 0.025;
  if( bestConfidence < secondConfidence + limit ) {
    bestType = -1;
  }

  return bestType;
}

void TeamInfo::parseSubstituteMessage(const Ath98String& message)
{
  int uid, playerType;
  int rc = sscanf( message.cString(),"(change_player_type %d %d)",
          &uid, &playerType );
  if( rc == 2 ){
    // wir haben ausgewechselt
    INFO( Log::DELIBERATOR, 5, idString(), globalLog->infoStream <<
          "Changed type of teammate player: " << uid << endl );
  }
  else{
    // Gegner hat ausgewechselt
    changedOpponentPlayerTypeByNumber[uid] = YES;
    INFO( Log::SENSORS, 5, idString(), globalLog->infoStream <<
          "Changed type of opponent player: " << uid << endl );
    int playerId = uid + MIN_OPPONENT_ID - 1;
    const Situation& situation = globalWorldModel->situation();
    for( int i = 0; i < situation.playerCount(); i++ ) {
      if( situation.playerWithIndex(i).id() == playerId ) {
        playerInfos[i].Initialize( i );
        break;
      }
    }
  }
}

