
/*
 * Aktueller Stand
 *
 *   $RCSfile: Timer.cc,v $
 *   $Revision: 1.9 $
 *   $Date: 2001/08/02 12:05:49 $
 *   $Author: wendler $ 
 *
 */

#include <sys/time.h>
#include <unistd.h>
#include <stdio.h>

#include "./Timer.h"
#include "./Deliberator.h"
#include "../serverInterface/Sensors.h"
#include "../serverInterface/Effectors.h"
#include "../../basic/Ath98String.h"
#include "../../worldmodel/WorldModelInterface.h"

Timer *globalTimer;

Timer::Timer( Sensors *sensors, Effectors *effectors, 
              Deliberator *deliberator) :
  theSensors( sensors ), theEffectors( effectors ), 
  theDeliberator( deliberator )
{
  setConstructionTime();
  theTimeOfLastCall = getAgentTime();
  theIdString = "Timer";
}

Timer::~Timer() 
{
}

bool Timer::controller() 
{
  int startTime = getAgentTime();
  bool isNewInfo;
  theSensors->sense();
  isNewInfo = theSensors->isNewInfo();

  if( isNewInfo ) {
    globalWorldModel->integrateSensorInfo();

    if( theSensors->isNewSeeInfo() ) {
      INFO( Log::WORLD_MODEL, 5, idString(), globalLog->infoStream <<
            *globalWorldModel );
      theDeliberator->execute();
      if( globalConf->coachUseInteractive ) { readAndSend(); }
    }

    theSensors->initSensors();
    INFO( Log::TIMER, 2, idString(), globalLog->infoStream <<
          "simstep:\t" << globalWorldModel->simStep() <<
          "\tcontroller needed time: " << (getAgentTime() - startTime) << 
          "\n" );
  }
  return isNewInfo;
}

void Timer::agentSleep() 
{
  globalLog->flushInfoStream();

  if( !globalConf->coachWithRecordFile ) {
    usleep(1000);
  }
}

int Timer::getAgentTime() const
{
  static timeval t; 
  gettimeofday( &t, NULL );
  return (t.tv_sec - theTimeOfConstruction) * 1000 + (t.tv_usec / 1000); 
}

void Timer::setConstructionTime()
{
  timeval t; 
  gettimeofday( &t, NULL );
  theTimeOfConstruction = t.tv_sec;
}

void Timer::readAndSend()
{
  static const int sendMax = 8192;
  Ath98String sendMessage = Ath98String("");
  static char line[256];
  cout << "?" << flush;
  
  while( YES ) {
    if( gets( line ) == NULL ) {
      cerr << "Error in reading from stream\n";
    }
    cout << line;
    if( line[0] != '\0' ) {
      sendMessage = sendMessage + Ath98String( line );
      cout << "\n?" << flush;
    }
    else {
      if( sendMessage.length() > 0 ) {
        cout << "Sending:(" << sendMessage.length() << ")\n" << 
          sendMessage << endl;
        if( sendMessage.length() > sendMax ) {
          cout << "Message to long\n";
        }
        theEffectors->sayMessage( sendMessage );
        break;
      }
    }
  }
}

string Timer::idString() const
{
  return theIdString;
}


