#ifndef _SENSORS_H_
#define _SENSORS_H_

#include "./MessageCoach.h"
#include "../basic/TeamInfo.h"
#include "../../basic/Exception.h"
#include "../../worldmodel/Situation.h"


/**
  Diese Klasse ist die Superklasse fuer moegliche Sensor-Klassen
 
  {\small {\bf Quelltext}

  [\URL[Header]{src/Sensors.h} | [\URL[Implementierung]{src/Sensors.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: Sensors.h,v $
  *   $Revision: 1.9 $
  *   $Date: 2001/08/02 11:52:29 $
  *   $Author: wendler $ 
  *
  **************************************************
  \end{verbatim} }*/
class Sensors
{
public:
  /** Interne Fehlerklasse. */
  class SensorsException: public Exception {
  public:
    SensorsException(const Ath98String& aDescription )
      : Exception( aDescription ) {}
  };

protected:
  /** Erzeugt das Sensorobjekt. */
  Sensors();

  virtual ~Sensors();

public:
  /** Initialisieren der Sensordaten.
      Diese Methode muss nach dem kompletten Lesen der aufgelaufenen
      Daten aufgerufen werden um damit kenntlich zu machen, dass man nur noch
      an neueren Daten interessiert ist. */
  void initSensors();

  /** Empfaengt Sensorstrings und verarbeitet sie zu Objekten. */
  virtual void sense();

  /** Wertet die Information nach Betreten des Spielfeldes aus. */
  virtual int checkEnterFieldForSide();

  /** Wertet die Paramter aus, die vom SoccerServer gesendet werden. */
  virtual void checkServerParams();

  virtual bool isNewOkInfo();
  void initOkInfo();

  /** Anfrage ob eine neue Sichtinformation vorliegt. Falls dies der Fall
      ist wird YES zurueckgegeben, andernfalls NO. Diese Methode dient dem
      Weltmodell um zu erkennen, ob sich die Welt geaendert hat. Neue Daten
      liegen vor, wenn nach dem letzten initSensors() Sensordaten ausgewertet
      wurden. */
  bool isNewSeeInfo();

  /** Anfrage ob eine neue Schiedsrichterentscheidungen vorliegen. Falls
      dies der Fall ist wird YES zurueckgegeben, andernfalls NO. Diese
      Methode dient dem Weltmodell um zu erkennen, ob sich die Welt
      geaendert hat. Neue Daten liegen vor, wenn nach dem letzten
      \Ref{initSensors} Schiedsrichterdaten ausgewertet wurden. */
  bool isNewRefereeInfo();

  /** Anfrage ob eine neue H\"orinformation vorliegt. Falls
      dies der Fall ist wird YES zurueckgegeben, andernfalls NO. Diese
      Methode dient dem Weltmodell um zu erkennen, ob sich die Welt
      geaendert hat. Neue Daten liegen vor, wenn nach dem letzten
      \Ref{initSensors} SchiedsrichterdatenH\"ordaten ausgewertet wurden. */ 
  bool isNewHearInfo();

  /** Anfrage ob eine Spielerwechsel-Information vorliegt. */
  bool isNewChangePlayerTypeInfo();

  /** Anfrage, ob ueberhaupt eine Information vorliegt, egal welche. */
  bool isNewInfo();

  /** Rueckgabe der beobachteten Situation. */
  const Situation& situation() const;

  /** Die gehoerten Nachrichten. Diese entstammen der hear-Nachricht des
      SoccerServers. */
  const vector<MessageCoach>& heardMessages() const;

protected:
  /** Der Simulationsschritt, zu dem die Sichtinformationen gehoeren. */
  long theSimStep;

  /** Die Sichtinformation als Situation. */
  Situation theSituation;

  /** Die gehoerten und aufbereiteten Nachrichten. */
  vector<MessageCoach> theHeardMessages;

  bool theIsNewOkInfo;

  /** Liegt ueberhaupt eine neue Information vor, egal welche? */
  bool theIsNewInfo;

  /** Liegt neue Sichtinformation vor? */
  bool theIsNewSeeInfo;
  
  /** Liegen neue Schiederichterentscheidungen vor? */
  bool theIsNewRefereeInfo;
  
  /** Liegt neue H\"orinformation vor. */
  bool theIsNewHearInfo;
  
  /** Liegt Information \"uber einen Spielerwechsel vor. */
  bool theIsNewChangePlayerTypeInfo;

  /** String der die Kurzkennung der Klasse enth\"alt */
  string theIdString;

};

#endif
