
/*
 * Aktueller Stand
 *
 *   $RCSfile: SensorsRecord.cc,v $
 *   $Revision: 1.9 $
 *   $Date: 2001/08/02 11:52:29 $
 *   $Author: wendler $ 
 *
 */
 
#include <arpa/inet.h>
#include "./SensorsRecord.h"
#include "../../basic/Common.h"
#include "../basic/CommonCoach.h"
#include "../../worldmodel/ModelledBall.h"
#include "../../worldmodel/ModelledPlayer.h"
#include "../../worldmodel/WorldModelInterface.h"

SensorsRecord::SensorsRecord( Ath98String recFile )
  : Sensors() 
{
  theIdString = "SensorsRecord";
  prepareRecordFile( recFile );
  playmode = 0; oldPlaymode = 0;
  teamInfo[0].name[0] = '\0';
  teamInfo[1].name[0] = '\0';
  INFO( Log::ANY, 2, idString(), globalLog->infoStream <<
	"SensorsRecord created. \n" );
}

SensorsRecord::~SensorsRecord()
{
  fclose( record );
}

void SensorsRecord::prepareRecordFile( Ath98String recFile )
{
  char localbuf[6];
  // open the file
  if( ( record = fopen( recFile.cString(), "rb" ) ) == NULL ) {
    throwException( SensorsException( "Can't open record file " + recFile ) );
  }

  version = REC_OLD_VERSION;

  // read version information
  if( fread( localbuf, 4, 1, record ) == 0 ) {
    // could not read from file
    fclose( record );
    throwException( SensorsException( "Can't read from record file " + 
                                      recFile ) );
  }

  // check version
  if( (localbuf[0] == 'U') && (localbuf[1] == 'L') && (localbuf[2] == 'G') ) {
    version = localbuf[3];
  }
  else {
    // no version information, so it is an old logfile
    // (or no logfile at all)
    rewind( record );
  }
  cout << "record file version: " << version << endl;
}

void SensorsRecord::sense()
{
  // read until a new showinfo is reached
  while( theIsNewSeeInfo == NO ) {
    if( readDisplayInfo() == NO ) {
      throwException( SensorsException( "End of record file reached." ) );
    }
  }
  theIsNewInfo = YES;
  cout << "Record time:";
  if( version == 1 || version == 2 ) {
    cout << ntohs(info1.body.show.time) << endl;
  }
  else {
    cout << ntohs(info2.body.show.time) << endl;
  }
}



bool SensorsRecord::readDisplayInfo(void)
{
  short len;
  long pos = ftell( record );
	
  INFO( Log::SENSORS, 7, idString(), globalLog->infoStream <<
        "Reading data from pos: " << pos << "\n" );

  switch( version ) {
  case REC_OLD_VERSION:
    if( fread( (char *) (&info1), sizeof(dispinfo_t), 1, record ) == 0 ) {
      return NO;
    }
    switch( ntohs(info1.mode) ) {
    case SHOW_MODE:
      playmode = info1.body.show.pmode;
      teamInfo[0] = info1.body.show.team[0];
      teamInfo[1] = info1.body.show.team[1];
      generateSituationFromInfo();
      return YES;
    case MSG_MODE:
      generateMessageFromInfo();
    default:
      return YES;
    }

  case REC_VERSION_2:
    if( fread( (char *)(&info1.mode), sizeof(info1.mode), 1, record ) == 0 ) {
      return NO;
    }

    switch( ntohs(info1.mode) ) {
    case SHOW_MODE:
      if(fread((char *)(&info1.body.show), sizeof(info1.body.show), 1,
               record) == 0) {
        return NO;
      }
      playmode = info1.body.show.pmode;
      teamInfo[0] = info1.body.show.team[0];
      teamInfo[1] = info1.body.show.team[1];
      generateSituationFromInfo();
      return YES;

    case MSG_MODE:
      if( fread( (char *)(&info1.body.msg.board), 
                 sizeof(info1.body.msg.board), 1, record) == 0 ) {
        return NO;
      }
      if( fread((char *)(&len), sizeof(len), 1, record) == 0 ) {
        return NO;
      }
      len = ntohs(len);		
      if( fread( info1.body.msg.message, len, 1, record ) == 0 ) {
        return NO;
      }
      generateMessageFromInfo();
      return YES;
    }

  case REC_VERSION_3:
    if( fread((char *)(&info2.mode), sizeof(info2.mode), 1, record ) == 0 ) {
      return NO;
    }

    switch( ntohs(info2.mode) ) {
    case PM_MODE:
      if( fread((char *)(&playmode), sizeof(char), 1, record) == 0 ) {
        return NO;
      }
      INFO( Log::SENSORS, 5, idString(), globalLog->infoStream <<
            "Reading new playmode: " << 
            PlayModeString[(int)playmode] << "\n" );
      return YES;
      
    case TEAM_MODE:
      if( fread((char *)(teamInfo), sizeof(team_t), 2, record) == 0) {
        return NO;
      }
      INFO( Log::SENSORS, 5, idString(), globalLog->infoStream <<
            "Reading new team info.\n" );
      return YES;
		  
    case SHOW_MODE:
      info2.body.show.pmode = playmode;
      info2.body.show.team[0] = teamInfo[0];
      info2.body.show.team[1] = teamInfo[1];
      
      if( fread( (char *)(&vision), sizeof(short_showinfo_t2), 1,
                 record ) == 0 ) {
        return NO;
      }
      info2.body.show.ball = vision.ball;
      
      for ( int i = 0; i < MAX_PLAYER * 2; i++ ) {
        info2.body.show.pos[i] = vision.pos[i];
      }  
      info2.body.show.time = vision.time;

      generateSituationFromInfo();
      return YES;

    case MSG_MODE:
      if( fread( (char *)(&info2.body.msg.board), 
                 sizeof(info2.body.msg.board), 1, record) == 0 ) {
        return NO;
      }
      if( fread((char *)(&len), sizeof(len), 1, record ) == 0 ) {
        return NO;
      }
      len = ntohs(len);
      if( fread( info2.body.msg.message, len, 1, record ) == 0 ) {
        return NO;
      }
      generateMessageFromInfo();
      return YES;

    case PARAM_MODE:
      if( fread( (char *)(&info2.body.sparams), sizeof(info2.body.sparams), 1, 
                 record ) == 0) {
        return NO;
      }
      readParameterFromInfo();
      return YES;
      
    case PPARAM_MODE:
      if( fread( (char *)(&info2.body.pparams), sizeof(info2.body.pparams), 1,
                 record ) == 0 ) {
        return NO;
      }
      readParameterFromInfo();
      return YES;

    case PT_MODE:
      if( fread( (char *)(&info2.body.ptinfo), sizeof(info2.body.ptinfo), 1,
                 record ) == 0 ) {
        return NO;
      }
      readParameterFromInfo();
      return YES;

    default:
      INFO( Log::SENSORS, 1, idString(), globalLog->infoStream <<
            "WARNING: Unrecognized MODE: " << ntohs(info2.mode) <<
            " at file position: " << pos << ".\n" );
      return NO;
    }
  default:
    INFO( Log::SENSORS, 1, idString(), globalLog->infoStream <<
          "WARNING: Unrecognized record file version: " << version << ".\n" );
    return NO;
  }
}



int SensorsRecord::checkEnterFieldForSide() 
{
  int side;

  while( teamInfo[0].name[0] == '\0' && teamInfo[1].name[0] == '\0' ) {
    if( readDisplayInfo() == NO ) {
      throwException( SensorsException( "End of record file reached." ) );
    }
  }

  side = sideByTeamNames();

  INFO( Log::SENSORS, 5, idString(), globalLog->infoStream <<
        "Recorded game between teams:\n" <<
        "on left  side:      " << teamInfo[0].name << "\n" <<
        "on right side:      " << teamInfo[1].name << "\n" <<
        "Coaching for team:  " << 
        ((side == RIGHT_SIDE_ID)?teamInfo[1].name:teamInfo[0].name) << "\n" ); 
  return side;
}

int SensorsRecord::sideByTeamNames() {
  if( Ath98String( teamInfo[0].name ) == globalWorldModel->teamName() ) {
    return LEFT_SIDE_ID;
  }  
  else if( Ath98String( teamInfo[1].name ) == globalWorldModel->teamName() ) {
    return RIGHT_SIDE_ID;
  }
  else {
    globalWorldModel->setTeamName( Ath98String( teamInfo[0].name ) );
    return LEFT_SIDE_ID;
  }
}

void SensorsRecord::checkServerParams() 
{
}


void SensorsRecord::generateSituationFromInfo()
{
  INFO( Log::SENSORS, 5, idString(), globalLog->infoStream <<
        "Reading new visual info.\n" );
  
  double x, y, xChange, yChange, bodyDir, faceDir;
  bool alive, isGoalie;
  int simStep, playerId, side, viewWidth, viewQuality;
  short playerType;
  double stamina, effort, recovery;

  // Situation und Ball
  if( version == 3 ) { //using info2 struct
    simStep = ntohs( info2.body.show.time );
    x = toReal2( info2.body.show.ball.x );
    y = -toReal2( info2.body.show.ball.y );
    xChange = toReal2( info2.body.show.ball.deltax );
    yChange = -toReal2( info2.body.show.ball.deltay );

    if( globalWorldModel->side() == RIGHT_SIDE_ID ) {
      x = -x; y = -y;
      xChange = -xChange; yChange = -yChange;
    }   
  }
  if( version == 1 || version == 2 ) {
    simStep = ntohs( info1.body.show.time );
    x = toReal1( info1.body.show.pos[0].x );
    y = -toReal1( info1.body.show.pos[0].y );
    xChange = UNDEF_DOUBLE;
    yChange = UNDEF_DOUBLE;

    if( globalWorldModel->side() == RIGHT_SIDE_ID ) {
      x = -x; y = -y;
    }   
  }
  theSituation = Situation( simStep );
  theSituation.setBall( ModelledBall( BALL_ID, 
        Vector( x, y ), Vector( xChange, yChange ) ) );

  // alle Spieler
  for( int i = 0; i < MAX_PLAYER * 2; i++ ) {
    if( version == 3 ) {
      alive = ( ntohs( info2.body.show.pos[ i ].mode ) & STAND ) != 0;
      INFO( Log::SENSORS, 12, idString(), globalLog->infoStream <<
            "Mode of player with index " << i << " : " << 
            ntohs( info2.body.show.pos[ i ].mode ) << "\n" );
    }
    if( version == 1 || version == 2 ) {
      alive = ( ntohs( info1.body.show.pos[ i+1 ].enable ) & STAND ) != 0;
    }
    
    // ein Spieler
    if( alive ) {
      if( version == 3 ) {
        playerId = i;
        side = ( ( playerId < 11 ) ? LEFT_SIDE_ID : RIGHT_SIDE_ID );
        playerId -= ( ( playerId < 11 ) ? 0 : 11 );
        int number = playerId+1;
        playerId += ( globalWorldModel->side() == side ) ? MIN_TEAMMATE_ID : 
          MIN_OPPONENT_ID;
        playerType = ntohs( info2.body.show.pos[i].type );
        INFO( Log::SENSORS, 15, idString(), globalLog->infoStream << 
              "id:" << playerId << " playerType: " << playerType << "\n" );
        if( globalPlayerTypes->typeOfPlayerWithId( playerId ) != playerType ) {
          INFO( Log::SENSORS, 5, idString(), globalLog->infoStream <<
                "Player type changed for playerId: " << playerId <<
                " to type: " << playerType << "\n" );
          if( globalWorldModel->side() != side ) {
            globalTeamInfo->playerInfos[i].Initialize( i );
            globalTeamInfo->changedOpponentPlayerTypeByNumber[number] = YES;
            INFO( Log::SENSORS, 5, idString(), globalLog->infoStream <<
                  "Changed type of opponent player: " << number << "\n" );
          }
        }
        globalPlayerTypes->setTypeOfPlayerWithId( playerId, playerType );
        x = toReal2( info2.body.show.pos[i].x );
        INFO( Log::SENSORS, 15, idString(), globalLog->infoStream << 
              "x gelesen: " << info2.body.show.pos[i].x << 
              "\nx ntohl:   " << (long)ntohl(info2.body.show.pos[i].x) <<
              "\nx double:  " << (double)(long)(ntohl(info2.body.show.pos[i].x)) <<
              "\nx scaled:  " << ( (double)(long)(ntohl(info2.body.show.pos[i].x)) /
                                 SHOWINFO_SCALE2 ) << "\n" );
        y = -toReal2( info2.body.show.pos[i].y );
        xChange = toReal2( info2.body.show.pos[i].deltax );
        yChange = -toReal2( info2.body.show.pos[i].deltay );
        bodyDir = -toReal2( info2.body.show.pos[i].body_angle );
        faceDir = -toReal2( info2.body.show.pos[i].head_angle ) + bodyDir;
        viewWidth = (int) toReal2( info2.body.show.pos[i].view_width );
        viewQuality = ntohs( info2.body.show.pos[i].view_quality );
        stamina = toReal2( info2.body.show.pos[i].stamina );
        effort = toReal2( info2.body.show.pos[i].effort );
        recovery = toReal2( info2.body.show.pos[i].recovery );
        isGoalie = ( ntohs( info2.body.show.pos[i].mode ) & GOALIE );

        if( globalWorldModel->side() == RIGHT_SIDE_ID ) {
          x = -x; y = -y;
          xChange = -xChange; yChange = -yChange;
          bodyDir = normalizeAngle( bodyDir - Ath98_PI );
          faceDir = normalizeAngle( faceDir - Ath98_PI );
        }        

        viewWidth = VIEW_WIDTH_WIDE;
        viewQuality = VIEW_QUALITY_HIGH;

      }
      if( version == 1 || version == 2 ) {
        playerId = ntohs( info1.body.show.pos[i+1].unum ) - 1;
        side = ( ( ntohs( info1.body.show.pos[i+1].side ) == 1 ) ? 
          LEFT_SIDE_ID : RIGHT_SIDE_ID );
        int ownSide;
        if( globalWorldModel->side() == NO_SIDE_ID ) { 
          globalWorldModel->setSide( sideByTeamNames() );
        }
        ownSide = globalWorldModel->side();
        playerId += ( ownSide == side ) ? MIN_TEAMMATE_ID : MIN_OPPONENT_ID;
        x = toReal1( info1.body.show.pos[i+1].x );
        y = -toReal1( info1.body.show.pos[i+1].y );
        xChange = UNDEF_DOUBLE;
        yChange = UNDEF_DOUBLE;
        bodyDir = radFromDegree( -ntohs( info1.body.show.pos[i+1].angle ) );
        faceDir = bodyDir;
        viewWidth = VIEW_WIDTH_WIDE;
        viewQuality = VIEW_QUALITY_HIGH;
        stamina = globalConf->playerStaminaMax;
        effort = 1.0;
        recovery = 1.0;
        isGoalie = ( ntohs( info1.body.show.pos[i+1].enable ) & GOALIE );

        if( globalWorldModel->side() == RIGHT_SIDE_ID ) {
          x = -x; y = -y;
          bodyDir = normalizeAngle( bodyDir - Ath98_PI );
          faceDir = bodyDir;
        }
      }

      theSituation.addPlayer( ModelledPlayer( playerId, Vector( x, y ),
          Vector( xChange, yChange ), faceDir, bodyDir, 
          viewWidth, viewQuality, stamina, effort, recovery, isGoalie ) );
    }
  }
  theIsNewSeeInfo = YES;

  updatePlaymodesAndScore();
}

void SensorsRecord::updatePlaymodesAndScore() 
{
  // Playmodes
  if( playmode != oldPlaymode ) { //playmode has been changed
    int gameState = 0, initiative = NO_TEAM_ID;
    switch( playmode ) {
    case PM_BeforeKickOff:
      gameState = WorldModelInterface::BEFORE_KICK_OFF;
      initiative = LEFT_SIDE_ID;
      break;
    case PM_TimeOver:
      gameState = WorldModelInterface::TIME_OVER;
      initiative = NO_TEAM_ID;
      break;
    case PM_PlayOn:
      gameState = WorldModelInterface::PLAY_ON;
      initiative = BOTH_TEAMS_ID;
      break;
    case PM_KickOff_Left:
      gameState = WorldModelInterface::KICK_OFF;
      initiative = LEFT_SIDE_ID;
      break;
    case PM_KickOff_Right:
      gameState = WorldModelInterface::KICK_OFF;
      initiative = RIGHT_SIDE_ID;
      break;
    case PM_KickIn_Left:
      gameState = WorldModelInterface::KICK_IN;
      initiative = LEFT_SIDE_ID;
      break;
    case PM_KickIn_Right:
      gameState = WorldModelInterface::KICK_IN;
      initiative = RIGHT_SIDE_ID;
      break;
    case PM_FreeKick_Left:
      gameState = WorldModelInterface::FREE_KICK;
      initiative = LEFT_SIDE_ID;
      break;
    case PM_FreeKick_Right:
      gameState = WorldModelInterface::FREE_KICK;
      initiative = RIGHT_SIDE_ID;
      break;
    case PM_CornerKick_Left:
      gameState = WorldModelInterface::CORNER_KICK;
      initiative = LEFT_SIDE_ID;
      break;
    case PM_CornerKick_Right:
      gameState = WorldModelInterface::CORNER_KICK;
      initiative = RIGHT_SIDE_ID;
      break;
    case PM_GoalKick_Left:
      gameState = WorldModelInterface::GOAL_KICK;
      initiative = LEFT_SIDE_ID;
      break;
    case PM_GoalKick_Right:
      gameState = WorldModelInterface::GOAL_KICK;
      initiative = RIGHT_SIDE_ID;
      break;
    case PM_AfterGoal_Left:
      gameState = WorldModelInterface::GOAL;
      initiative = RIGHT_SIDE_ID;
      break;
    case PM_AfterGoal_Right:
      gameState = WorldModelInterface::GOAL;
      initiative = LEFT_SIDE_ID;
      break;
    case PM_Drop_Ball:
      gameState = WorldModelInterface::DROP_BALL;
      initiative = BOTH_TEAMS_ID;
      break;
    case PM_OffSide_Left:
      gameState = WorldModelInterface::OFFSIDE;
      initiative = RIGHT_SIDE_ID;
      break;
    case PM_OffSide_Right:
      gameState = WorldModelInterface::OFFSIDE;
      initiative = LEFT_SIDE_ID;
      break;
    default:
      INFO( Log::SENSORS, 1, idString(), globalLog->infoStream <<
            "WARNING: Unknown playmode: " << PlayModeString[(int)playmode] <<
            "\n");
      break;
    }
    globalWorldModel->setGameState( gameState );
    if( globalWorldModel->side() == NO_SIDE_ID ) { 
      globalWorldModel->setSide( sideByTeamNames() );
    }
    globalWorldModel->setInitiative( initiative );
    oldPlaymode = playmode;
    theIsNewRefereeInfo = YES;
    INFO( Log::SENSORS, 7, idString(), globalLog->infoStream <<
          "New playmode: " << PlayModeString[(int)playmode] << "\n" );
  }

  // Score
  if( globalWorldModel->score( LEFT_SIDE_ID ) != ntohs(teamInfo[0].score) ) { 
    // left score changed
    globalWorldModel->setScore( ntohs(teamInfo[0].score), LEFT_SIDE_ID );
  }
  if( globalWorldModel->score( RIGHT_SIDE_ID ) != ntohs(teamInfo[1].score) ) { 
    // right score changed
    globalWorldModel->setScore( ntohs(teamInfo[1].score), RIGHT_SIDE_ID );
  }
}

void SensorsRecord::generateMessageFromInfo()
{
  //info2.body.msg.board = info1.body.msg.board;
  //info2.body.msg.message = info1.body.msg.message;
  INFO( Log::SENSORS, 5, idString(), globalLog->infoStream <<
        "Reading a new message: " << info2.body.msg.message << "\n" );

  // to fill: creating a Message with the content

}

void SensorsRecord::readParameterFromInfo()
{
  // in Abhaengigkeit von ntohs(info2.mode)
  // to fill: reading the server parameter from info2 to globalConf
  // to fill: reading the player parameter from info2 to globalConf
  // to fill: reading the player types from info2 to globalPlayerTypes
  if( version == 3 ) {
    Ath98String infoString;
    switch( ntohs(info2.mode) ) {
    case PARAM_MODE:
      INFO( Log::SENSORS, 5, idString(), globalLog->infoStream <<
            "Reading server parameter\n" );
      
      break;
    case PPARAM_MODE:
      INFO( Log::SENSORS, 5, idString(), globalLog->infoStream <<
            "Reading player parameter\n" );
      
      break;
    case PT_MODE:
      INFO( Log::SENSORS, 5, idString(), globalLog->infoStream <<
            "Reading player type info for type " << info2.body.ptinfo.id << 
            ":\n" );
      infoString = Ath98String( "(player_type " ) + 
        ntohs( info2.body.ptinfo.id ) + " " +
        toReal2( info2.body.ptinfo.player_speed_max ) + " " + 
        toReal2( info2.body.ptinfo.stamina_inc_max ) + " " +
        toReal2( info2.body.ptinfo.player_decay ) + " " +
        toReal2( info2.body.ptinfo.inertia_moment ) + " " +
        toReal2( info2.body.ptinfo.dash_power_rate ) + " " +
        toReal2( info2.body.ptinfo.player_size ) + " " +
        toReal2( info2.body.ptinfo.kickable_margin ) + " " +
        toReal2( info2.body.ptinfo.kick_rand ) + " " +
        toReal2( info2.body.ptinfo.extra_stamina ) + " " +
        toReal2( info2.body.ptinfo.effort_max ) + " " +
        toReal2( info2.body.ptinfo.effort_min ) + ")";
      INFO( Log::SENSORS, 5, idString(), globalLog->infoStream << 
            infoString << "\n" );
      globalPlayerTypes->parsePlayerTypeMessage( infoString );
      break;
    default:
      break;
    }
  }
}

bool SensorsRecord::isNewOkInfo() {
  return YES;
}
