#ifndef _SENSORS_RECORD_H_
#define _SENSORS_RECORD_H_

#include <stdio.h>
#include "./Sensors.h"
#include "./ServerTypes.h"
#include "../../basic/Ath98String.h"

/**
  Diese Klasse ist fuer das Auslesen eines Recfiles zustaendig und
  reagiert nach aussen genauso, als wenn es die Daten vom SoccerServer
  lesen wuerde.
 
  {\small {\bf Quelltext}

  [\URL[Header]{src/SensorsRecord.h} | [\URL[Implementierung]{src/SensorsRecord.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: SensorsRecord.h,v $
  *   $Revision: 1.5 $
  *   $Date: 2001/08/02 11:52:29 $
  *   $Author: wendler $ 
  *
  **************************************************
  \end{verbatim} }*/
class SensorsRecord : public Sensors
{
public:

  /** Erzeugt das Sensorobjekt. */
  SensorsRecord( Ath98String recFile );

  ~SensorsRecord();

 /** Empfaengt Sensorstrings und verarbeitet sie zu Objekten.
     Es werden die Daten fuer eine Situation aus dem recFile gelesen. */
  void sense();

  /** Wertet die Information nach Betreten des Spielfeldes aus.
      Im Weltmodell wird die Seite auf der das Team des Coaches spielt
      eingetragen.
      Das Weltmodell muss also vor Aufruf dieser Methode bereits initialisiert
      sein. */
  int checkEnterFieldForSide();

  /** Wertet die Paramter aus, die vom SoccerServer gesendet werden. */
  void checkServerParams();

  bool isNewOkInfo();

private:
  /** Oeffnet das Record-File und bereitet es zum Auslesen vor. */
  void prepareRecordFile( Ath98String recFile );

  /** Liest die naechste Information vom Record-File.
      @return YES wenn eine Information gelesen wurde, sonst NO
  */
  bool readDisplayInfo(void);

  void generateSituationFromInfo();

  void generateMessageFromInfo();

  void readParameterFromInfo();

  int sideByTeamNames();

  void updatePlaymodesAndScore();

private:
  /** Pointer aufs Record-File das abgespielt werden soll. */
  FILE *record;
  /** Version des Record-Files */
  int version;

  /** Puffer fuer einen dispinfo_t Struct */
  dispinfo_t info1;
  /** Puffer fuer einen dispinfo_t2 Struct */
  dispinfo_t2 info2;
  /** Puffer fuer den Playmode unabhaengig ob info1 oder info2 immer richtig
   Info auch in info1 oder info2 enthalten. */
  char playmode, oldPlaymode;
  /** Puffer fuer eine Team-Information, die jeweils aus Namen und
      Score besteht, unabhaengig ob info1 oder info2 immer richtig
      Info auch in info1 oder info2 enthalten. */
  team_t teamInfo[2];

  /** Puffer fuer eine einfachen Situations-Struct. (version 3) */
  short_showinfo_t2 vision;
};

#endif
