#include "GoToIntention.h"

#include "../conditions/PlayerDependentConditions.h"
#include "../conditions/KickConditions.h"
#include "../conditions/PositionConditions.h"
#include "../skills/Skills.h"

#include "../../analyzer/Analyzer.h"
#include "../../worldmodel/Neck.h"

GoToIntention::GoToIntention( const Target& aTarget, double maxDesiredPower,
			                  double staminaLimit, const int specialBehaviors ) :
  	  Intention( theId, aTarget ),
	  theMaxDesiredPower( maxDesiredPower ),
	  theStaminaLimit( staminaLimit ),
	  theSpecialBehaviour( specialBehaviors )
{

//cout << globalWorldModel->simStep() << "  GoToIntention() KONSTR.:: Target = " << aTarget << endl;
  thePreviousTargetPos.setCartesian(0,0);
  addPreCondition( new StaminaCondition( theStaminaLimit ) );
//  addPreCondition( new KickImpossibleCondition() );
  addPostCondition( new PositionCondition( aTarget, 0.1 ) );
  INFO( Log::INTENTIONS , 7 , idString(),	globalLog->infoStream << idString()	<< "GoToIntention::Konstruktor 2\n" ); 
}


Action GoToIntention::execute() {
  INFO( Log::INTENTIONS , 7 , idString(),	globalLog->infoStream << idString()	<< "ENTER GoToIntention::execute\n" );

  Action anAction;
  try {
    INFO( Log::INTENTIONS , 6 , idString(), globalLog->infoStream << idString() << " calculate new Plan\n" );

    theCurrentTargetPos = target().pos();

    const ModelledPlayer& mySelf = globalWorldModel->player();
    Vector mySelfToTarget = theCurrentTargetPos - mySelf.pos();
    Vector mySelfToOldTarget = thePreviousTargetPos - mySelf.pos();
    double deviationBetweenTargets = deviationBetween( mySelfToTarget.angle(), mySelfToOldTarget.angle() );
    double distBetweenTargets = (theCurrentTargetPos - thePreviousTargetPos).length();
    double distToTarget = mySelfToTarget.length();

/*    if( ( deviationBetweenTargets < radFromDegree( 4.5*log(1+distToTarget) ) ) // siehe auch Skills::gotoRegion()
        && ( distBetweenTargets < log(1 + 0.1*distToTarget) ) ) {
      // voriges Ziel beibehalten
      theCurrentTargetPos = thePreviousTargetPos;
    }*/

    // voriges Ziel setzen
    thePreviousTargetPos = theCurrentTargetPos;

    // nicht ueber den Rand des Feldes laufen
    if( theCurrentTargetPos.x() > globalWorldModel->field().enemyGoal().x() ){
      theCurrentTargetPos.setCartesian( globalWorldModel->field().enemyGoal().x(), theCurrentTargetPos.y() );
    }
    if( theCurrentTargetPos.x() < globalWorldModel->field().ourGoal().x() ){
      theCurrentTargetPos.setCartesian( globalWorldModel->field().ourGoal().x(), theCurrentTargetPos.y() );
    }
    if( theCurrentTargetPos.y() > globalWorldModel->field().fieldWidth()/2 ){
      theCurrentTargetPos.setCartesian( theCurrentTargetPos.x(), globalWorldModel->field().fieldWidth()/2 );
    }
    if( theCurrentTargetPos.y() < -globalWorldModel->field().fieldWidth()/2 ){
      theCurrentTargetPos.setCartesian( theCurrentTargetPos.x(), -globalWorldModel->field().fieldWidth()/2 );
    }

    INFO( Log::INTENTIONS , 6 , idString(), globalLog->infoStream << idString() << "execute with targetPos:  " << target().pos() << endl );


    globalSkills->resetSituation( globalWorldModel->situation() );

//  cout << globalWorldModel->simStep() << "  GoToIntention :: Target = " << theCurrentTargetPos << endl;
//  cout << globalWorldModel->simStep() << "  GoToIntention :: myPosition = " << mySelf.pos() << endl;

// HACK !!!  wir mssen fr die GoaliePositionierung die Bewegungsrichtung bestimmen
    if(    ( target().id() == Target::GOALIE_STANDARD_POSITION) 
        && ( globalWorldModel->player().isGoalie() == YES ) ) {
      bool runBackwards = NO;
      Vector targetToPlayer = mySelf.pos() - theCurrentTargetPos;
      Vector playerToTarget = theCurrentTargetPos - mySelf.pos();
      Vector targetToBall = globalWorldModel->ball().pos() - theCurrentTargetPos;
      Vector playerToBall = globalWorldModel->ball().pos() - mySelf.pos();
      Vector targetToPlayerNorm = targetToPlayer;
      targetToPlayerNorm.normalize();
      double angle = deviationBetween( playerToTarget.angle(), playerToBall.angle() );

      // Ball ruht u. ist relativ weit weg
      if( ( (globalWorldModel->gameState() & WorldModelInterface::INACTIVE_BALL) != 0 )
           && (playerToBall.length() > 20.0 )
           && (globalAnalyzer->getNearestEnemyStepsToBall() > 3) ) {
        runBackwards = NO;
      }
      // Ball ist sehr weit weg
      else if( ( playerToBall.length() > 25.0 )
      				&& (globalAnalyzer->getNearestEnemyStepsToBall() > 3) ) {
        runBackwards = NO;
      }
      // Ball bewegt sich von uns weg (nicht zu langsam)
      else if( ( ( targetToPlayerNorm * globalWorldModel->ball().speed() ) > 1.0 ) 
          			&& ( playerToBall.length() > 10 )
        				&& (globalAnalyzer->getNearestEnemyStepsToBall() > 3) ) {
        runBackwards = NO;
      }
      // Ball im Blickwinkel halten
      else if( fabs(angle) > Ath98_PI / 2.0 ) {
        runBackwards = YES;
      }
      else {
        runBackwards = NO;
      }

      if(runBackwards == YES) {
        theSpecialBehaviour = Skills::RUN_BACKWARDS;
      }
    }
//cout << globalWorldModel->simStep() << " GoToIntention::execute()   theCurrentTargetPos = " << theCurrentTargetPos << endl;
    thePlan = globalSkills->gotoRegion( theCurrentTargetPos,   // dorthin wollen wir
                0.1,                                           // wenn wir soweit vom Target entfernt sind, sind wir angekommen
                theStaminaLimit,                               // soviel Kraft mu brigbleiben
                theSpecialBehaviour,                              // Spezialverhalten? (z.B. rckwrts)
                theMaxDesiredPower );                     // max. Dashpower
    if( thePlan.size() > 0 ) {
      anAction = thePlan[0];
    }
    else {
      INFO( Log::INTENTIONS , 6 , idString(), globalLog->infoStream << idString() << "ZERO PLAN\n" );
//cout << globalWorldModel->simStep() << " GoToIntention::planSize = 0 -> send EmtyAction\n";
      anAction = Action( EMPTY_ACTION, globalWorldModel->simStep() );
    }

  }
  catch( ... ) {
    INFO( Log::INTENTIONS,  6,  idString() , globalLog->infoStream << "catching in GoToIntention\n" );
//cerr << "catching in GoToIntention\n";
    anAction = Action( EMPTY_ACTION, globalWorldModel->simStep() );
  }
  INFO( Log::INTENTIONS, 7 , idString() , globalLog->infoStream << " LEAVE " << idString() << " with Action: " << anAction << endl );
  return anAction;
}

