#include "FreeKickIntention.h"

#include "../conditions/AllConditions.h"
#include "GetBallIntention.h"
#include "PassIntention.h"
#include "DribbleIntention.h"
#include "PositionIntention.h"
#include "PPositionIntention.h"
#include "PPositionIntention2.h"
#include "BuildChainIntention.h"
#include "EmergencyIntention.h"

#include "../basicIntentions/KickIntention.h"
#include "../basicIntentions/TurnIntention.h"
#include "../basicIntentions/WaitIntention.h"
#include "../basicIntentions/GoToIntention.h"
#include "../basicIntentions/DummyIntention.h"


#include "../../analyzer/Analyzer.h"

#include "../../basic/Common.h"
#include "../../basic/StraightLine.h"
#include "../../basic/Timer.h"
#include "../../basic/Configuration.h"

#include <iostream>

FreeKickIntention::FreeKickIntention() :
    Intention( theId) {

  theTimeOfLastCheck = globalTimer->getAgentTime();

  Situation situation( globalWorldModel->situation() );
  situation.removeOwnGoalie();
  theInitialDistToBallOfNearestPlayer =
    situation.getBallDistOfXNearestPlayer(1);

  theSimStepsBetweenChecks = max(situation.getNearestTeammateSteps()*2, 10);
  theNumberOfPlayersToGetTheBall = 1;
  theNumberOfRecalculations = 0;
  theTask = NO_TASK;
  initDefaultPreConditions();
  initDefaultPostConditions();  
  initialize();
}

void FreeKickIntention::initialize() {
  INFO( Log::INTENTIONS,  10,  idString() , globalLog->infoStream << 
          idString() << "::initialize\n" ); 
  calculateIntentionOrder();
}


void FreeKickIntention::calculateIntentionOrder() {
  INFO( Log::INTENTIONS,  10,  idString() , globalLog->infoStream << 
	idString() <<	"::calculateIntentionOrder\n");
  // cerr << "calculate Intention Order" << endl;

  // saeubern
  eraseAllPreConditions();
  eraseAllPostConditions();
  eraseAllIntentions(); 

  // Auswahl zw. Freistoss der eigenen oder der gegnerischen
  // Mannschaft
  if( globalWorldModel->initiative()==THEIR_TEAM_ID ){
    theTask = DEFENSE;
  }
  else{
    // Aufgabe innerhalb des FreeKick berechenen
    theTask = calculateTask();
  }
  switch( theTask ){
  case KICK:
    calcKickList();
    break;
  case CHAIN:
    calcChainList();
    break;
  case NORMAL:
    calcNormalList();
    break;
  case DEFENSE:
    calcDefenseList();
    break;
  default:
    ERROR( "No role is calculated for FreeKickIntention!!\n" );
    initDefaultPreConditions();
    initDefaultPostConditions();  
    addIntention( new EmergencyIntention );
  }  
  current = theIntentions.begin();
}

bool FreeKickIntention::needToRecalculateIntentionOrder(){
  bool result = NO;
  
  /* fuer passiven Freistoss muss nicht neu ueberlegt werden, da alle das selbe
  machen */
  if( theTask != DEFENSE ){
  
    /* Testen ob schon genug Zeit vergangen ist  */  
    if( (globalTimer->getAgentTime() - theTimeOfLastCheck) >  
        theSimStepsBetweenChecks * globalConf->simulatorStep * globalConf->slowDownFactor ){
        
      Situation situation( globalWorldModel->situation() );
      situation.removeOwnGoalie();

      /* Testen, ob ich selbst den Schuss ausfuehren soll, aber jemand anders
         dichter am Ball ist */
      if( (theTask == KICK) &&
          (globalWorldModel->id() != situation.index2Id(situation.getIndexOfXNearestPlayer(1))) ){
        theNumberOfPlayersToGetTheBall = 1;
        theSimStepsBetweenChecks = max(situation.getNearestTeammateSteps()*2, 5);
        result = YES;
      }
      /* Testen, ob seit dem letzten mal kein Spieler dichter an den Ball
         rangekommen ist */
      else if( (theTask != KICK) &&
          (situation.getBallDistOfXNearestPlayer(1) > 3) &&
          (situation.getBallDistOfXNearestPlayer(1)>= theInitialDistToBallOfNearestPlayer) ){
        if (theNumberOfPlayersToGetTheBall<10){
          theNumberOfPlayersToGetTheBall++;
        }
        theSimStepsBetweenChecks = max(situation.getNearestTeammateSteps()*2, 5);        
        result = YES;
      }
      /* Wenn der naechste Spieler den Ball in KickRange hat, aendert sich
      nichts mehr */
//      else if( situation.getNearestTeammateSteps()<= 1 ){
//        theSimStepsBetweenChecks = 1000;        
//      }
      theInitialDistToBallOfNearestPlayer = situation.getBallDistOfXNearestPlayer(1);
      theTimeOfLastCheck = globalTimer->getAgentTime();
    }

  }
  if( result == YES ){
    theNumberOfRecalculations++;
  }
  return result;
}

FreeKickIntention::FreeKickTask FreeKickIntention::calculateTask(){
  INFO( Log::INTENTIONS,  6,  idString() , globalLog->infoStream 
   	<< idString() << "::calculateTask\n");

  Situation situation( globalWorldModel->situation() );
  situation.removeOwnGoalie();

  if (theTask != KICK){
    /* der naechste Spieler rennt zum Ball; Sollte keiner hinrennen, wird das
    Kriterium zur Auswahl des KICKERs ggf. gelockert. */
    for(unsigned int i=1; i<=theNumberOfPlayersToGetTheBall; i++) {
      if( globalWorldModel->id() ==
          situation.index2Id(situation.getIndexOfXNearestPlayer(i)) ) {
        return KICK;
      }
    }
  }
  else{
    /* Evtl. rennen zu viele Spieler zum Ball -> ich renne nur hin, wenn ich am
       dichtesten dran bin */
       
    if( globalWorldModel->id() ==
        situation.index2Id(situation.getIndexOfXNearestPlayer(1)) ) {
    
        return KICK;
    }
    else{
      theNumberOfPlayersToGetTheBall=1;
    }
         
  } 
  // pruefen ob Kettenspieler: liefert Target ein gueltiges Ziel, dann
  // ist der Spieler ein Kettenspieler
  Target aTarget( Target::DP_PASS );
//  if( aTarget.pos().isValid() ){
  if( NO ){
    return CHAIN;
  }

  return NORMAL;
}

void FreeKickIntention::calcKickList(){
  INFO( Log::INTENTIONS,  6,  idString() , globalLog->infoStream 
	<< "ENTER " << idString() << "::calcKickList\n" );
  
  Vector target( globalWorldModel->
		 preferredKickDirection( globalWorldModel->ball().pos() ) );
  target.normalize();
  target = 3 * target + globalWorldModel->ball().pos();
  INFO( Log::INTENTIONS,  6,  idString() , globalLog->infoStream 
	<<  "TARGET: " << target << '\n' );
  addIntention( new GoToIntention( Target( target ), 45,
				   globalConf->gotoBaseStaminaLimit,
				   Skills::AVOID_OBSTACLES ) );
  addIntention( new TurnIntention( Target( Target::BALL ) ) );
  addIntention( new GetBallIntention( 45 ) );
  addIntention( new TurnIntention( globalWorldModel->
				   preferredKickDirection( globalWorldModel->
							   ball().pos() ) ) );
  addIntention( new WaitIntention( 50 ) );
  addIntention( new PassIntention );
  
  addIntention( new KickIntention( target ) );

  // Conditions setzen
  initDefaultPreConditions();
  initDefaultPostConditions();
  INFO( Log::INTENTIONS,  6,  idString() , globalLog->infoStream 
	<< "LEAVE " << idString() << "::calcKickList\n" );
}

void FreeKickIntention::calcChainList(){ 
  addIntention( new PositionIntention );
  initDefaultPreConditions();
  initDefaultPostConditions();
}

void FreeKickIntention::calcNormalList(){
// die normalen Spieler sollen sich gleich positionieren und nicht rumstehen wie die Oelgoetzen
//  if( theNumberOfRecalculations == 0 ){
//    addIntention( new WaitIntention( 20 ) );
//  }

   if( globalWorldModel->id() <= 423 ) { // Stuermer
//	  addIntention( new PositionIntention() );
  	  addIntention( new GoToIntention(Target(Target::COVER)) );
  	  addIntention( new TurnIntention( Target( Target::BALL) ) );
   }
   else if( globalWorldModel->id() <= 426 ) { // Mittelfeld
  	  addIntention( new GoToIntention(Target(Target::COVER)) );
  	  addIntention( new TurnIntention( Target( Target::BALL) ) );
//	  addIntention( new PositionIntention() );
   }
   else {
	  addIntention( new PositionIntention() );
//  	  addIntention( new GoToIntention(Target(Target::DP_PASS)) );
//  	  addIntention( new TurnIntention( Target( Target::BALL) ) );
   }



  addIntention( new PositionIntention );

  // Conditions setzen
  initDefaultPreConditions();
  initDefaultPostConditions();
}

void FreeKickIntention::calcDefenseList(){
   eraseAllIntentions();

   if( globalWorldModel->id() <= 423 ) { // Stuermer
//	  addIntention( new PositionIntention() );
  	  addIntention( new GoToIntention(Target(Target::COVER)) );
  	  addIntention( new TurnIntention( Target( Target::BALL) ) );
   }
   else if( globalWorldModel->id() <= 426 ) { // Mittelfeld
  	  addIntention( new PositionIntention() );
   }
   else {
  	  addIntention( new GoToIntention(Target(Target::DP_PASS)) );
  	  addIntention( new TurnIntention( Target( Target::BALL) ) );
   }


  // Conditions setzen
  initDefaultPreConditions();
  initDefaultPostConditions();
}

void FreeKickIntention::initDefaultPreConditions(){
  addPreCondition( new GameStateCondition( WorldModelInterface::FREE_KICK |
					   WorldModelInterface::KICK_IN   |
					   WorldModelInterface::GOAL_KICK |
					   WorldModelInterface::CORNER_KICK |
					   WorldModelInterface::OFFSIDE ) );
}

void FreeKickIntention::initDefaultPostConditions(){
  addPostCondition( new GameStateCondition( WorldModelInterface::PLAY_ON ) );
}


