#ifndef _FREEKICK_INTENTION_H_
#define _FREEKICK_INTENTION_H_

#include "../Intention.h"

/** Ich sehe im Moment nicht, das es einen funktionalen Unterschied zwischen
Intention und ComplexIntention gibt. Das sollte nach Seattle geandert werden. */

class  FreeKickIntention : public Intention {
 public:
  /** Aufz"ahlungstyp f"ur die Aufgaben, die die Spieler beim Freistoss
  uebernehmen. */
  enum FreeKickTask { NO_TASK, KICK, CHAIN, NORMAL, DEFENSE };

  static const IntentionId theId = 13;
  FreeKickIntention();
  
  int utility(){ return 0; }

  void calculateIntentionOrder();
  
  /** Prueft, ob eine Neuverteilung der Aufgaben beim aktiven Freistoss notwendig ist.
      Damit soll sichergestellt werden, dass in mehrdeutigen Situationen --
      mehrere Spieler gleich weit vom Ball entfernt -- mindestens einer zum Ball
      laeuft bzw. dass nicht zuviele Spieler hinlaufen, nachdem bei Eintreten
      des ersten Falls die Auswahlbedingungen fuer den KICK-Task gelockert
      wurden */
  bool needToRecalculateIntentionOrder();

 private:
  /** Die Aufgabe, die der Spieler beim aktuellen Freistoss einnimmt. */
  FreeKickTask theTask;
  

  /** Zeitpunkt an dem das letze mal die Notwendigkeit einer Neuberechnung
  geprueft wurde */
  int theTimeOfLastCheck;
  /** Anzahl der SimSteps zwischen 2 Recalculateaufrufen */  
  double theSimStepsBetweenChecks;
  /** Urspruengliche Entfernung des naechsten Spielers zum Ball */
  double theInitialDistToBallOfNearestPlayer;
  /** Anzahl der Spieler die zum Ball rennen */
  unsigned int theNumberOfPlayersToGetTheBall;
  /** Anzahl der durchgefuehrten Neuordnungen der IntentionListe */
  unsigned int theNumberOfRecalculations;
  
  /** Berechung der Aufgabe, die der Spieler beim aktuellen Freistoss
      uebernimmt.
  @return Die Aufgabe. */
  FreeKickTask calculateTask();
  
  /** Berechnen der aufgabenabh"angigen Intentionslisten incl.
     der zugeh"origen Condition-Listen. */
  /** Intentionsliste f"ur den Spieler, der den Freistoss ausf"uhrt. */
  void calcKickList();
  /** Berechnen der Intentionsliste f"ur einen Spieler, der sich als
      Anspielpartner aufstellt. */
  void calcChainList();
  /** Berechnen der Intentionsliste f"ur einen Spieler, der keine
      aktive Rolle beim Freistoss einnimmt. */
  void calcNormalList();
  /** Berechnen der Intentionsliste f"ur den gegnerischen Freistoss. */
  void calcDefenseList();

  /** Initialisierung des Freekick. */
  void initialize();
  void initDefaultPreConditions();
  void initDefaultPostConditions();

  //@}
};

#endif
