#include "StandardRole.h"
#include "Role.h"

#include "../../analyzer/Analyzer.h"

#include "../../basic/Timer.h"

#include "../conditions/WorldModelStatesConditions.h"

#include "../basicIntentions/InitialScanIntention.h"
#include "../basicIntentions/MoveIntention.h"
#include "../basicIntentions/WaitIntention.h"
#include "../basicIntentions/DummyIntention.h"
#include "../basicIntentions/TurnIntention.h"
#include "../basicIntentions/KickIntention.h"
#include "../basicIntentions/PMoveIntention.h"
#include "../basicIntentions/GoToIntention.h"

#include "../complexIntentions/GetBallIntention.h"
#include "../complexIntentions/InterceptBallIntention.h"
#include "../complexIntentions/FreeKickIntention.h"
#include "../complexIntentions/ParallelWalkIntention.h"
#include "../complexIntentions/DribbleIntention.h"
#include "../complexIntentions/TrivialPassIntention.h"
#include "../complexIntentions/PassIntention.h"
#include "../complexIntentions/PositionIntention.h"
#include "../complexIntentions/ScoreGoalIntention.h"
#include "../complexIntentions/ScanFieldIntention.h"
#include "../complexIntentions/PPositionIntention.h"
#include "../complexIntentions/PPositionIntention2.h"
#include "../complexIntentions/EmergencyIntention.h"

#include "../../worldmodel/WorldModelInterface.h"


#include "../skills/Skills.h"

StandardRole::StandardRole() : Role(100, "StandardRole")
{
  addIntention( new DummyIntention() );
  current = theIntentions.begin();

  setHomePosition( globalWorldModel->basePosition(WorldModelInterface::PLAY_ON, OUR_TEAM_ID) );
  setBeamPositionOurKickOff( globalWorldModel->initialPosition(WorldModelInterface::KICK_OFF, OUR_TEAM_ID) );
  setBeamPositionEnemyKickOff( globalWorldModel->initialPosition(WorldModelInterface::KICK_OFF, THEIR_TEAM_ID) );
  
  needNewIntentionList = YES;
  theGameState = -1; 
}


StandardRole::~StandardRole()
{
}


void StandardRole::deliberate()
{
  INFO( Log::ROLE,  6,  name() , globalLog->infoStream << "ENTER " << name() << "::deliberate\n");
  globalAnalyzer->analyze();

  if( ( theGameState != globalWorldModel->gameState() ) 
  	   || (globalWorldModel->simStepOfLastRefereeDecision() == globalWorldModel->simStep() ) ){
    INFO( Log::ROLE,  6,  name() , globalLog->infoStream << name() << ": Der Gamestate hat sich geaendert\n" );
    needNewIntentionList = YES;
    sitClassChange = YES;
    theGameState = globalWorldModel->gameState();
  }
  else {
    sitClassChange = NO;
  }
//cout << "SitClassChanged : " << (sitClassChange? "YES" : "NO")  << endl;

  setNeckBehaviour();


  if( globalWorldModel->ball().posConfidence().reliability() < 0.5 ){ // HACK! muss mit dem HACK im Neck konform sein!
    // wenn der Ball zu lange nicht mehr gesehen wurde, mu Scanfield ihn suchen
    if( lastIntentionId() != ScanFieldIntention::theId ){
      INFO( Log::ROLE,  6,  name() , globalLog->infoStream << name() <<
      ": ScanfieldIntention ausloesen\n" );
      eraseAllIntentions();
      addIntention( new ScanFieldIntention() );
      current = theIntentions.begin();
    }
    needNewIntentionList = NO;
  }
  
  /** wenn beim Freistoss niemand zum Ball rennt (jeder denkt jemand anders ist
  schneller dran), muessen die Aufgaben nochmal neu verteilt werden */

  if( current!= theIntentions.end() ){
    if( (* current)->id() == FreeKickIntention::theId ){
       if( ((FreeKickIntention *)(* current))->needToRecalculateIntentionOrder() == YES ){
         ((FreeKickIntention *)(* current))->calculateIntentionOrder();
      }
    }
  }

 
  /** wenn die alte Intentionliste abgearbeitet ist, oder Grnde fr eine Neuberechnung sprechen
      wird eine neue Intentionliste angelegt */

  if( needNewIntentionList == YES ){
    INFO( Log::ROLE,  6,  name() , globalLog->infoStream << name() << ": Neue Intentionliste generieren\n" );
    // gamestate = before_kick_off oder kick_off => Anstoss
	if( (globalWorldModel->gameState() & (WorldModelInterface::KICK_OFF | WorldModelInterface::BEFORE_KICK_OFF | WorldModelInterface::GOAL) ) != 0 ){
      INFO( Log::ROLE,  6,  name() , globalLog->infoStream << name() << ": Situation vor dem Anstoss\n" );
      eraseAllIntentions();
      Vector beamPosition = (globalWorldModel->initiative() == OUR_TEAM_ID) ? beamPositionOurKickOff() : beamPositionEnemyKickOff();
      addIntention( new MoveIntention( beamPosition ) );
      addIntention( new TurnIntention( Vector( 0, 0 ) ) );
      if( globalWorldModel->player().distTo( globalWorldModel->ball().pos() ) < globalWorldModel->player().kickRange()
          && (globalWorldModel->gameState() & WorldModelInterface::KICK_OFF) != 0 ){
//cerr << "ANSTOSS !!!!\n";          	
          // Anstoss und ich habe den Ball in meiner Kickrange -> passen zum 3. angemeldeten Spieler
   	      addIntention( new PassIntention() );
          addIntention( new KickIntention( Target(Target::PLAYER, 422), 0.5, 2 ) );
          addIntention( new KickIntention( Target(Vector(-3.0, 8.0)), 0.8, 2 ) );
      }
      else{
      	addIntention( new InitialScanIntention() );
        addIntention( new TurnIntention( Vector( 0, 0 ) ) );
        addIntention( new WaitIntention( new GameStateCondition( WorldModelInterface::PLAY_ON ) ) );
      }
    } 
    else if( ( ( ( theGameState & ( WorldModelInterface::FREE_KICK | WorldModelInterface::CATCH_BALL_AND_FREE_KICK ) ) != 0 )
               && ( globalWorldModel->initiative() == THEIR_TEAM_ID ) )
             ||( ( ( theGameState & ( WorldModelInterface::GOAL_KICK ) ) != 0 )
               && ( globalWorldModel->initiative() == OUR_TEAM_ID ) ) ) {
      eraseAllIntentions();
      if( globalWorldModel->id() <= 423 ) { // Stuermer
    	  addIntention( new PositionIntention() );
      }
      else if( globalWorldModel->id() <= 426 ) { // Mittelfeld
    	  addIntention( new PositionIntention() );
      }
      else {
    	  addIntention( new GoToIntention(Target(Target::DP_PASS)) );
    	  addIntention( new TurnIntention( Target( Target::BALL) ) );
      }
    }
    else if( ( theGameState & ( WorldModelInterface::FREE_KICK | WorldModelInterface::CORNER_KICK |
    							WorldModelInterface::KICK_IN | /*WorldModelInterface::GOAL_KICK |*/
    							WorldModelInterface::OFFSIDE | WorldModelInterface::FOUL ) ) != 0 ) {
      eraseAllIntentions();
      addIntention( new FreeKickIntention() );
      //addIntention( new InterceptBallIntention() );
      //addIntention( new PassIntention() ); 
      //addIntention( new KickIntention( globalWorldModel->player().pos(), 0.01, 2 ) ); // HACK
      //addIntention( new PPositionIntention() );
    }
    else{
      INFO( Log::ROLE,  6,  name() , globalLog->infoStream << name() << ": Normale Spielsituation\n" );
      eraseAllIntentions();

      if( globalWorldModel->id() <= 423 ) { // Stuermer
        addIntention( new InterceptBallIntention() );
        addIntention( new ScoreGoalIntention() );
        addIntention( new PassIntention() );
        addIntention( new ScoreGoalIntention( YES ) ); // risky scoreGoal
        addIntention( new DribbleIntention() );
        addIntention( new KickIntention( Target(Target::DRIBBLE), 0.6) ); //HACK
  	  	addIntention( new PPositionIntention() );
	    	addIntention( new PPositionIntention2() );
      }
      else if( globalWorldModel->id() <= 426 ) { // Mittelfeld
        addIntention( new InterceptBallIntention() );
        addIntention( new ScoreGoalIntention() );
        addIntention( new PassIntention() );
        addIntention( new ScoreGoalIntention( YES ) ); // risky scoreGoal
//        addIntention( new DribbleIntention() );
        addIntention( new KickIntention( Target(Target::DRIBBLE), 1.0) ); //HACK
  	  	addIntention( new PPositionIntention() );
	    	addIntention( new PPositionIntention2() );
      }
      else {                                      //Verteidiger
        addIntention( new InterceptBallIntention() );
        addIntention( new PassIntention() );
        addIntention( new KickIntention( Target(Target::DRIBBLE), 2.7 ) ); //HACK
  	  	addIntention( new PPositionIntention() );
	    	addIntention( new PPositionIntention2() );
      }

    }

    // Iterator auf aktuell auszufhrende Intention an den Anfang setzen
    current = theIntentions.begin();
    needNewIntentionList = NO;
//    sitClassChange = YES; 
    // HACK
    setLastIntentionId( DummyIntention::theId /*(*current)->id()*/ );
  }

  INFO( Log::ROLE,  6,  name() , globalLog->infoStream	<< "LEAVE " << name() <<"::deliberate()\n" );
}// end of deliberate()

