
#include "Communicator.h" 
#include "../basic/Log.h"
#include "../worldmodel/WorldModelInterface.h"
	 
Communicator::Communicator(const Ath98String& hostName, int port) { 
  ownSocket = -1;  
  communicationAvailable = NO;  
  openConnection(hostName, port); 
  theIdString = "Communicator"; 
  theSendedActions.reserve( 10 ); 
} 
	 
Communicator::~Communicator() { 
  closeConnection();  
} 
 
void Communicator::sendMessage(const Ath98String& theMessage ) { 

  INFO( Log::COMMUNICATOR,  4, idString(), globalLog->infoStream << 
	" Send: " <<    theMessage  <<  " \n"  ) ; 
  
 
  if ( sendto(ownSocket, theMessage.cString(), theMessage.length(), 
	      0, (struct sockaddr *) &hostAddress,  
	      sizeof(hostAddress)) != theMessage.length() ) { 
    throwException( CommunicatorException( "Send failed") );  
  }
  STAMP ( theSendedActions.push_back( string( theMessage.cString() ) ); );
} 
		 
Ath98String Communicator::getNextMessage(int seconds) { 
  fd_set fileDescriptor;  
  struct timeval timeInterval;  
  struct timeval* pointerToTimeInterval;  
  static char buffer[MAX_MESSAGE];  
  static int lengthOfMessage;  
 
  FD_ZERO(&fileDescriptor);  
  FD_SET(ownSocket, &fileDescriptor);  
  if ( seconds == WAIT_FOREVER) { 
    pointerToTimeInterval = NULL;  
  } 
  else { 
    pointerToTimeInterval = &timeInterval;  
    timeInterval.tv_sec = seconds;  
    timeInterval.tv_usec = 0;  
  } 
 
  if ( select( ownSocket+1, &fileDescriptor, NULL, NULL,  
	      pointerToTimeInterval ) > 0 ) { 
    lengthOfMessage = recvfrom(ownSocket, buffer, MAX_MESSAGE, 0, 
			       (struct sockaddr *) &hostAddress, 
			       &lengthOfHostAddress);  
    buffer[lengthOfMessage++] = '\0';  
     
    INFO( Log::COMMUNICATOR  ,  4 ,  idString() , globalLog->infoStream << 
	  " Receive at Simstep: " << globalWorldModel->simStep() << "\n"  <<  buffer ) ; 
  
     
    if (lengthOfMessage == 0) { 
      throwException( NothingReceived() );
      // return Ath98String();  // hack micha um den Segmentationfault am anfang zu beheben
    } 
    else { 
      return Ath98String(buffer);  
    } 
  } 
  else { 
    throwException( NothingReceived() );  
  } 
} 
 
Ath98String Communicator::getLastMessage() { 
  Ath98String message;  
 
  message = getNextMessage(WAIT_FOREVER);  
  try { 
    while(1) { 
      message = getNextMessage(0);  
    } 
  } 
  catch (NothingReceived) {} 
  return message;  
} 
 
void Communicator::openConnection(const Ath98String& hostName, int port) { 
  struct hostent* hostEnvironment;  
  struct in_addr* hostInternetAddress;  
  struct sockaddr_in clientAddress;  
  int socketID;  
  unsigned int bufferSize = 64000;  
  const char* host = hostName.cString();  
   
  // find the host 
  if (( hostEnvironment = (struct hostent*) gethostbyname(host)) == NULL) { 
    if ( ((long) inet_addr(host)) == -1) { 
      throwException(  
          CommunicatorException( "Can't find address of the host") );  
    } 
  } 
  else { 
    hostInternetAddress = (struct in_addr*) *hostEnvironment->h_addr_list;  
    host = inet_ntoa (*hostInternetAddress);  
  } 
   
  // get the socket 
  if (( socketID = socket(AF_INET, SOCK_DGRAM, 0)) < 0) { 
    throwException(  
        CommunicatorException( "Initialization of socket failed") );  
  } 
  if (( setsockopt (socketID, SOL_SOCKET, SO_RCVBUF,  (char*) &bufferSize, 
		    sizeof (int)) ) < 0) { 
    throwException( 
        CommunicatorException( "Initialization of socket failed") );  
  } 
 
  // bind the socket 
  bzero( (char *) &clientAddress, sizeof(clientAddress) );  
  clientAddress.sin_family = AF_INET;  
  clientAddress.sin_addr.s_addr = htonl(INADDR_ANY);  
  clientAddress.sin_port = htons(0);  
  if ( bind(socketID, (struct sockaddr *) &clientAddress,  
	    sizeof(clientAddress)) < 0) { 
    throwException( CommunicatorException( "Binding of socket failed") );  
  } 
 
  ownSocket = socketID;  
 
  bzero( (char *) &hostAddress, sizeof(hostAddress) );  
  hostAddress.sin_family = AF_INET;  
  hostAddress.sin_addr.s_addr = inet_addr(host);  
  hostAddress.sin_port = htons(port);  
   
  lengthOfHostAddress = sizeof(hostAddress);  
  communicationAvailable = YES;  
} 
void Communicator::closeConnection() { 
  if (communicationAvailable) { 
    close(ownSocket);  
    communicationAvailable = NO;  
  } 
} 
