#include "Message.h"

#include <strings.h> 
 
#include "../basic/Exception.h"
#include "../basic/Configuration.h"

char Message::theTeamnameCheckSum[4] = "\0\0\0";
Ath98String Message::theTeamname = Ath98String( "" );
 
Message::Message( const Ath98String& information )
{ 
  theTime = UNDEF_TIME;
  setByHearInfo( information );
} 

Message::Message( const Ath98String& information, long simStep )
{
  theTime = UNDEF_TIME;
  setBySayInfo( information, simStep );
}
 
Message::Message()  
{ 
  theTime = UNDEF_TIME;  
}

Message::~Message()
{
}

void Message::setTeamname( const Ath98String& teamname ) {
  theTeamname = teamname;
  char name[teamname.length() + 3];
  char* namePointer = name;
  memcpy( namePointer, teamname.cString(), teamname.length() );
  bzero( namePointer + teamname.length(), 3 );

  for( int i = 0; i < teamname.length(); i+=3 ) {
    theTeamnameCheckSum[0] += namePointer[0];
    theTeamnameCheckSum[1] += namePointer[1];
    theTeamnameCheckSum[2] += namePointer[2];
    namePointer += 3;
  }
}

void Message::setByHearInfo( const Ath98String& information )
{
  // parse hear-message
  parseHearMessage( information );
  if( theSenderInfo == UNDEF_TEAM_ID) {
    // decode hear-message
    // HACK
    // decodeHearMessage( theMessage );  
  }
}

void Message::setBySayInfo( const Ath98String& information, long simStep )
{
  theMessage = information;
  theTime = simStep;
  theSenderInfo = SELF_ID;
}


void Message::parseHearMessage( const Ath98String& information )  
{ 
  if( !(information.beginsWith( "(hear") ) ) { 
    throwException( ( Exception("Message was incorrect used") ) );  
  } 
  char message[MAX_MESSAGE];  
  char sender[30];  
  sscanf( information.cString(), "(hear %ld %s %[^)]", &theTime, 
	  sender, message );  
  if ( strcmp(sender, "referee") == 0 ) { 
    theSenderInfo = REFEREE_ID;  
  } 
  else if( strcmp(sender, "self") == 0 ) { 
    theSenderInfo = SELF_ID;  
  } 
  else { 
    theSenderInfo = UNDEF_TEAM_ID;  
  } 
  theMessage = Ath98String(message); 
}


Ath98String Message::encodeSayMessage( const Ath98String& info, 
                                       long simStep ) const
{
  INFO( Log::WORLD_MODEL, 7, "", globalLog->infoStream <<
        "Encoding sayMessage for string: " << info << "\n" );
  int headerLengthIn = 6;
  int lengthIn = info.length();
  if( lengthIn % 3 != 0 ) { lengthIn += 3 - lengthIn % 3;}
  int lengthOut = ((lengthIn + headerLengthIn) / 3) * 4;

  if( lengthOut > globalConf->sayMessageSize ) {
    INFO( Log::WORLD_MODEL, 0, "", globalLog->infoStream <<
          "ERROR: encodeSayMessage: message is bigger than allowed, " <<
          "truncating.\n" );
    lengthOut = (globalConf->sayMessageSize / 4) * 4;
    lengthIn = (lengthOut / 4) * 3 - headerLengthIn;
  }
  INFO( Log::WORLD_MODEL, 9, "", globalLog->infoStream <<
        "message: in length: " << lengthIn << "+" << headerLengthIn << 
        " out length: " << lengthOut << "\n" );

  unsigned char inArray[lengthIn + headerLengthIn + 1];
  unsigned char* inPointer = inArray;
  bzero(inPointer, (lengthIn + headerLengthIn + 1)* sizeof(unsigned char));
  inPointer += 3; //keeping place for checksum
  addSimStep( inPointer, simStep ); //adding simStep
  memcpy( inPointer + 3, info.cString(), lengthIn );
  memcpy( inPointer + 3 + info.length(), "   ", lengthIn - info.length() );
  addCheckSum( inPointer, lengthIn + 3, inArray );
  inPointer = inArray;
  INFO( Log::WORLD_MODEL, 9, "", globalLog->infoStream <<
        "Message before encoding: " << inArray + headerLengthIn << "\n" );

  unsigned char outArray[lengthOut + 1];
  unsigned char* outPointer = outArray;
  bzero(outPointer, (lengthOut + 1) * sizeof(unsigned char));
  for( int i = 0; i < lengthOut; i += 4 ) {
    encode( inPointer, outPointer );
    inPointer += 3; outPointer += 4;
  }
  INFO( Log::WORLD_MODEL, 7, "", globalLog->infoStream <<
        "Generated encodedSayMessage: " << outArray << "\n" );

  return Ath98String( (char *)outArray );
}

void Message::decodeHearMessage( const Ath98String& info )
{
  INFO( Log::WORLD_MODEL, 7, "", globalLog->infoStream <<
        "Decoding hearMessage for string: " << info << "\n" );
  int lengthIn = info.length();
  if( lengthIn % 4 != 0 || lengthIn < 8 ) { 
    INFO( Log::WORLD_MODEL, 9, "", globalLog->infoStream <<
          "Message has not correct length -> it is from opponent." );
    theSenderInfo = THEIR_TEAM_ID;
    theMessage = Ath98String();
    return;
  }
  int headerLengthOut = 6;
  int lengthOut = (lengthIn / 4) * 3;
  INFO( Log::WORLD_MODEL, 9, "", globalLog->infoStream <<
        "message: in length: " << lengthIn << 
        " out length: " << lengthOut << "-" << headerLengthOut << "\n" );

  unsigned char inArray[lengthIn + 1];
  unsigned char* inPointer = inArray;
  bzero(inPointer, (lengthIn + 1)* sizeof(unsigned char));
  memcpy( inPointer, info.cString(), lengthIn );
  INFO( Log::WORLD_MODEL, 9, "", globalLog->infoStream <<
        "Message before decoding: " << inArray << "\n" );

  unsigned char outArray[lengthOut + 1];
  unsigned char* outPointer = outArray;
  bzero(outPointer, (lengthOut + 1) * sizeof(unsigned char));
  for( int i = 0; i < lengthOut; i +=3 ) {
    decode( inPointer, outPointer );
    inPointer += 4; outPointer += 3;
  }
  outPointer = outArray;
  INFO( Log::WORLD_MODEL, 7, "", globalLog->infoStream <<
        "Generated decodedSayMessage: " << outArray + headerLengthOut << 
        "\n" );

  unsigned char checkSum[3];
  addCheckSum( outPointer + 3 , lengthOut - 3, checkSum );
  if( memcmp( outPointer, checkSum, 3 ) != 0 ) {
    INFO( Log::WORLD_MODEL, 9, "", globalLog->infoStream <<
          "Message has not correct checksum -> it is from opponent." );
    theSenderInfo = THEIR_TEAM_ID;
    theMessage = Ath98String();
    return;
  }
  //checksum correct
  theSenderInfo = OUR_TEAM_ID;
  extractSimStep( outPointer + 3 );
  theMessage = Ath98String( (char *)(outPointer + headerLengthOut) );
  INFO( Log::WORLD_MODEL, 6, "", globalLog->infoStream <<
        "Message decoded: " << *this );
  return;
}

void Message::encode( const unsigned char* in, unsigned char*& out ) const
{
  out[0] = in[0]>>2;
  out[1] = (in[1] >> 4) + ((in[0] & 0x03) << 4);
  out[2] = (in[2] >> 6) + ((in[1] & 0x0f) << 2);
  out[3] = in[2] & 0x3f;
  for( int i = 0; i < 4; i++ ) {
    out[i] = globalConf->encodetable[out[i]];
  }
} 
  
void Message::decode( const unsigned char* inP, unsigned char*& out ) const
{
  unsigned char in[4];  
  for( int i = 0; i < 4; i++ ) {
    in[i] = globalConf->decodetable[inP[i]];
  } 
  out[0] = (in[0] << 2) + (in[1] >> 4);
  out[1] = ((in[1] & 0x0f) << 4) + (in[2] >> 2);
  out[2] = ((in[2] & 0x03) << 6) +  in[3];
} 

void Message::addCheckSum( const unsigned char* in, const int lengthOfIn, 
                           unsigned char* out ) const
{
  out[0] = out[1] = out[2] = 0;
  const unsigned char* inPointer = in;
  for( int i = 0; i < lengthOfIn; i+=3 ) {
    out[0] += inPointer[0];
    out[1] += inPointer[1];
    out[2] += inPointer[2];
    inPointer += 3;
  }
  out[0] += theTeamnameCheckSum[0];
  out[1] += theTeamnameCheckSum[1];
  out[2] += theTeamnameCheckSum[2];
}  

void Message::addSimStep( unsigned char* out, long simStep ) const
{
  out[0] = (unsigned char)simStep;
  out[1] = (unsigned char)(simStep>>8);
}

void Message::extractSimStep( const unsigned char* in )
{
  theTime = (long)((((unsigned short)in[1])<<8) + (unsigned short)in[0]);
}


ostream& operator<< (ostream& os, const Message& message ) {
  os << 
    "( message sender: " << message.senderInfo() <<
    ", time: " << message.time() <<
    ", message: " << message.message() << " )\n";
  return os;
}

long Message::time() const {
  return theTime;
}

const Ath98String& Message::message() const {
  return theMessage;
}

int Message::senderInfo() const {
  return theSenderInfo;
}

Ath98String Message::messageForSend() const {
  return encodeSayMessage( theMessage, theTime );
}
;  
