#ifndef _CONFIDENCE_H_
#define _CONFIDENCE_H_

#include <stream.h>

#include "../basic/Ath98String.h"
#include "../basic/Common.h"


/**
  Klasse, die die Verlaesslichkeit von Positionen und Geschwindigkeiten
  erfasst. Sie wird als Member von MovingObject vererbt. Um den Fehlerradius
  eines Objekts zu berechnen, ist die Entfernung in der es gesehen wurde ntig.
  Diese Klasse speichert alle relativen Koordinaten, die vom Soccerserver fr 
  ein Objekt bermittelt wurden.
 
  {\small {\bf Quelltext}

  [\URL[Header]{src/Confidence.h} | [\URL[Implementierung]{src/Confidence.cc}]

  {\bf CVS Informationen zum Interface}

  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: Confidence.h,v $
  *   $Revision: 1.1.1.1 $
  *   $Date: 2001/07/31 22:23:09 $
  *   $Author: rami $ 
  *
  **************************************************
  \end{verbatim} }*/

class Confidence
{
public:	
  
  /// Vergleichskonstanten fr das Alter des Objektes
  static const unsigned long UNDEF_AGE = UNDEF_LONG/2;


  /** Konstruktor fuer ein neues Confidence-Objekt
      @param   */
  Confidence::Confidence();  
  
  Confidence::Confidence(const Confidence&  ); 
  
  /** Konstruktor fuer ein neues Confidence-Objekt
      @param   */
  Confidence( int simStep, double aMaxError, double aReliability );

  void setSimStep( int aSimStep );

  /** Verringert die Confidence --> ist noch ein Hack */
  void decrease( double factor = 0.9 );

  /** liefert das Alter des Objektes, also die Anzahl
      der simsteps, die seit seiner beobachtung vergangen sind
      @return  */
  int age() const;

  /** */
  int simStep() const;  

  /** */
  double maxError( ) const;

  /** */
  double reliability( ) const;

  /** Ausgabe auf einen String-Stream */
  friend ostream& operator<< (ostream&, const Confidence& );

  Ath98String toString() const;

 private:
  long theSimStep; 
  double theMaxError;
  double theReliability;
};

inline void Confidence::decrease( double factor ){
  theReliability *= factor;
}

inline int Confidence::simStep() const {
  return theSimStep;
}

inline void Confidence::setSimStep( int aSimStep ){
  theSimStep = aSimStep;
}

inline double Confidence::reliability( ) const {
  return theReliability;
}

inline double Confidence::maxError( ) const {
  return theMaxError;
}

#endif





