#include <math.h>

#include "ConfiguredPlayer.h"
#include "WorldModelInterface.h"

#include "../basic/Common.h"
#include "../basic/Configuration.h"
#include "../basic/Log.h"
#include "../basic/StraightLine.h"

#include "PlayerTypes.h"

/** 
 * Aktueller Stand 
 * 
 *   $RCSfile: ConfiguredPlayer.cc,v $ 
 *   $Revision: 1.1.1.1 $ 
 *   $Date: 2001/07/31 22:23:09 $ 
 *   $Author: rami $  
 * 
 */ 

 
double ConfiguredPlayer::theStaminaMax;  
double ConfiguredPlayer::theRecoverDecThr;  
double ConfiguredPlayer::theRecoverDec;  
double ConfiguredPlayer::theRecoverMin;  
double ConfiguredPlayer::theEffortDecThr;  
double ConfiguredPlayer::theEffortDec;  
double ConfiguredPlayer::theEffortIncThr;  
double ConfiguredPlayer::theEffortInc;  
double ConfiguredPlayer::theMaxPower;  
double ConfiguredPlayer::theMinPower;  
double ConfiguredPlayer::theMaxMoment;  
double ConfiguredPlayer::theMinMoment;  
double ConfiguredPlayer::theMaxNeckMoment;  
double ConfiguredPlayer::theMinNeckMoment;
double ConfiguredPlayer::theMaxNeckAngle;  
double ConfiguredPlayer::theMinNeckAngle; 
double ConfiguredPlayer::theNormVisibleAngle;  
short  ConfiguredPlayer::isSetOnce=NO;  
 
 
ConfiguredPlayer::ConfiguredPlayer( int id,
				    const Vector& aPos,
				    const Vector& aSpeed,
				    double stamina,
				    double effort,
				    double recovery,
            const Confidence& aPosConfidence,
            const Confidence& aSpeedConfidence )	
  : ModelledObject( id, aPos, aSpeed, aPosConfidence, aSpeedConfidence ),
  theStamina( stamina ),
  theEffort( effort ),
  theRecovery( recovery )
{
  thePlayerType = globalPlayerTypes->typeOfPlayerWithId(id);

  if ( ConfiguredPlayer::isSetOnce == NO ) { 
    setClassVars();  
  } 
}
 
ConfiguredPlayer::ConfiguredPlayer(  
    const MovingPlayer& aMovingPlayer,   
    double stamina, 
    double effort, 
    double recovery,
    const Confidence& aPosConfidence,
    const Confidence& aSpeedConfidence )	
  : ModelledObject( aMovingPlayer.id(),
		    aMovingPlayer.pos(),  
		    aMovingPlayer.speed(),
		    aPosConfidence,
		    aSpeedConfidence ),
  theStamina( stamina ),
  theEffort( effort ),
  theRecovery( recovery )
{
  thePlayerType = globalPlayerTypes->typeOfPlayerWithId(aMovingPlayer.id());

  if ( isSetOnce == NO ) {
    setClassVars();  
  } 
}

void ConfiguredPlayer::setClassVars()  
{ 
  setStaminaMax(globalConf->playerStaminaMax);  
  setRecoverDecThr(globalConf->playerRecoverDecThr);  
  setRecoverDec(globalConf->playerRecoverDec);  
  setRecoverMin(globalConf->playerRecoverMin);  
  setEffortDecThr(globalConf->playerEffortDecThr);  
  setEffortDec(globalConf->playerEffortDec);
  setEffortIncThr(globalConf->playerEffortIncThr);  
  setEffortInc(globalConf->playerEffortInc);  
  setMaxPower(globalConf->maxPower);  
  setMinPower(globalConf->minPower);  
  setMaxMoment(radFromDegree(globalConf->maxMoment));  
  setMinMoment(radFromDegree(globalConf->minMoment));  
  setMaxNeckMoment(radFromDegree(globalConf->maxNeckMoment));  
  setMinNeckMoment(radFromDegree(globalConf->minNeckMoment)); 
  setMaxNeckAngle(radFromDegree(globalConf->maxNeckAngle));  
  setMinNeckAngle(radFromDegree(globalConf->minNeckAngle)); 
  setNormVisibleAngle(globalConf->visibleAngle);  
  ConfiguredPlayer::isSetOnce=YES;
}

void ConfiguredPlayer::setStaminaValues( double newStamina, double newEffortFactor, double newRecoveryFactor ) 
{ 
  setStamina( newStamina );
  setEffort( newEffortFactor );
  setRecovery( newRecoveryFactor );
}

double ConfiguredPlayer::effectiveDashPower( double power ) const
{
  power = normalizeDashPower( power );

  double stamina = min( theStamina, globalConf->playerStaminaMax );

  if( power > stamina ) {
    // Staminaverbrauch grer als Staminavorrat
    power = stamina;
  }
  if( -2.0 * power > stamina ) {
    // Staminaverbrauch grer als Staminavorrat
    // ( rckwrts laufen verbraucht doppelt Stamina )
    power = stamina / -2.0;
  }

  // HACK
  return power * min(theEffort, 1.0);
}

double ConfiguredPlayer::dashPowerForSend( double power ) const
{
  power = normalizeDashPower( power );
  // HACK
  power = power / min(theEffort,1.0);
  double thePowerForSend = normalizeDashPower( power );

  if ( fabs( thePowerForSend ) > stamina() ) {
    WARNING( Log::MODELLED_OBJECTS, "dashPowerForSend: thePowerForSend > stamina" );
  }
  if( isAlmostEqual( thePowerForSend, power ) == NO ) {
    WARNING( Log::MODELLED_OBJECTS, "dashPowerForSend: power value may be out of range" );  
  } 
 
  return thePowerForSend;  
} 
 

double ConfiguredPlayer::effectiveTurnAngle( double angle ) const  
{ 
  double relAngle = ( angle > Ath98_PI ) ? (2.0 * Ath98_PI) - angle : (-1.0 * angle);
  double relTurnAngle = relAngle / turnAngleMult();
  return (relTurnAngle >= 0.0) ? (2.0 * Ath98_PI) - relTurnAngle : (-1.0 * relTurnAngle);
} 

 
double ConfiguredPlayer::turnAngleForSend( double angle ) const  
{ 
  angle = normalizeAngle( angle );  
  if( angle > Ath98_PI ){
    double deviation = (2.0 * Ath98_PI) - angle;
    deviation *= turnAngleMult();
    deviation = min(deviation, Ath98_PI);
    angle = (2.0 * Ath98_PI) - deviation;
  }
  else{
    angle *=turnAngleMult();
    angle = min(angle, Ath98_PI);
  }
  return normalizeTurnAngle(angle);
} 
 

double ConfiguredPlayer::turnAngleMult() const  
{ 
  return( 1.0 + inertiaMomentOfTurn() * speed().length() );  
} 

 
void ConfiguredPlayer::recover( int simSteps )  
{
  while( simSteps > 0 ) {
    /* RecoveryFactor ausrechnen */ 
    if( theStamina <= recoverDecThr() * staminaMax() ) {
      if( theRecovery <= recoverMin() ){ 
        /* Minimum erreicht */ 
      }  
      else{
        theRecovery -= recoverDec();  
      } 
    } 

    /* EffortFactor ausrechnen */ 
    if( theStamina <= effortDecThr() * staminaMax() ) { 
      if( theEffort <= effortMin() ) { 
        /* Minimum erreicht */ 
      } 
      else{  
        theEffort -= effortDec();  
        theEffort = max( theEffort, effortMin() );
      } 
    } 
    
    if( theStamina >= effortIncThr() * staminaMax() ) { 
      if( theEffort >= effortMax() ) { 
        /* Maximum erreicht */ 
      } 
      else { 
        theEffort += effortInc();  
        theEffort = min( theEffort, effortMax() );
      } 
    } 
 
    /* Stamina ausrechnen */
    theStamina += (theRecovery * staminaIncMax());
    theStamina = min( theStamina, staminaMax() );   
    --simSteps;
  }
} 
