
#include <math.h>

#include "Field.h"
#include "WorldModelInterface.h"

#include "../basic/Configuration.h"
#include "../basic/InvalidIdException.h"
#include "../basic/Log.h"

/** 
 * Aktueller Stand 
 * 
 *   $RCSfile: Field.cc,v $ 
 *   $Revision: 1.1.1.1 $ 
 *   $Date: 2001/07/31 22:23:09 $ 
 *   $Author: rami $
 * 
 */ 
  
  
const double Field::theFieldLength         = 105;  
const double Field::theFieldWidth          = 68;  
const double Field::thePenaltyAreaLength   = 16.5;  
const double Field::thePenaltyAreaWidth    = 40.32;  
double Field::theGoalWidth                 = 7.32;  
const double Field::theKickoffCircleRadius = 9.15;  

Vector Field::theFlagPositions[] = { 
  // UPPER_LEFT_FLAG_ID 
  Vector( -Field::theFieldLength / 2, Field::theFieldWidth / 2 ),  
  // UPPER_MIDDLE_FLAG_ID 
  Vector( 0, Field::theFieldWidth / 2 ), 
  // UPPER_RIGHT_FLAG_ID 
  Vector( Field::theFieldLength / 2, Field::theFieldWidth / 2 ), 
  // LOWER_LEFT_FLAG_ID 
  Vector( -Field::theFieldLength / 2, -Field::theFieldWidth / 2 ), 
  // LOWER_MIDDLE_FLAG_ID 
  Vector( 0, -Field::theFieldWidth / 2 ),
  // LOWER_RIGHT_FLAG_ID
  Vector( Field::theFieldLength / 2, -Field::theFieldWidth / 2 ), 
  // UPPER_LEFT_PENALTY_FLAG_ID
  Vector( -Field::theFieldLength / 2 + Field::thePenaltyAreaLength,  
          Field::thePenaltyAreaWidth / 2 ), 
  // MIDDLE_LEFT_PENALTY_FLAG_ID 
  Vector( -Field::theFieldLength / 2 + Field::thePenaltyAreaLength, 0 ), 
  // LOWER_LEFT_PENALTY_FLAG_ID 
  Vector( -Field::theFieldLength / 2 + Field::thePenaltyAreaLength,  
          -Field::thePenaltyAreaWidth / 2 ), 
  // UPPER_RIGHT_PENALTY_FLAG_ID 
  Vector( Field::theFieldLength / 2 - Field::thePenaltyAreaLength,
          Field::thePenaltyAreaWidth / 2 ), 
  // MIDDLE_RIGHT_PENALTY_FLAG_ID 
  Vector( Field::theFieldLength / 2 - Field::thePenaltyAreaLength, 0 ), 
  // LOWER_RIGHT_PENALTY_FLAG_ID 
  Vector( Field::theFieldLength / 2 - Field::thePenaltyAreaLength,  
          -Field::thePenaltyAreaWidth / 2 ), 
  // LEFT_GOAL_CENTER_FLAG_ID 
  Vector( -Field::theFieldLength / 2, 0), 
  // RIGHT_GOAL_CENTER_FLAG_ID 
  Vector( Field::theFieldLength / 2, 0), 
 
  // CENTER_FLAG_ID 
  Vector( 0, 0), 
  // LEFT_GOAL_TOP_FLAG_ID 
  Vector( -Field::theFieldLength / 2, Field::theGoalWidth / 2 ), 
  // LEFT_GOAL_BOTTOM_FLAG_ID
  Vector( -Field::theFieldLength / 2, -Field::theGoalWidth / 2 ), 
  // RIGHT_GOAL_TOP_FLAG_ID 
  Vector( Field::theFieldLength / 2, Field::theGoalWidth / 2 ), 
  // RIGHT_GOAL_BOTTOM_FLAG_ID 
  Vector( Field::theFieldLength / 2, -Field::theGoalWidth / 2 ), 
 
  // OUTSIDE_TOP_LEFT_10_FLAG_ID 
  Vector( -10, Field::theFieldWidth / 2 + 5 ), 
  // OUTSIDE_TOP_LEFT_20_FLAG_ID 
  Vector( -20, Field::theFieldWidth / 2 + 5 ),
  // OUTSIDE_TOP_LEFT_30_FLAG_ID 
  Vector( -30, Field::theFieldWidth / 2 + 5 ),
  // OUTSIDE_TOP_LEFT_40_FLAG_ID 
  Vector( -40, Field::theFieldWidth / 2 + 5 ), 
  // OUTSIDE_TOP_LEFT_50_FLAG_ID 
  Vector( -50, Field::theFieldWidth / 2 + 5 ),
  // OUTSIDE_TOP_RIGHT_10_FLAG_ID 
  Vector( 10, Field::theFieldWidth / 2 + 5 ), 
  // OUTSIDE_TOP_RIGHT_20_FLAG_ID 
  Vector( 20, Field::theFieldWidth / 2 + 5 ), 
  // OUTSIDE_TOP_RIGHT_30_FLAG_ID 
  Vector( 30, Field::theFieldWidth / 2 + 5 ), 
  // OUTSIDE_TOP_RIGHT_40_FLAG_ID 
  Vector( 40, Field::theFieldWidth / 2 + 5 ), 
  // OUTSIDE_TOP_RIGHT_50_FLAG_ID 
  Vector( 50, Field::theFieldWidth / 2 + 5 ), 
  // OUTSIDE_BOTTOM_LEFT_10_FLAG_ID 
  Vector( -10, -Field::theFieldWidth / 2 - 5 ), 
  // OUTSIDE_BOTTOM_LEFT_20_FLAG_ID 
  Vector( -20, -Field::theFieldWidth / 2 - 5 ), 
  // OUTSIDE_BOTTOM_LEFT_30_FLAG_ID 
  Vector( -30, -Field::theFieldWidth / 2 - 5 ), 
  // OUTSIDE_BOTTOM_LEFT_40_FLAG_ID
  Vector( -40, -Field::theFieldWidth / 2 - 5 ),
  // OUTSIDE_BOTTOM_LEFT_50_FLAG_ID 
  Vector( -50, -Field::theFieldWidth / 2 - 5 ), 
  // OUTSIDE_BOTTOM_RIGHT_10_FLAG_ID 
  Vector( 10, -Field::theFieldWidth / 2 - 5 ), 
  // OUTSIDE_BOTTOM_RIGHT_20_FLAG_ID 
  Vector( 20, -Field::theFieldWidth / 2 - 5 ), 
  // OUTSIDE_BOTTOM_RIGHT_30_FLAG_ID 
  Vector( 30, -Field::theFieldWidth / 2 - 5 ), 
  // OUTSIDE_BOTTOM_RIGHT_40_FLAG_ID 
  Vector( 40, -Field::theFieldWidth / 2 - 5 ), 
  // OUTSIDE_BOTTOM_RIGHT_50_FLAG_ID 
  Vector( 50, -Field::theFieldWidth / 2 - 5 ), 
 
  // OUTSIDE_LEFT_TOP_10_FLAG_ID
  Vector( -Field::theFieldLength / 2 - 5, 10 ), 
  // OUTSIDE_LEFT_TOP_20_FLAG_ID
  Vector( -Field::theFieldLength / 2 - 5, 20 ),
  // OUTSIDE_LEFT_TOP_30_FLAG_ID 
  Vector( -Field::theFieldLength / 2 - 5, 30 ), 
  // OUTSIDE_LEFT_BOTTOM_10_FLAG_ID 
  Vector( -Field::theFieldLength / 2 - 5, -10 ), 
  // OUTSIDE_LEFT_BOTTOM_20_FLAG_ID 
  Vector( -Field::theFieldLength / 2 - 5, -20 ),
  // OUTSIDE_LEFT_BOTTOM_30_FLAG_ID 
  Vector( -Field::theFieldLength / 2 - 5, -30 ), 
  // OUTSIDE_RIGHT_TOP_10_FLAG_ID 
  Vector( Field::theFieldLength / 2 + 5, 10 ), 
  // OUTSIDE_RIGHT_TOP_20_FLAG_ID 
  Vector( Field::theFieldLength / 2 + 5, 20 ), 
  // OUTSIDE_RIGHT_TOP_30_FLAG_ID 
  Vector( Field::theFieldLength / 2 + 5, 30 ), 
  // OUTSIDE_RIGHT_BOTTOM_10_FLAG_ID 
  Vector( Field::theFieldLength / 2 + 5, -10 ), 
  // OUTSIDE_RIGHT_BOTTOM_20_FLAG_ID 
  Vector( Field::theFieldLength / 2 + 5, -20 ), 
  // OUTSIDE_RIGHT_BOTTOM_30_FLAG_ID 
  Vector( Field::theFieldLength / 2 + 5, -30 ), 
  // OUTSIDE_TOP_CENTER_FLAG_ID
  Vector( 0, Field::theFieldWidth / 2 + 5 ), 
  // OUTSIDE_BOTTOM_CENTER_FLAG_ID 
  Vector( 0, -Field::theFieldWidth / 2 - 5 ), 
  // OUTSIDE_LEFT_CENTER_FLAG_ID 
  Vector( -Field::theFieldLength / 2 - 5, 0 ), 
  // OUTSIDE_RIGHT_CENTER_FLAG_ID 
  Vector( Field::theFieldLength / 2 + 5, 0 ) 
};  
 
const Vector Field::theGoalPositions[] = { 
  Vector( Field::theFieldLength / 2, 0 ), 
  Vector( -Field::theFieldLength / 2, 0 )
};  

 
Field::Field()  
{ 
  theIdString = "Field";
  // siehe auch Zeile 302 "Field" 
  theGoalWidth = globalConf->goalWidth;  
  // Neusetzen der von der Torbreite abhaengigen Flaggen 
  theFlagPositions[LEFT_GOAL_TOP_FLAG_ID] = 
    Vector( -Field::theFieldLength / 2, Field::theGoalWidth / 2 );  
  theFlagPositions[LEFT_GOAL_BOTTOM_FLAG_ID] = 
    Vector( -Field::theFieldLength / 2, -Field::theGoalWidth / 2 );  
  theFlagPositions[RIGHT_GOAL_TOP_FLAG_ID] =
    Vector( Field::theFieldLength / 2, Field::theGoalWidth / 2 );  
  theFlagPositions[RIGHT_GOAL_BOTTOM_FLAG_ID] =
    Vector( Field::theFieldLength / 2, -Field::theGoalWidth / 2 );  
} 
 
 
const Vector& Field::posOfPoint(int anId) const 
{ 
  if ( anId >= MIN_FLAG_ID && anId <= MAX_EXISTING_FLAG_ID ) { 
    return theFlagPositions[anId - MIN_FLAG_ID];  
  } 
  if ( anId >= MIN_GOAL_ID && anId <= MAX_EXISTING_GOAL_ID ) { 
    return theGoalPositions[anId - MIN_GOAL_ID];
  }  
  else {
    throwException( InvalidIdException(
      "Id of point was neither a flag-id nor a goal-id." ) );
  }
}


bool Field::isInsideField( const Vector& point )
{
  if ( ( fabs( point.x() ) > (double)fieldLength() / 2.0 ) ||
       ( fabs( point.y() ) > fieldWidth() / 2 )  ) {
    return NO;
  }
  else {
    return YES;
  }
}

Vector Field::calculatePosition( const vector<StableObject>& seenFlags ) const
{
  INFO( Log::WORLD_MODEL , 8 , idString() , globalLog->infoStream 
    << "Zahl der gesehenen Flaggen: " << seenFlags.size() << endl );

  // die Odometrieposition --> wurde beim abschicken der letzten Aktion schon berechnet
  // --> steht also in der aktuellen Situation
  Vector position = globalWorldModel->player().pos();

  int flagCount = seenFlags.size();
  if ( flagCount == 0 ) {
    WARNING( Log::WORLD_MODEL, "No flags seen!  Use simulated Position! " );
    return position;
  }

  // Fehler, gesehene Entfernung und Position der Flagge schon mal ermitteln
  vector <double> seenDistance;
  vector <double> flagError;
  vector <Vector> flagPosition;

  for( unsigned int i = 0; i < seenFlags.size(); ++i ){
    seenDistance.push_back( seenFlags[i].pos().length() );
    flagError.push_back( 0.0045 * seenDistance[i] + 0.05 );
    flagPosition.push_back( posOfPoint( seenFlags[i].id() ) );
  }

  if( globalWorldModel->side() == RIGHT_SIDE_ID ) {
    position = Vector( -position.x(), -position.y() );
  }

  // GradientenAlgorithmus
  int iteration = 0, index;
  double distance, angle, maxDistance = 1.0;
  while( (iteration < 20) && (fabs(maxDistance) > 0.001) ){
    ++iteration;
    maxDistance = 0.0;
    index = -1;
    for( unsigned int i = 0; i < seenFlags.size(); ++i ){
      distance = fabs( (flagPosition[i] - position).length() - seenDistance[i] ) - flagError[i];
      if( distance < 0.0 ) distance = 0.0;
      if( distance > maxDistance ){
        maxDistance = distance;
        index = i;
      }
    }
    if( index > -1 ){
      distance = (flagPosition[index] - position).length() - seenDistance[index];
      if( distance < 0 ) distance += flagError[index];
      else distance -= flagError[index];
      angle = (flagPosition[index] - position).angle();
      /* Sinus und Cosinus noch rausnehmen und durch Tabelle ersetzen */
      position.set( position.x() + distance * cos(angle),
                    position.y() + distance * sin(angle) );
    }
  }

  if( globalWorldModel->side() == RIGHT_SIDE_ID ) {
    position = Vector( -position.x(), -position.y() );
  }

  INFO( Log::WORLD_MODEL, 7, idString(), globalLog->infoStream 
    << " ermittelte Position: " << position << endl ) ;

  return position;
}


double Field::computeFaceDir( const vector<StableObject>& seenLines )
{
  double angle = 0.0;
  int lineId = 0;

  // zuerst die Linie auswaehlen, deren Winkel ausgewertet wird.
  switch ( seenLines.size() ) {
    case 0:
      // wird keine line gesehen, wird die simulierte FaceDir bernommen
      WARNING( Log::WORLD_MODEL,
        "Cannot compute face dir. Number of seen lines is 0." );
      return globalWorldModel->player().faceDir();
      break;
    case 1:
      // Position ist innerhalb des Spielfeldes.
      angle = seenLines[0].pos().angle();
      lineId = seenLines[0].id();
      break;
    default:
      // Die zweite Linie wird von innen gesehen, daher kann der Winkel
      // berechnet werden als waere man im inneren des Feldes.
      int i;
      if( seenLines[0].pos().length() > seenLines[1].pos().length() ) {
        i = 0;
      }
      else {
        i = 1;
      }
      angle = seenLines[i].pos().angle();
      lineId = seenLines[i].id();
      break;
  }

  if( lineId == 0 ) cerr << "Achtung : lineId = 0 !!!! --> Field.cc , computeFaceDir" << endl;

  // Jetzt abhaengig von der LinienId die absolute Blickrichtung berechnen
  INFO( Log::WORLD_MODEL , 10 , "Field" , globalLog->infoStream 
    << lineId << " Winkel: " << degFromRadiant( angle ) << endl );

  switch ( lineId ) {
  case UPPER_LINE_ID:
    if ( angle > Ath98_PI ) {
      return 2.0*Ath98_PI - angle;
    }
    else {
      return Ath98_PI - angle;
    }
    break;
  case LOWER_LINE_ID:
    if ( angle > Ath98_PI ) {
      return 3.0*Ath98_PI - angle;
    }
    else {
      return 2.0*Ath98_PI - angle;
    }
    break;
  case LEFT_LINE_ID:
    if ( angle > Ath98_PI ) {
      return 2.5*Ath98_PI - angle;
    }
    else {
      return 1.5*Ath98_PI - angle ;
    }
    break;
  case RIGHT_LINE_ID:
    if ( angle > Ath98_PI ) {
      return 3.5*Ath98_PI - angle;
    }
    else {
      return 0.5*Ath98_PI - angle;
    }
    break;

  default:
    WARNING( Log::WORLD_MODEL, "Unknown line id in computeFaceDir" );
    return globalWorldModel->player().faceDir();
    break;
  }
}
