#ifndef _MOVINGPLAYER_H_
#define _MOVINGPLAYER_H_

#include "MovingObject.h"

/**
  Ein gesehender Spieler.
  Ein MovingPlayer hat zusaetzlich zu den Eigenschaften eines MovingObject noch
  eine Blickrichtung. Die Blickrichtung gibt an, in welche Richtung der Spieler
  guckt, also auch in welche Richtung er beim naechsten Dash beschleunigen
  wird.
  
  MovingPlayer wird in \Ref{Sensors} eingesetzt, um die wahrgenommenen Spieler
  zu repr\"asentieren.

  {\small {\bf Quelltext}

  [\URL[Header]{src/MovingPlayer.h} | [\URL[Implementierung]{src/MovingPlayer.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: MovingPlayer.h,v $
  *   $Revision: 1.1.1.1 $
  *   $Date: 2001/07/31 22:23:09 $
  *   $Author: rami $ 
  *
  **************************************************
  \end{verbatim} }*/
class MovingPlayer : public MovingObject
{
public:

  /** @name Erzeugen und Zerst\"oren */
  //@{  
    /** Default-Konstruktor. Alle Werte werden auf undef gesetzt. */
    MovingPlayer();

    /** Neuer MovingPlayer mit Id, Position, Geschwindigkeit und
      Blickrichtung. Geschwindigkeit und Blickrichtung sind optional.
      Sind sie nicht angegeben, werden die Werte auf undef gesetzt.
      @param id Die Id des Spielers
      @param aPos Die Position.
      @param aSpeed Die Geschwindigkeit.
      @param aFaceDir Die Blickrichtung. */
    MovingPlayer( int id,
                  bool isGoalie,
                  const Vector& aPos, 
                  const Vector& aSpeed = Vector(),
                  double aFaceDir = 0.0,
                  double aBodyDir = 0.0 );

  //@}

  /** @name Werte abfragen */
  //@{
    /** liefert Team-Id */
     int teamId() const;

    /** Liefert die Blickrichtung. */
    double faceDir() const;
    
    /** Liefert die Laufrichtung */
    double bodyDir() const;

    /** Liefert die Goalie-Eigenschaft */
    bool isGoalie() const;

  //@}
  
  /** @name Werte setzen */
  //@{
    /** Setzt die Blickrichtung neu. */
    void setFaceDir( double newFaceDir );
    
    /** Setzt die Laufrichtung neu. */
    void setBodyDir( double newBodyDir );

    /** Setzt die Goalie-Eigenschaft neu. */
    void setIsGoalie( bool isGoalie );
    
  //@}

private:  
  /** Handelt es sich um den Golie ? */
  bool isTheGoalie;

  /** Die Blickrichtung als absoluter Winkel im Bogenmass. */
  double theFaceDir;

  /** Die Laufrichtung als absoluter Winkel im Bogenmass. */
  double theBodyDir;
};

inline double MovingPlayer::faceDir() const {
  return theFaceDir;
}

inline double MovingPlayer::bodyDir() const {
  return theBodyDir;
}

inline bool MovingPlayer::isGoalie() const {
  return isTheGoalie;
}

inline void MovingPlayer::setBodyDir(double newBodyDir) {
  theBodyDir = newBodyDir;
}

inline void MovingPlayer::setIsGoalie(bool isGoalie) {
  isTheGoalie = isGoalie;
}

#endif
