#include "PlayerTypes.h"
#include "../basic/Configuration.h"
#include "ModelledPlayer.h"

PlayerTypes *globalPlayerTypes; 

PlayerTypes::PlayerTypes()
{
  for(int i=0; i<numberOfPlayerTypes; ++i){
// hier sollten eventuell die Standardwerte eingelesen werden (von wo auch immer)
    thePlayerMaxSpeed[i] = 1.0;
    theStaminaIncMax[i] = 45.0;
    thePlayerDecay[i] = 0.4;
    theInertiaMoment[i] = 5.0;
    theDashPowerRate[i] = 0.006;
    thePlayerSize[i] = 0.3;
    theKickableMargin[i] = 0.7;
    theKickRand[i] = 0.0;
    theExtraStamina[i] = 0.0;
    theEffortMax[i] = 1.0;
    theEffortMin[i] = 0.6;

    calculateDependentParameters(i);
  }

  for(int i=0; i<22; ++i){
    thePlayerMap[i] = 0;
  }
  for(int i=0; i<numberOfPlayerTypes; ++i){
    parsedPlayers[i] = NO;
  }
}

PlayerTypes::~PlayerTypes()
{
}

void PlayerTypes::parsePlayerTypeMessage(const Ath98String& message)
{
  int id;
  double playerSpeedMax, staminaIncMax, playerDecay, inertiaMoment, dashPowerRate, playerSize, kickableMargin, kickRand, extraStamina, effortMax, effortMin;
  sscanf( message.cString(),"(player_type %d %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf)",
          &id, &playerSpeedMax, &staminaIncMax, &playerDecay, &inertiaMoment, &dashPowerRate,
          &playerSize, &kickableMargin, &kickRand, &extraStamina, &effortMax, &effortMin );
  thePlayerMaxSpeed[id] = playerSpeedMax;
  theStaminaIncMax[id]  = staminaIncMax;
  thePlayerDecay[id]    = playerDecay;
  theInertiaMoment[id]  = inertiaMoment;
  theDashPowerRate[id]  = dashPowerRate;
  thePlayerSize[id]     = playerSize;
  theKickableMargin[id] = kickableMargin;
  theKickRand[id]       = kickRand;
  theExtraStamina[id]   = extraStamina;
  theEffortMax[id]      = effortMax;
  theEffortMin[id]      = effortMin;

  calculateDependentParameters(id);

  parsedPlayers[id] = YES;

  bool test = YES;
  for(int i=0; i<numberOfPlayerTypes; ++i){
    test &= parsedPlayers[i];
  }
  if( test == YES ){
    generatePlayerDistTable();
//    printPlayerDistTable();
  }
}


void PlayerTypes::parseSubstituteMessage(const Ath98String& message)
{
  int uid, playerType;
  int rc = sscanf( message.cString(),"(change_player_type %d %d)",
          &uid, &playerType );
  if( rc == 2 ){
    // wir haben ausgewechselt
    thePlayerMap[uid-1] = playerType;
  }
  else{
    // Gegner hat ausgewechselt
    // wir wissen aber vorerst nicht welchen Typ!
    // das macht der Coach spter (hoffentlich)
    thePlayerMap[uid-1+11] = 0;
  }
}


void PlayerTypes::calculateDependentParameters(int aId)
{
  theKickRange[aId] = theKickableMargin[aId] + thePlayerSize[aId] + globalConf->ballSize;
}


void PlayerTypes::generatePlayerDistTable()
{
  for(int i=0; i<numberOfPlayerTypes; ++i) {
    ModelledPlayer mPlayer = ModelledPlayer(420, Vector(0.0,0.0), Vector(0.0,0.0), 0.0, 0.0, 
                               VIEW_WIDTH_WIDE, VIEW_QUALITY_HIGH, globalConf->playerStaminaMax,
                               1.0, 1.0, NO, Confidence( 0, 0.0, 1.0 ), Confidence( 0, 0.0, 1.0 ),
                               Confidence( 0, 0.0, 1.0 ), Confidence( 0, 0.0, 1.0 ), Confidence( 0, 0.0, 1.0 ),
                               Confidence( 0, 0.0, 1.0 ), Confidence( 0, 0.0, 1.0 ) );
    mPlayer.setPlayerType(i);
    double neededPower = 0;
    double power = 0;
    for( int o = 0; o <= 8; ++o) {
      mPlayer = ModelledPlayer(420, Vector(0.0,0.0), Vector(0.1 * o,0.0), 0.0, 0.0,
                               VIEW_WIDTH_WIDE, VIEW_QUALITY_HIGH, globalConf->playerStaminaMax,
                               1.0, 1.0, NO, Confidence( 0, 0.0, 1.0 ), Confidence( 0, 0.0, 1.0 ),
                               Confidence( 0, 0.0, 1.0 ), Confidence( 0, 0.0, 1.0 ), Confidence( 0, 0.0, 1.0 ),
                               Confidence( 0, 0.0, 1.0 ), Confidence( 0, 0.0, 1.0 ) );

      for( int t = 0; t < 101; ++t ){
        theMaxPlayerDist[i][t][o] = mPlayer.pos().length();
        neededPower = ( mPlayer.maxSpeed() - mPlayer.speed().length() ) / mPlayer.dashPowerRate();
        power = neededPower / mPlayer.effort() ;
        power = min( power, mPlayer.maxPower() );
        power = min( power, mPlayer.stamina()-1200 );
        mPlayer.accelerate(power);
        mPlayer.simulateSteps(1);
      }
    }
  }
}


void PlayerTypes::printPlayerDistTable() const
{
cout << "\n***************************************************\n";
  for(int t=0; t<=50; ++t) {
//    for(int i=0; i<numberOfPlayerTypes; ++i) {
    for(int i=0; i<=8; ++i) {
      cout << theMaxPlayerDist[0][t][i] << '\t' << '\t' ;
    }
    cout << endl;
  }
}