
/* 
 * Aktueller Stand 
 * 
 *   $RCSfile: Situation.cc,v $ 
 *   $Revision: 1.2 $ 
 *   $Date: 2001/08/02 08:09:42 $add 
 *   $Author: rami $  
 * 
 */ 
  
#include "Situation.h" 
#include "../basic/Log.h"
#include "../basic/InvalidIdException.h"
#include "../worldmodel/WorldModelInterface.h"
 

Situation::Situation( int aSimStep )
  : theSimStep( aSimStep ),
    thePlayers( vector<ModelledPlayer>(0) ),
    areStepsToBallEstimated( NO ), 
    isDistToBallCalculated( NO ), 
    nearestTeammate(UNDEF_INT),
	  secondNearestTeammate(UNDEF_INT),
	  nearestEnemy(UNDEF_INT),
    theOffsideLineX( UNDEF_DOUBLE ), 
    theOffsideLineXConfidence()
{
} 
 
 
void Situation::setPlayer( const ModelledPlayer& newPlayer ) 
{ 
  int index = indexOfPlayerWithId( newPlayer.id() );  

  if( index == UNDEF_INT ) {
    addPlayer( newPlayer );
  }
  else{
    setPlayer( newPlayer, index );  
  }
}


int Situation::indexOfPlayerWithId( int aPlayerId ) const
{
  if( aPlayerId == SELF_ID ){
    aPlayerId = globalWorldModel->id();
  }
  
  for( unsigned int i = 0;  i < thePlayers.size();  ++i  ) {
    if ( thePlayers[i].id() == aPlayerId ) {
      return i;
    }
  }
  return UNDEF_INT;
}


int Situation::idOfLastPlayer() const
{
  int index = UNDEF_INT;
  double minX = UNDEF_INT;
  for( unsigned int i = 0;  i < thePlayers.size();  ++i ) {
    if( ( thePlayers[i].team() == OUR_TEAM_ID ) &&
	    ( thePlayers[ i ].isGoalie() == NO ) &&
	    ( thePlayers[i].pos().x() < minX ) ){
        index = i;
    }
  }
  if( index == UNDEF_INT ){
    ERROR( "No last Player found\n" );
    throw;
  }
  return thePlayers[index].id();
}


void Situation::removePlayerWithIndex( int index )
{
  if ( playerWithIndex( index ).team() == THEIR_TEAM_ID ) {
    theOffsideLineX = UNDEF_DOUBLE;
  }
  thePlayers.erase( &thePlayers[index] );
  areStepsToBallEstimated = NO;
}


void Situation::removePlayerWithId( int aPlayerId )
{
  int playerIndex = indexOfPlayerWithId( aPlayerId );
  if( playerIndex != UNDEF_INT ){
    removePlayerWithIndex( playerIndex );
  }
}


void Situation::removeOwnGoalie(){
  for( unsigned int i = 0; i < thePlayers.size(); ++i ){
    if( thePlayers[ i ].isGoalie() &&	thePlayers[ i ].team() == OUR_TEAM_ID ){
      thePlayers.erase( &thePlayers[i] );
      return;
    }
  }
}


const ModelledPlayer& Situation::playerWithId( int aPlayerId ) const 
{ 
  int playerIndex = indexOfPlayerWithId( aPlayerId );
  if( playerIndex == UNDEF_INT ){
    // HACK
    //Um nicht mit einer Exception zu beenden, wenn wir in einer Methode keinen Spieler
    //zurckgeben knnen, definieren wir einen static UNDEF_PLAYER
    static ModelledPlayer UNDEF_PLAYER = ModelledPlayer();
    
    return UNDEF_PLAYER;
  }
  return thePlayers[playerIndex];  
} 


const int Situation::canISeePlayerWithId( int aPlayerId ) const
{
  for( unsigned int i = 0; i < thePlayers.size(); ++i ) {
    if ( thePlayers[i].id() == aPlayerId ) {
      return YES;
    }
  }
  return NO;
}


const ModelledPlayer& Situation::playerWithIndex( int index ) const
{ 
  return thePlayers[index];  
} 

 
bool Situation::containsPlayerWithId( int aPlayerId ) const 
{ 
  int playerIndex = indexOfPlayerWithId( aPlayerId );  
  if( playerIndex == UNDEF_INT ){
    return NO;  
  } 
  else{ 
    return YES;  
  } 
} 


void Situation::simulateSteps( unsigned int stepCount )  
{ 
	stepCount = max(stepCount, 0);
	stepCount = min(stepCount, 100);

  Vector playerToBall;  

  // Kollisionstest scheint nach mehreren Simulationsschritten nicht mehr sinnvoll,
  // daher testen wir nur nach den ersten beiden Simulationsschritten auf Kollision
  // und setzen pos und speed entsprechend zurck 

  for( unsigned int count = 1; count <= stepCount; ++count){
    if( theBall.posConfidence().reliability() > 0.01 ){
      theBall.simulateSteps( 1 );
    }
    for( int playerIndex = playerCount() - 1;  playerIndex >= 0;  --playerIndex ) {
      thePlayers[playerIndex].simulateSteps( 1 );  
      playerToBall = theBall.pos() - thePlayers[playerIndex].pos();

      // kollidiert ein Spieler mit dem Ball ?
      if ( (playerToBall.length() < thePlayers[playerIndex].size() + globalConf->ballSize) && (count < 3 ) ){
        INFO( Log::SITUATION , 10 , idString() , globalLog->infoStream 
          << " Situation::simulateSteps collision with ball detected\n"  ) ; 

        playerToBall.setPolar( thePlayers[playerIndex].size() + globalConf->ballSize, playerToBall.angle() );  
        theBall.setPos( thePlayers[playerIndex].pos() + playerToBall );  
        theBall.setSpeed( -0.1 * theBall.speed() );
        thePlayers[playerIndex].setSpeed( -0.1 * thePlayers[playerIndex].speed() );
      }
    }
  }
  theSimStep += stepCount;  
  areStepsToBallEstimated = NO;  
  theOffsideLineX = UNDEF_DOUBLE;
}

 
int Situation::indexOfNearestPlayer( int teamId ) const 
{ 
  if( areStepsToBallEstimated == NO ) { 
    estimateForAllStepsToBall();  
  } 
  if( teamId == BOTH_TEAMS_ID ) {
    if( nearestEnemySteps <= nearestTeammateSteps ) {
      return nearestEnemy;
    }
    else {
      return nearestTeammate;		
     }
  }
  else if( teamId == OUR_TEAM_ID ) {
    return nearestTeammate;
  }
  else if( teamId == THEIR_TEAM_ID ) {
    return nearestEnemy;
  }
  else {
    cerr << "Situation::indexOfNearestPlayers unknown TEAM_ID\n";
  }
  // WICHTIG!
  // an allen Stellen, wo indexOfNearestPlayer gerufen wird, muss berprft werden,
  // ob der Rckgabewert UNDEF_INT ist ! - sonst Absturz mglich
  cerr << "Situation::indexOfNearestPlayer returns UNDEF_INT\n";
  return UNDEF_INT;
}

int Situation::indexOfSecondNearestTeammate() const
{
  if( areStepsToBallEstimated == NO ) { 
    estimateForAllStepsToBall();  
  } 
  return secondNearestTeammate;
}	


const ModelledPlayer& Situation::nearestPlayer( int teamId ) const 
{ 
  int aPlayerIndex = indexOfNearestPlayer( teamId );
  if( aPlayerIndex == UNDEF_INT){
    // HACK
    //Um nicht mit einer Exception zu beenden, wenn wir in einer Methode keinen Spieler
    //zurckgeben knnen, definieren wir einen static UNDEF_PLAYER
    static ModelledPlayer UNDEF_PLAYER = ModelledPlayer();
    return UNDEF_PLAYER;
  }
  return thePlayers[ aPlayerIndex ];  
} 


int Situation::stepsOfNearestPlayer( int teamId ) const 
{ 
  if( areStepsToBallEstimated == NO ) { 
    estimateForAllStepsToBall();  
  } 
  if ( nearestEnemySteps == UNDEF_INT ) nearestEnemySteps = 1000;
  if ( nearestTeammateSteps == UNDEF_INT ) nearestTeammateSteps = 1000;

  if( teamId == BOTH_TEAMS_ID ) {
    if( nearestEnemySteps <= nearestTeammateSteps ) {
      return nearestEnemySteps;
    }
    else {
      return nearestTeammateSteps;		
     }
  }
  else if( teamId == OUR_TEAM_ID ) {
    return nearestTeammateSteps;
  }
  else if( teamId == THEIR_TEAM_ID ) {
    return nearestEnemySteps;
  }
  else {
    cerr << "Situation::stepsOfNearestPlayers unknown TEAM_ID\n";
  }
  cerr << "Situation::stepsOfNearestPlayer returns 1000\n"; // UNDEF_INT\n";
  return 1000; // UNDEF_INT;
} 


void Situation::estimateForAllStepsToBall() const
{ 
	nearestTeammate = UNDEF_INT;
	secondNearestTeammate = UNDEF_INT;
	nearestEnemy = UNDEF_INT;

	nearestTeammateSteps = 1000;
	secondNearestTeammateSteps = 1000;
	nearestEnemySteps = 1000; 
	Vector ballPos = theBall.pos();
	Vector ballSpeed = theBall.speed();

	int takt = 1; // Takt 0 sollte besser nicht geprueft werden, damit wir den Ball im Konfliktfall wegpassen

	if ( fabs(ballSpeed.x()) > 1000 ) { // Ballspeed ungltig?
		ballSpeed = Vector( 0.0, 0.0 );
	}
 
	ballPos += ballSpeed;
	ballSpeed = globalConf->ballDecay * ballSpeed;	

	if( playerCount() > 0 && fabs(ballPos.x()) < 1000 ) { // Test auf gltigen Ball 
		while( takt < 50 && ( nearestTeammate == UNDEF_INT || secondNearestTeammate == UNDEF_INT || nearestEnemy == UNDEF_INT) && globalWorldModel->field().isInsideField(ballPos)) { 
			for( int i = 0; i < playerCount(); ++i ) {
				if( playerWithIndex(i).getMaxDistAfterSteps( takt ) > playerWithIndex(i).distTo( ballPos )) {  
					if( playerWithIndex(i).team() == OUR_TEAM_ID) {
						if( nearestTeammate == UNDEF_INT ) {
							nearestTeammate = i;
							nearestTeammateSteps = takt;
						}	
						else if( (secondNearestTeammate == UNDEF_INT) && (nearestTeammate != i) ) {
							secondNearestTeammate = i;
							secondNearestTeammateSteps = takt;
						}	
					}
					else if( nearestEnemy == UNDEF_INT ) {
						nearestEnemy = i;
						nearestEnemySteps = takt;
					}
				}
			}	
			++takt;
			ballPos += ballSpeed;
			ballSpeed = globalConf->ballDecay * ballSpeed;	

		}
	}
	
	areStepsToBallEstimated = YES; 
} 

 


int Situation::indexOfNearestPlayerToPos( const Vector& otherPos,  
    const int teamId ) const 
{ 
  double dist;  
  double mindist = DBL_MAX;  
  int index = UNDEF_INT;  

  if ( teamId != OUR_TEAM_ID && teamId != THEIR_TEAM_ID && teamId != BOTH_TEAMS_ID ) { 
    cerr << "Situation::indexOfNearestPlayerToPos  undef TeamId\n";
    return UNDEF_INT;
  } 
  
  for ( int i = 0;  i < playerCount();  ++i ) { 
      // Alle Spieler ueberspringen, die nicht zu dem erfragten Team gehoeren 
    int playerTeamId = thePlayers[i].team();  
    if ( playerTeamId == UNDEF_TEAM_ID ) { 
      playerTeamId = THEIR_TEAM_ID;  
    } 
    if ( playerTeamId != teamId && teamId != BOTH_TEAMS_ID ) { 
      continue;  
    } 
 
    dist = ( thePlayers[i].pos() - otherPos ).length();  
    if ( dist < mindist ) { 
      mindist = dist;  
      index = i;  
    } 
  } 
  return index;  
}


double Situation::distOfNearestPlayerToPos( const Vector& aPos, const int teamId ) const 
{ 
  int index = indexOfNearestPlayerToPos(aPos, teamId);
  if( index == UNDEF_INT ){
    cerr << "Situation::distOfNearestPlayerPos  undef PlayerIndex\n";
    return 1000.0;
  }
  return ( thePlayers[index].pos() - aPos ).length();  
} 


Ath98String Situation::toString() const 
{ 
  Ath98String output;  
  output = "( Situation at simStep " + Ath98String( simStep() ) + ":\n"
    + "ball: " + ball().toString() + ",\n" + Ath98String( playerCount() ) + " players:";  
  for( int i = 0;  i < playerCount();  ++i) { 
    output = output + "\n   " + playerWithIndex(i).toString();  
  } 
  if( playerCount() > 0 ) { 
    output = output + "\n relative ball: Pos: " 
      + (ball().pos() - playerWithIndex(0).pos()).toString() + " Speed: " 
      + (ball().speed() - playerWithIndex(0).speed()).toString();  
  } 
  output = output + " )\n";  
  return output;  
} 


double Situation::offsideLineX() const 
{ 
  if ( theOffsideLineX == UNDEF_DOUBLE ) { 
    theOffsideLineX = computeOffsideLineX();    
  } 
  return theOffsideLineX;  
} 


double Situation::computeOffsideLineX() const 
{ 
  INFO( Log::SITUATION , 3 , idString() , globalLog->infoStream 
    << " ENTER Situation::computeOffsideLineX\n" ); 

  if( ball().pos().x() < 0.0 - ball().posConfidence().maxError() ){
    // befindet sich der Ball in unserer Spielhlfte (mit kleiner Sicherheit)
    // dann gibt es kein Abseits und die Abseitslinie wird auf die gegnerische
    // Grundlinie verschoben 
    // nicht ganz, Problem: wenn der Ball die Mittellinie berquert, stehen wir sofort im Abseits
    // darum Abseitslinie wie sonst
//    return globalWorldModel->field().fieldLength() / 2.0;
  }
      
  // der Ball ist auf jeden Fall in der gegnerischen Hlfte
  // die Abseitslinie muss also mindestens auf Ballhhe sein
 
  double xOfLastEnemy = 0.0;
  double xOf2ndLastEnemy = 0.0;
  // x-Koordinaten des letzten gegnerischen Spielers (nicht Goalie!) bestimmen. (also vorletzter) (OK, nicht ganz sauber; aber hey...)
  for( int i = 0;  i < playerCount();  ++i) {
    const ModelledPlayer& player = playerWithIndex( i );  
    if ( (player.team()  != OUR_TEAM_ID) && (player.isGoalie() == NO) ) { 
      INFO( Log::SITUATION  , 8 , idString() , globalLog->infoStream 
        << " << Checking player " << i << " << ( " << player.id()  
        << " ) x: " << player.pos().x() << endl ); 
  
      if ( player.pos().x() - player.posConfidence().maxError() > xOfLastEnemy ) { 
        xOf2ndLastEnemy = xOfLastEnemy;
        xOfLastEnemy = player.pos().x();
      } 
    } 
  }
  double offsideLineX = max( ball().pos().x() , 0.0 );
  offsideLineX = max( offsideLineX, xOf2ndLastEnemy );
  theOffsideLineXConfidence = Confidence( simStep(), 1.0, 1.0 ); // HACK
   
  INFO( Log::SITUATION, 5, idString(), globalLog->infoStream
    << " offsideLineX: " << offsideLineX << endl) ; 
  INFO( Log::SITUATION, 3, idString(), globalLog->infoStream
    << " LEAVE Situation::computeOffsideLineX\n" ) ; 
   
  return offsideLineX;  
}   

 
bool Situation::isOffsideWithIndex( int index ) const 
{ 
  if( index == UNDEF_INT ){
    cerr << "Situation::isOffsideWithIndex   Index = UNDEF_INT\n";
    return NO;
  }
  if( playerWithIndex( index ).team() == THEIR_TEAM_ID ) { 
    cerr << "Situation::isOffsideWithIndex   Can only check offside condition for team members.\n";  
    return NO;
  } 
  return ( playerWithIndex( index ).pos().x() > offsideLineX() );  
} 

 
bool Situation::isOffsideWithId( int playerId ) const 
{ 
  if( playerId == UNDEF_PLAYER_ID ){
    return NO;
  }
  int index = indexOfPlayerWithId( playerId );  
  if( index == UNDEF_INT ){
    return NO;
  }
  return ( isOffsideWithIndex( index ) );  
} 

     
Situation* Situation::copy() const {
  return new Situation( *this );  
} 

 

int Situation::idOfGoalie( int aTeamId ){
  for( int i = 0; i < playerCount(); ++i ){
    if( (thePlayers[i].isGoalie() == YES) && (thePlayers[i].team() == aTeamId) ){
      return thePlayers[i].id();
    }
  }
  return UNDEF_PLAYER_ID;
}


void Situation::calculatePlayerDistToBallTable() {
  theIndexOfXNearestPlayer.clear();
  theBallDistOfXNearestPlayer.clear();

  Vector ballPos = ball().pos();
  double dist;
  int pos=0;
  for(int i=0; i<playerCount(); ++i) {
    if( playerWithIndex(i).team() == OUR_TEAM_ID ) {
      dist = playerWithIndex(i).distTo(ballPos);
      for( pos=0; pos < (int)theBallDistOfXNearestPlayer.size(); ++pos ) {
        if( theBallDistOfXNearestPlayer[pos] > dist) {
          break;
        }
      }
      theBallDistOfXNearestPlayer.insert(theBallDistOfXNearestPlayer.begin()+pos, dist);
      theIndexOfXNearestPlayer.insert(theIndexOfXNearestPlayer.begin()+pos, i);
    }
  }
  isDistToBallCalculated = YES;
}



    