/**
 * Aktueller Stand
 *
 *   $RCSfile: WorldModelCore.cc,v $
 *   $Revision: 1.1.1.1 $
 *   $Date: 2001/07/31 22:23:09 $
 *   $Author: rami $
 *
 */

#include <math.h>

#include "WorldModelCore.h"
#include "WorldModelCommunicator.h"
#include "WorldModelIntegrator.h"


#include "../basic/Log.h"
#include "../basic/Timer.h"


 
WorldModelCore::WorldModelCore(Sensors *sensors) : theSensors( sensors ) 
{
  // create situation array
  historyCount = globalConf->historyCount;
  situationFieldSize = 2*historyCount + 1;
  theSituations = vector<Situation>( situationFieldSize );
  for ( int i = 0;  i < situationFieldSize;  ++i ) {
    theSituations.push_back( Situation( 0 ) );
  }
  currentSituationIndex = historyCount;
  theCorrectSituation=Situation( 0 );
  
  // zu Spielbeginn wissen wir, das der Ball auf (0,0) liegt und sich nicht bewegt
  ModelledBall ball = ModelledBall( BALL_ID, Vector(0, 0), Vector(0, 0),
                      UNDEF_TEAM_ID, -1,
                      Confidence( 0, 0.0, 1.0 ),
                      Confidence( 0, 0.0, 1.0 ) );
  theSituations[currentSituationIndex].setBall( ball );

  // initialize player related values
  // bei der Initialisierung des Weltmodells liegen diese Daten noch
  // nicht vor, da sich der Spieler noch nicht beim Server angemeldet hat
  thePlayerNumber = UNDEF_INT;
  theRole = GENERIC_ROLE;
  theTeamName = Ath98String( "Unknown Team" );
  theSide = NO_SIDE_ID;
  theScore[0] = 0; theScore[1] = 0;
  theInitiative = NO_TEAM_ID;
  theGameState = WorldModelInterface::BEFORE_KICK_OFF;
  theStartTimeOfElapsingGameState = 0;
  theInitialPositions.push_back( Vector() );
  theInitialPositions.push_back( Vector() );
  theInitialPositions.push_back( Vector() );
  theInitialPositions.push_back( Vector() );
  theBasePositions = theInitialPositions;

  // initialize simstep related values
  theSimStepOfLastCatch = 0;
  theSimStepWhenCatchedBallIsReleased = 0;
  theCanGoalieMove = NO;
  theSimStepOfLastSay = 0;
  theSimStepOfLastRefereeDecision = -1; // berechnet in SS0 sonst stndig eine neue IntentionListe
  theSimStepOfLastFreekick = 0;
  theSimStepOfLastSee = 0;
  simStepOfLastMichaelsError = -1;
  theIdString = "WorldModelCore";
} 

WorldModelCore::~WorldModelCore()
{
  delete theWorldModelIntegrator;
  delete theWorldModelCommunicator;
}
 
void WorldModelCore::initializeCore() 
{
  theWorldModelIntegrator = new WorldModelIntegrator();
  theWorldModelIntegrator->setWorldModelReference( this );
  theWorldModelIntegrator->setSensorsReference( theSensors );
  theWorldModelCommunicator = new WorldModelCommunicator();
  theWorldModelCommunicator->setWorldModelReference( this );
  
  // prepare [en|de]coding for communication
  // theWorldModelCommunicator->prepareEncodingAndDecoding();
 }

void WorldModelCore::integrateSensorInfo() {
  theWorldModelIntegrator->integrateSensorInfo();
}

void WorldModelCore::integrateCorrectSensorInfo() {
  theWorldModelIntegrator->integrateCorrectSensorInfo();
}
 
void WorldModelCore::registerExecution( const Action& anAction ) {
  theWorldModelIntegrator->registerExecution( anAction );
} 

const ModelledPlayer& WorldModelCore::player( int playerId ) const
{ 
  if ( playerId == SELF_ID ) {
    playerId = id();  
  } 
  if ( playerId == UNDEF_PLAYER_ID ) { 
    throwException( InvalidIdException(
        "Id was UNDEF_PLAYER_ID in WorldModel::player( int )") );
  }
  return theSituations[currentSituationIndex].playerWithId( playerId );  
} 

void WorldModelCore::setPlayerNumber( int newPlayerNumber, bool isGoalie ) 
{
  // wird von von Effector - enterField gerufen, nachdem der Server
  // nach der Anmeldung die Playernummer bermittelt
  thePlayerNumber = newPlayerNumber;  

  if( situation().playerWithId( id() ).id() == UNDEF_ID ){
    cerr << "WorldModelCore::setPlayerNumber    self id == UNDEF_INT\n";
    // ich fge mich in die aktuelle Situation ein
    ModelledPlayer playerSelf =
      ModelledPlayer( id(), Vector( -5.0, 37.0 ), Vector( 0.0, 0.0 ),
                      0.0, 0.0, VIEW_WIDTH_WIDE, VIEW_QUALITY_HIGH, globalConf->playerStaminaMax,
/*HACK*/              1.0/*EFFORT_MAX --> player.conf*/, 1.0 /*RECOVER*/,
                      isGoalie,
                      Confidence( simStep(), 0.0, 1.0 ), Confidence( simStep(), 0.0, 1.0 ),
                      Confidence( simStep(), 0.0, 1.0 ), Confidence( simStep(), 0.0, 1.0 ),
                      Confidence( simStep(), 0.0, 1.0 ) );
    playerSelf.setVisibleAngle( VIEW_WIDTH_NORMAL );
    theSituations[currentSituationIndex].addPlayer( playerSelf );

    INFO( Log::WORLD_MODEL, 7, idString(), globalLog->infoStream 
      << "New initial situation:\n" << theSituations[currentSituationIndex] );
    INFO( Log::WORLD_MODEL, 7, idString(), globalLog->infoStream 
      << "New game state: " << gameState() << " New initiative: " 
      << initiative() << " side: " << side() << endl );
  } 
  
  // Positionen und Rolle setzen
  setInitialPosition( WorldModelInterface::BEFORE_KICK_OFF, OUR_TEAM_ID,
      globalConf->kickoffPositionsOffensive[newPlayerNumber - 1] );
  setInitialPosition( WorldModelInterface::BEFORE_KICK_OFF, THEIR_TEAM_ID,
      globalConf->kickoffPositionsDefensive[newPlayerNumber - 1] );
  setInitialPosition( WorldModelInterface::PLAY_ON, OUR_TEAM_ID,
      globalConf->homePositionsOffensive[newPlayerNumber - 1] );
  setInitialPosition( WorldModelInterface::PLAY_ON, THEIR_TEAM_ID,
      globalConf->homePositionsDefensive[newPlayerNumber - 1] );
  theBasePositions = theInitialPositions;
  setRole( globalConf->roles[newPlayerNumber-1] );
}


const ModelledBall& WorldModelCore::ball() const
{
  return theSituations[currentSituationIndex].ball();
}

const Situation& WorldModelCore::situation( int timeStepOffset = 0 ) const
{
  if ( abs(timeStepOffset) > historyCount ) {
    throwException( "TimeStep offset too large." );
  }
  int index = (currentSituationIndex + timeStepOffset + situationFieldSize) %
    situationFieldSize;
  if ( index < 0 ) {
    throwException( "Negative index is illegal" );
  }

  if ( timeStepOffset > 0 &&
       theSituations[currentSituationIndex].simStep() + timeStepOffset >
       theSituations[index].simStep() ) {
    theSituations[index] = theSituations[currentSituationIndex];
    theSituations[index].simulateSteps( timeStepOffset );
  }
  return theSituations[index];
}

const Situation& WorldModelCore::correctSituation() const
{
  return theCorrectSituation;
}

bool WorldModelCore::hasInitiative( int aTeamId ) const
{
  bool answer;
  switch ( aTeamId ) {
  case OUR_TEAM_ID:
    answer = ( initiative() == OUR_TEAM_ID ||
               initiative() == BOTH_TEAMS_ID );
    break;
  case THEIR_TEAM_ID:
    answer = ( initiative() == THEIR_TEAM_ID ||
               initiative() == BOTH_TEAMS_ID );
    break;
  default:
    throwException( InvalidIdException(
        "Only OUR_TEAM_ID and THEIR_TEAM_ID are allowed as parameters." ) );
  }
  return answer;
}

void WorldModelCore::setGameState( int newGameState )
{
  if( gameState() == WorldModelInterface::CATCH_BALL ){
    if( newGameState == WorldModelInterface::FREE_KICK ){
      theGameState = WorldModelInterface::CATCH_BALL_AND_FREE_KICK;
      setCanGoalieMove( YES );
    }
  }
  else if( gameState() == WorldModelInterface::CATCH_BALL_AND_FREE_KICK ){
    setCanGoalieMove( NO );
    theWorldModelIntegrator->setSimStepWhenCatchedBallIsReleased( simStep() );
    theGameState = newGameState;
  }
  else if( newGameState == WorldModelInterface::FREEZE ){
    setSimStep( 0 );
    theGameState = newGameState;
  }
  else if( gameState() == WorldModelInterface::FREEZE ){
    if( newGameState == WorldModelInterface::PLAY_ON ){
      theGameState = newGameState;
    }
  }
  else {
    theGameState = newGameState;
  }
  // Zeit merken
  if ( isTimeElapsingGameState() == NO ) {
    theStartTimeOfElapsingGameState = globalTimer->getAgentTime();
//    cerr << "WMC-set-elapsing-Time " << theStartTimeOfElapsingGameState << endl;
  }
  else {
    theStartTimeOfElapsingGameState = UNDEF_INT;
//    cerr << "WMC-reset-elapsing-Time " << theStartTimeOfElapsingGameState << endl;
  }
} 

void WorldModelCore::setInitiative( int newInitiative )  
{ 
  if ( newInitiative == OUR_TEAM_ID || newInitiative == THEIR_TEAM_ID || 
       newInitiative == BOTH_TEAMS_ID ) { 
    theInitiative = newInitiative;  
  } 
  else if ( newInitiative == LEFT_SIDE_ID || newInitiative == RIGHT_SIDE_ID) { 
    if ( side() == newInitiative ) { 
      theInitiative = OUR_TEAM_ID;   
    } 
    else { 
      theInitiative = THEIR_TEAM_ID;  
    } 
  } 
  else { 
    theInitiative = NO_TEAM_ID;  
  }
}

Vector WorldModelCore::preferredKickDirection( const Vector& pos ) const
{
  INFO( Log::WORLD_MODEL, 3, idString(), globalLog->infoStream <<
        "    ENTER WorldModel::preferredKickDirection\n" );

  Vector toTarget;
  const Vector WING_ATTRACTOR( 0, 34 );
  Vector opponentAttractor;
  if ( player().pos().y() > 0 ) {
    opponentAttractor.set( 52.5, 4 );
  }
  else {
    opponentAttractor.set( 52.5, -4 );
  }

  const double targetVectorLength = 5.0;
  // Die Vorzugsrichtung ist abhaengig von der Region.
  RegionOnField regionOfPos = region( pos );
  INFO( Log::WORLD_MODEL, 8, idString(), globalLog->infoStream <<
	"I am in region " << (int)regionOfPos << "\n" );
  switch ( regionOfPos ) {
  case TO_UPPER:
    // Raus auf den oberen Fluegel
    toTarget = WING_ATTRACTOR - pos;  
    toTarget.setPolar( targetVectorLength, toTarget.angle() );
    break;  
  case TO_LOWER: 
    // Raus auf den unteren Fluegel 
    toTarget = ( -1.0 * WING_ATTRACTOR ) - pos;  
    toTarget.setPolar( targetVectorLength, toTarget.angle() );  
    break;  
  case REST: 
  case UPPER_MIDDLE: 
  case LOWER_MIDDLE: 
    // waagrecht rueber in die gegnerische Haelfte 
    toTarget = opponentAttractor;  
    toTarget.setPolar( targetVectorLength, toTarget.angle() );  
    break;  
  case UPPER: 
  case LOWER: 
    // weg vom Rand, direkt auf das Tor des Gegners
    toTarget = opponentAttractor - pos;  
    toTarget.setPolar( targetVectorLength, toTarget.angle() );  
    break;  
  case ENEMY_GOAL: 
    // direkt aufs gegnerische Tor 
    toTarget = opponentAttractor - pos;
    toTarget.setPolar( targetVectorLength, toTarget.angle() );  
    break;  
  case UPPER_EDGE: 
  case LOWER_EDGE: 
    toTarget = Vector( opponentAttractor.x() - 11, opponentAttractor.y() ) - 
      pos;  
    toTarget.setPolar( targetVectorLength, toTarget.angle() );  
    break;  
  default: 
    throwException( InvalidIdException( "Invalid region number." ) );  
  } 
  INFO( Log::WORLD_MODEL, 6, idString(), globalLog->infoStream <<
	"Preferred kick direction is " << toTarget << "\n" );

  return toTarget;  
} 

Vector WorldModelCore::backKickDirection( const Vector& pos ) const 
{
  INFO( Log::WORLD_MODEL, 3, idString(), globalLog->infoStream <<
        "    ENTER WorldModel::backKickDirection\n" );
  
  Vector toTarget;  
  const Vector WING_ATTRACTOR( 15, 34 );  
  Vector opponentAttractor;  
  if ( player().pos().y() > 0 ) { 
    opponentAttractor.set( 52.5, 4 );  
  } 
  else { 
    opponentAttractor.set( 52.5, -4 );  
  } 
  Vector selfAttractor;  
  if ( player().pos().y() > 0 ) { 
    opponentAttractor.set( -52.5, 30 );  
  }
  else {
    opponentAttractor.set( -52.5, -30 );
  }

  // ??? hartcodiert
  const double targetVectorLength = 5.0;
//  const double maximumSpeed = globalConf->ballSpeedMax / 1.5;
//  const double slowSpeed = maximumSpeed / 4;  
  // Die Vorzugsrichtung ist abhaengig von der Region. 
  RegionOnField regionOfPos = region( pos );  
  INFO( Log::WORLD_MODEL, 8, idString(), globalLog->infoStream <<
	"I am in region " << (int)regionOfPos << "\n" );
  switch ( regionOfPos ) { 
  case TO_UPPER: 
    // In die Mitte 
    toTarget = WING_ATTRACTOR - pos;
    toTarget.setPolar( targetVectorLength, toTarget.angle() + Ath98_PI );  
    break;  
  case TO_LOWER: 
    // In die Mitte 
    toTarget = ( -1.0 * WING_ATTRACTOR ) - pos;  
    toTarget.setPolar( targetVectorLength, toTarget.angle() + Ath98_PI );  
    break;  
  case REST: 
  case UPPER_MIDDLE:
  case LOWER_MIDDLE: 
    // Gerade zurueck 
    toTarget = opponentAttractor;  
    if( player().pos().x() < 40 ){ 
      toTarget.setPolar( targetVectorLength, toTarget.angle() + Ath98_PI );  
    } 
    else{ 
      toTarget.setPolar( targetVectorLength, toTarget.angle() + Ath98_PI );  
    } 
    break;  
  case UPPER: 
  case LOWER: 
    // Auf der Linie zurueck 
    toTarget = selfAttractor - pos;  
    toTarget.setPolar( targetVectorLength, toTarget.angle() );  
    break;  
  case ENEMY_GOAL: 
    // Auf die Fluegel zurueck 
    toTarget = opponentAttractor - pos;  
    toTarget.setPolar( targetVectorLength, toTarget.angle() + Ath98_PI );  
    break;  
  case UPPER_EDGE: 
  case LOWER_EDGE:
    toTarget = Vector( opponentAttractor.x() - 11, opponentAttractor.y() ) - 
      pos;  
    toTarget.setPolar( targetVectorLength, toTarget.angle() );  
    break;  
  default: 
    throwException( InvalidIdException( "Invalid region number." ) );  
  } 
  INFO( Log::WORLD_MODEL, 6, idString(), globalLog->infoStream <<
        "Preferred back-kick direction is " << toTarget << "\n" );
  
  return toTarget;  
} 

const double WorldModelCore::OUTER_LIMIT = 28.0;
const double WorldModelCore::INNER_LIMIT = 20.0;

RegionOnField WorldModelCore::region( const Vector& pos ) const
{
 
  const double MAIN_DIVIDER = 15;  
  const double WING_DISTRIBUTOR_DIVIDER = -20;  
  const double WING_DISTRIBUTOR_INNER_LIMIT = 7.1;  
  double x = pos.x();  
  double y = pos.y();  
  if ( x < MAIN_DIVIDER ) { 
    // Teil des Spielfelds, der die Mittellinie und das eigene Tor 
    // enthaelt 
    if ( y > OUTER_LIMIT ) { 
      // oben, Feldrand 
      return UPPER;  
    } 
    if ( y > INNER_LIMIT ) { 
      // oben, Streifen darunter 
      return UPPER_MIDDLE;  
    } 
    if ( y < -OUTER_LIMIT ) { 
      // unten, Feldrand 
      return LOWER;  
    } 
    if ( y < -INNER_LIMIT ) { 
      // unten, darueber 
      return LOWER_MIDDLE;
    } 
    if ( x < WING_DISTRIBUTOR_DIVIDER ) { 
      if ( y > WING_DISTRIBUTOR_INNER_LIMIT ) { 
        // Verteiler auf den oberen Fluegel 
        return TO_UPPER;  
      } 
      if ( y < -WING_DISTRIBUTOR_INNER_LIMIT ) { 
        // Verteiler auf den unteren Fluegel 
        return TO_LOWER;  
      } 
    } 
    // Der Rest ist eine T-foermige Region mit dem Querbalken quer zur 
    //Mittellinie und dem Laengsbalken zum eigenen Tor hin 
    return REST;  
  } 
  else { 
    // Teil des Spielfelds, das das gegnerische Tor enthaelt 
    double m;
    double c;  
    // Steigung(m) und Achsenabschnitt(c) einer Geraden berechnen, die 
    // vom oberen Torpfosten zum Punkt (18.5, 34) geht 
    m = ( theField.goalWidth()/2 - theField.fieldWidth()/2 ) /  
      ( theField.fieldLength()/2 - 18.5 );  
    c = theField.fieldWidth()/2 - 18.5 * m;  
    if ( y > m * x + c || y > OUTER_LIMIT ) { 
      // y liegt oberhalb der Geraden oder oberhalb von OUTER_LIMIT 
      return UPPER_EDGE;
    } 
    // Steigung(m) und Achsenabschnitt(c) einer Geraden berechnen, die 
    // vom unteren Torpfosten zum Punkt (18.5, -34) geht 
    m = ( -theField.goalWidth()/2 + theField.fieldWidth()/2 ) /  
      ( theField.fieldLength()/2 - 18.5 );  
    c = -theField.fieldWidth()/2 - 18.5 * m;  
    if ( y < m * x + c || y < -OUTER_LIMIT ) { 
      // y liegt unterhalb der Geraden oder unterhalb von OUTER_LIMIT 
      return LOWER_EDGE;  
    } 
    // Das uebrigbleibende Rechteck mit zwei abgeschnittenen Ecken. Die kurze 
    // Seite ist das gegnerische Tor 
    return ENEMY_GOAL;  
  } 
} 

 
int WorldModelCore::simStepOfSee() const 
{
  return theSensors->simStep();
}

/* INLINE FROM WM-CORE */
 
  NearestPlayerIndex::NearestPlayerIndex( double aDist, int aCrossIndex)
  : dist( aDist ), crossIndex( aCrossIndex ) {};


int WorldModelCore::playerNumber() const 
{
  return thePlayerNumber;
}

const Ath98String& WorldModelCore::teamName() const
{
  return theTeamName;
}

int WorldModelCore::side() const 
{
  return theSide;
}

int WorldModelCore::role() const 
{ 
  return theRole;      
}
 
int WorldModelCore::id() const
{
  return ( thePlayerNumber == UNDEF_INT) ? 
    UNDEF_PLAYER_ID : MIN_TEAMMATE_ID - 1 + thePlayerNumber;
}

Vector WorldModelCore::basePosition( int aGameState, int anInitiative ) const
{
  return theBasePositions[positionIndexFromState( aGameState, anInitiative) ];
}

Vector WorldModelCore::initialPosition( int aGameState, int anInitiative ) const
{
  return theInitialPositions[positionIndexFromState( aGameState, anInitiative) ];
}

int WorldModelCore::positionIndexFromState( int aGameState, int anInitiative ) const
{
  int index;
  if ( aGameState == UNDEF_INT ) {
    aGameState = gameState();
  }
  if ( anInitiative == UNDEF_INT ) {
    anInitiative = initiative();
  }

  switch ( aGameState ) {
  case WorldModelInterface::KICK_OFF:
  case WorldModelInterface::BEFORE_KICK_OFF:
  case WorldModelInterface::GOAL:
    if ( anInitiative == OUR_TEAM_ID ) {
      index = 0;
    }
    else {
      index = 1;
    }
    break;
  case WorldModelInterface::PLAY_ON:
    if ( anInitiative == OUR_TEAM_ID ) {
      index = 2;
    }
    else {
      index = 3;
    }
    break;
  default:
    cerr << "Problem in WorldModelCore::positionIndexFromState --> index unklar";
    index = 3;
  }
  return index;
}

  int WorldModelCore::simStep() const
{
  return theSituations[currentSituationIndex].simStep();
}

  int WorldModelCore::initiative() const
{
  return theInitiative;
}

  int WorldModelCore::gameState() const
{
  return theGameState;
}


  int WorldModelCore::score( int aTeamId ) const
{ 
  switch ( aTeamId ) { 
  case OUR_TEAM_ID: 
    return theScore[0];  
    break;  
  case THEIR_TEAM_ID: 
    return theScore[1];  
    break;  
  default: 
    throwException( InvalidIdException(  
        "Score for unknown team id demanded." ) );  
  } 
} 

const Field& WorldModelCore::field() const
{
  return theField;
}

string WorldModelCore::idString() const
{
  return theIdString;
}

ostream& operator<< (ostream& os, const WorldModelCore& worldmodel ) {
  os <<
    "WorldModel Info: simStep: " << worldmodel.simStep() <<
    "\n  gameState: " << worldmodel.gameState() <<
    "\n  initiative: " << worldmodel.initiative()  <<
    "\n  simStepOfLastCatch:" << worldmodel.simStepOfLastCatch() <<
    "\n  current Situation: " << worldmodel.situation() <<
    "\n  correct Situation: " << worldmodel.correctSituation() <<
    "  Base Positions: (" ;
  for( unsigned int i = 0; i < worldmodel.theBasePositions.size(); ++i ) {
    os << "\n   " <<  worldmodel.theBasePositions[i];
  }
  os <<  " )\n\n";
  return os;
}

void WorldModelCore::toStampStream( ostream& os) const {
  os << " " << WORLD_MODEL_MARK
     << " " << gameState()
     << " " << initiative();
  situation().toStampStream( os );
}


int WorldModelCore::simStepOfLastCatch() const
{
  return theSimStepOfLastCatch;
}

int WorldModelCore::simStepWhenCatchedBallIsReleased() const{
  return theSimStepWhenCatchedBallIsReleased;
}

int WorldModelCore::simStepOfLastSay() const
{
  return theSimStepOfLastSay;
}

int WorldModelCore::simStepOfLastRefereeDecision() const
{
  return theSimStepOfLastRefereeDecision;
}

int WorldModelCore::simStepOfLastFreekick() const
{
  return theSimStepOfLastFreekick;
}

int WorldModelCore::simStepOfLastSee() const 
{
  return theSimStepOfLastSee;
}


bool WorldModelCore::isTimeElapsingGameState() const
{
  return( gameState() != WorldModelInterface::BEFORE_KICK_OFF &&
          gameState() != WorldModelInterface::TIME_OVER &&
          gameState() != WorldModelInterface::GOAL &&
          gameState() != WorldModelInterface::OFFSIDE );
}


int WorldModelCore::startTimeOfElapsingGameState() const
{
  return  theStartTimeOfElapsingGameState;
}


void WorldModelCore::setTeamName( const Ath98String& newTeamName )
{
  theTeamName = newTeamName;
}

void WorldModelCore::setSide( int newSide )
{
  theSide = newSide;
}

void WorldModelCore::setRole( int newRole )
{
  theRole = newRole;
}

void WorldModelCore::setSimStep( int newValue )
{
  theSituations[currentSituationIndex].setSimStep( newValue );
}

void WorldModelCore::setScore(int newScore, int aTeamId )
{
  switch ( aTeamId ) {
  case OUR_TEAM_ID:
    theScore[0] = newScore;
    break;
  case THEIR_TEAM_ID:
    theScore[1] = newScore;
    break;
  default:
    throwException( InvalidIdException(
        "You try to set score for an unknown team id." ) );
  }
}



void WorldModelCore::setBasePosition( int aGameState, int anInitiative, const Vector& pos )
{
  theBasePositions[positionIndexFromState( aGameState, anInitiative)] = pos;
}

void WorldModelCore::setInitialPosition( int aGameState, int anInitiative, const Vector& pos )
{
  theInitialPositions[positionIndexFromState( aGameState, anInitiative)] = pos;
}

bool WorldModelCore::goalieCanMove() const{
  return theCanGoalieMove;
}

void WorldModelCore::setCanGoalieMove( bool hasMoved ){
  theCanGoalieMove = hasMoved;
}

Ath98String WorldModelCore::saySituation( const Situation& aSituation ) {
//cout << simStep() << "   WMCORE::SAYSITUATION = " << situation() << endl << endl;
    return theWorldModelCommunicator->saySituation( aSituation );
}


int WorldModelCore::getAgentTimeOfLastSeeInfo() {
	return theSensors->agentTimeOfLastSeeInfo;
}


