#ifndef _WORLDMODELINTERFACE_H_
#define _WORLDMODELINTERFACE_H_

// #include "WorldModelCore.h"
// die Vermeidung dieses includes sorgt fr die Unabhngigkeit der Klassen
// von WorldModelCore.h
// weil alle Klassen WorldModelInterface.h einbinden muessen

#include "../basic/Ath98String.h"
#include "../basic/Vector.h"
#include "Situation.h"
#include "Field.h"

class WorldModelCore;
class ModelledBall;
class ModelledPlayer;
class Sensors;
class Action;


/**
 
  Die Klasse ist nicht zur Vererbung geeignet.
  
  {\small {\bf Quelltext}

  [\URL[Header]{src/WorldModelInterface.h} | [\URL[Implementierung]{src/WorldModelInterface.cc}]

  {\bf CVS Informationen zum Interface}
  \begin{verbatim}
  **************************************************
  *
  * Aktueller Stand
  *
  *   $RCSfile: WorldModelInterface.h,v $
  *   $Revision: 1.1.1.1 $
  *   $Date: 2001/07/31 22:23:09 $
  *   $Author: rami $
  *
  **************************************************
  \end{verbatim} }*/


class WorldModelInterface
{
  friend class WorldModelCore;
  
 private:
  WorldModelCore* theWorldModelCore;
 
 public:
  
  /** @name RegionOnField
      Regionen auf dem Feld
      Symbolische Bezeichner f\"ur die Aufteilung des Feldes in verschiedene
      Regionen, in denen die preferredKickDirection verschieden berechnet
      wird.
      @see preferredKickDirection
  */
  /* enum RegionOnField { UPPER = 6, UPPER_MIDDLE = 4,
                       LOWER = 7, LOWER_MIDDLE = 5,
                       TO_UPPER = 1, TO_LOWER = 2,
                       REST = 3,
                       UPPER_EDGE = 9, LOWER_EDGE = 10,
                       ENEMY_GOAL = 8 }; 
  */
  /** @name M\"ogliche Spielzust\"ande
      Ein Wechsel des Spielzustands wird normalerweise vom Schiedsrichter
      bekanntgegeben. */
  //@{
    ///
    static const long GAME_STATE_MIN = 1;
    ///
    static const long BEFORE_KICK_OFF = 1;//0x00000001
    ///
    static const long TIME_OVER  = BEFORE_KICK_OFF<<1;//0x00000002
    ///
    static const long PLAY_ON = TIME_OVER<<1;//0x00000004
    ///
    static const long KICK_OFF = PLAY_ON<<1;//0x00000008
    ///
    static const long KICK_IN = KICK_OFF<<1;//0x00000010
    ///
    static const long FREE_KICK = KICK_IN<<1;//0x00000020
    ///
    static const long CORNER_KICK = FREE_KICK<<1;//0x00000040
    ///
    static const long GOAL_KICK = CORNER_KICK<<1;//0x00000080
    ///
    static const long GOAL = GOAL_KICK<<1;//0x00000100
    ///
    static const long FOUL = GOAL<<1;
    ///
    static const long HALF_TIME = FOUL<<1;
    ///
    static const long TIME_UP = HALF_TIME<<1;
    ///
    static const long TIME_UP_WITHOUT_A_TEAM = TIME_UP<<1;
    ///
    static const long TIME_EXTENDED =  TIME_UP_WITHOUT_A_TEAM<<1;
    ///
    static const long OFFSIDE = TIME_EXTENDED<<1;
    ///
    static const long DROP_BALL = OFFSIDE<<1;
    ///
    static const long CATCH_BALL = DROP_BALL<<1;
    ///
    static const long CATCH_BALL_AND_FREE_KICK = CATCH_BALL<<1;
    /// zum Einlesen neuer Parametersets beim Lernen
    static const long FREEZE = CATCH_BALL_AND_FREE_KICK<<1;
    ///
    static const long GAME_STATE_MAX = FREEZE;
    ///
    static const long INACTIVE_BALL = BEFORE_KICK_OFF & KICK_OFF & KICK_IN & FREE_KICK & CORNER_KICK
									  & GOAL_KICK & GOAL & FOUL & OFFSIDE & CATCH_BALL 
									  & CATCH_BALL_AND_FREE_KICK;
  //@}


  /** @name WorldModel anlegen */
  //@{
    /** Erzeugt ein neues Weltmodell, das seine Informationen aus der
      Sensorik-Komponente Sensors bezieht. Das Weltmodell legt ein internes
      Feld (Field) an, welches die Positionen der moeglichen Flaggen
      enthaehlt und eine Berechnung des eigenen Blickpunktes aus gesehenen
      Objekten erlaubt.
      @param Pointer auf das aktive Sensorik-Objekt */
    WorldModelInterface( Sensors * theSensors );
  //@}
 
    ~WorldModelInterface();

  /** @name Anfragen zum Agenten */
  //@{
    /** Liefert die Spielernummer des Agenten. */
    int playerNumber() const;
    
    /** Liefert den Namen des Teams, zu dem der Agent gehoert. */
    const Ath98String& teamName() const;
    
    /** Liefert die Seite des Agenten auf der das Tor seiner Mannschaft steht.
     */
    int side() const;

    /** Liefert die Id des Agenten. */
    int id() const;
    
    /** Liefert die Rolle des Agenten.
        Die Rolle bestimmt das Verhalten des Spielers. Anfangs wird die
        Rolle aus der Konfigurationsdatei eingelesen. Steht dort ein
        ung\"ultiger Wert, wird auch von dieser Funktion ein ung\"ultiger
        Wert geliefert.
        @return Der Wert einer der Rollenkonstanten.
        @see \Ref{Rollen} */
    int role() const;

    /** Liefert die Basisposition des Agenten. 
        Die BasisPosition wird anfangs aus der Konfigurationsdatei eingelesen,
        kann sich aber waehrend des Spiels \"andern. Je nach Spielzustand und
        Initiative wird eine andere Basisposition geliefert. Werden daf\"ur
        keine Werte angegeben, werden die aktuellen Werte aus dem Weltmodell 
        verwendet.
        @param aGameState Einer der m\"oglichen Spielzust\"ande.
        @param anInitiative Einer der Werte von in \Ref{Teams}.
        @return absoluter Vektor zur Basisposition.*/
    Vector basePosition( int aGameState = UNDEF_INT,
                         int anInitiative = UNDEF_INT ) const;

    /** Liefert die Anfangsposition des Agenten.
        Je nach Spielzustand und Initiative wird eine andere Anfangsposition
        geliefert. Werden daf\"ur keine Werte angegeben, werden die aktuellen
        Werte aus dem Weltmodell verwendet. Die Anfangsposition wird aus der
        Konfigurationsdatei ausgelesen und \"andert sich nur bei einem
        Rollentausch. Als Laufziel sollte immer ein Wert von
        \Ref{basePosition}() verwendet werden.
        @param aGameState Einer der m\"oglichen Spielzust\"ande.
        @param anInitiative Einer der Werte von in \Ref{Teams}.
        @return absoluter Vektor zur Anfangsposition.*/
    Vector initialPosition( int aGameState = UNDEF_INT,
                            int anInitiative = UNDEF_INT ) const;

  //@}

  /** @name Werte des Agenten setzen */
  //@{
    /** Setzt die Spielernummer des Agenten.
        Ein ModelledPlayer mit seiner Id wird in die aktuelle 
        Situation aufgenommen, falls er dort noch nicht vorhanden ist.
        Die Anfangs- und die Basispositionen  werden aus \Ref{globalConf}
        gelesen und gesetzt. */
    void setPlayerNumber( int newPlayerNumber, bool isGoalie );

    /** Setzt den Namen des Teams, zu dem der Agent gehoert. */
    void setTeamName( const Ath98String& newTeamName );

    /** Setzt die Seite, auf der der Agent spielt. */
    void setSide( int newSide );

    /** Setzt die Rolle eines Spielers.
        Die Rolle bestimmt das Verhalten des Spielers. Anfangs wird die
        Rolle aus der Konfigurationsdatei eingelesen. Die G\"ultigkeit der
        Rolle wird nicht \"uberpr\"uft.
        @param newRole neue Rolle des Spielers.
        @see \Ref{Rollen} */
    void setRole( int newRole );

    /** Setzt die Basisposition des Agenten.
        Hier m\"ussen der Spielzustand und die Initiative angegeben werden.
        Gilt dieselbe Position f\"ur verschiedene Zust\"ande, so modifiziert
        das Setzen in einem dieser Zust\"ande die Position f\"ur alle diese
        Zust\"ande.
        @param aGameState Einer der m\"oglichen Spielzust\"ande.
        @param anInitiative Einer der Werte von in \Ref{Teams}. */
    void setBasePosition( int aGameState, int anInitiative, 
                          const Vector& pos );



 private:
    /** Setzt die Anfangsposition des Agenten.
        Hier m\"ussen der Spielzustand und die Initiative angegeben werden.
        Gilt dieselbe Position f\"ur verschiedene Zust\"ande, so modifiziert
        das Setzen in einem dieser Zust\"ande die Position f\"ur alle diese
        Zust\"ande.
        Nur die Initialisierungs- und der Rollentauschroutine setzen die
        Anfangspositionen um.
        @param aGameState Einer der m\"oglichen Spielzust\"ande.
        @param anInitiative Einer der Werte von in \Ref{Teams}. */
    void setInitialPosition( int aGameState, int anInitiative, 
                             const Vector& pos );

 public:
  //@}

  /**@name Anfragen zur aktuellen Situation */
  //@{
    /** Liefert aktuelle Daten ueber den gewuenschten Spieler. 
        Wird kein Spieler angegeben, werden die Daten ueber den Agenten selbst
        geliefert. \\
        @param playerId die Id des gewuenschten Spielers. (default ist id() )
        @exception Situation::NoPlayerWithThisIdException, wenn der angegebene
        Spieler in der aktuellen Situation nicht existiert.
        @exception InvalidIdException, wenn die playerId ungueltig ist.
        @return ModelledPlayer mit den gewuenschten Daten. */
    const ModelledPlayer& player( int playerId = SELF_ID ) const;

    /** Liefert aktuelle Daten ueber den Ball.
        Wenn es keine Daten ueber den  Ball gibt, wird ein BallObjekt mit
        undefinierten Werten zurueckgegeben. */
    const ModelledBall& ball() const;

    /** Liefert den Simulationstakt der aktuellen Situation. 
        Die zu diesem Zeittakt gehoerende Situation wird von situation() 
        geliefert. */
    int simStep() const;

    //** */
    int simStepOfSee() const;  

    /** Liefert aktuelle, vergangene und zukuenftige Situationen.
        Es wird eine gewisse, vom historyCount-Parameter in agent.conf
        abhaengige Zahl von alten Situationen gespeichert. Aufgrund von
        Synchronisationsproblemen oder eines Spielzustands, in dem das Spiel
        nicht laeuft, lassen sich nicht immer alle Situationen bis zum Takt
        simStep-historyCount abfragen. \\
        Die haeufigste Verwendung ist sicher die Frage nach der aktuellen
        Situation. Der Parameter timeStepOffset ist per Default auf 0 gesetzt,
        so dass mit situation() auf die aktuelle Situation zugegriffen
        werden kann. \\
        @param timeStepOffset Mit timeStepOffset < 0 koennen vergangene
        (d.h. gespeicherte) Situationen abgerufen werden. Mit timeStepOffset >
        0 koennen zukuenftige Situationen vorausberechnet werden. Zulaessige
        Werte liegen zwischen -historyCount und +historyCount (wird in
        agent.conf gesetzt).
        @exception Array<Situation>::IndexOutOfBoundsException, wenn der
        timeStepOffset einen ungueltigen Wert hatte.
        @return Eine Referenz, die die angefragte Situation enth\"alt. Diese
        veraendert sich moeglicherweise mit Aenderung des simSteps(). */
    const Situation& situation( int timeStepOffset = 0 ) const;


    /** Liefert die aktuelle korrekte Situation aus den Fullstate-
        Informationen. Diese Situation ist stets vollstaendig und korrekt, 
        steht aber nur fuer Testspiele zur Verfuegung.
        @return Eine Referenz, die die angefragte Situation enth\"alt. Diese
        veraendert sich moeglicherweise mit Aenderung des simSteps(). */
    const Situation& correctSituation() const;


    /** Liefert den Spielzustand zurueck. Dieser ist als int kodiert. */
    int gameState() const;

    /** Liefert die Initiative.
        @return Einer der Werte OUR_TEAM_ID, THEIR_TEAM_ID, BOTH_TEAMS_ID,
        NO_TEAM_ID */
    int initiative() const;

    /** Gibt an, ob ein bestimmtes Team die Initiative hat. 
        Diese Funktion erleichtert das Umgehen mit der Initiative, da mehrfache
        Abfragen unn\"otig werden. Wir haben die Initiative, wenn, die
        Initiative auf OUR_TEAM_ID oder auf BOTH_TEAMS_ID steht.
        Entsprechendes gilt fuer THEIR_TEAM_ID.
        @param aTeamId Das Team, von erfragt werden soll, ob es die Initiative
        hat. M\"ogliche Werte sind OUR_TEAM_ID, THEIR_TEAM_ID.
        @exception InvalidIdException, wenn die aTeamId einen ungueltigen Wert 
        hat.
        @return YES oder NO. */
    bool hasInitiative( int aTeamId ) const;

    /** Anfrage ob wir uns gerade in einem Spielzustand befinden in dem
        die Zeit im SoccerServer in jedem Simulationstakt weitergezaehlt
        wird oder nicht
        @return YES wenn die Zeit im aktuellen Spielzustand fortlaeuft
        @return NO wenn die Zeit im aktuellen Spielzustand still steht */
    bool isTimeElapsingGameState();
    
    /** Anfrage wann der Spielzustand begann */
    int startTimeOfElapsingGameState() const;
  //@}

  /** @name Werte der aktuellen Situation setzen */
  //@{
    /** Setzt den Simulationstakt der aktuellen Situation. 
        Diese Funktion sollte mit Vorsicht verwendet werden, damit nicht die
        interne Synchronisation des Weltmodells durcheinander geraet. */
    void setSimStep( int newValue );

    /** Setzt den Spielzustand. */
    void setGameState( int newGameState );

    /** Setzt die Initiative. 
        Dabei kann auch eine Seite angegeben werden.  Spielen wir gerade auf
        dieser Seite, wird die Initiative auf OUR_TEAM_ID gesetzt, andernfalls
        auf THEIR_TEAM_ID. \\
        @param newValue Entweder die Id eines Teams ( OUR_TEAM_ID, 
        THEIR_TEAM_ID, BOTH_TEAMS_ID ) oder die Id einer Seite (LEFT_SIDE_ID,
        RIGHT_SIDE_ID ) */
    void setInitiative(int newValue);

 public:
  //@}

  /**@name Allgemeine und andere Methoden*/
  //@{
    /** Liefert die Zahl der Tore, die ein Team bisher erzielt hat.
        @param aTeamId Id des Teams. Gueltig sind OUR_TEAM_ID und THEIR_TEAM_ID
        @exception InvalidIdException, wenn die aTeamId einen ungueltigen Wert
        hat.
        @return Die Zahl der Tore dieses Teams. */
    int score(int aTeamId) const;

    /** Setzt die Zahl der Tore fuer ein Team.
        @param newScore Die neue Zahl der Tore
        @param aTeamId Id des Teams. Gueltig sind OUR_TEAM_ID und THEIR_TEAM_ID
        @exception InvalidIdException, wenn die aTeamId einen ungueltigen Wert
        hat */
    void setScore( int newScore, int aTeamId );

    /** Liefert das Feld, auf dem der Agent spielt. */
    const Field& field() const;

    /** Bestimmt anhand von Reliability-Daten von Objekten, Wichtigkeit
        dieser Objekte im aktuellen Spielverlauf und Informationen \"uber
        gesehene Bereiche die Blickrichtung des Necks, um ein m\"oglichst
        vollst\"andiges Weltbild zu erhalten **/ 
    void calculateBestViewDirection();


 private:
    /** Liefert einen internen Index, an dem die Position 
        des Agenten fuer den angegebenen Spielzustand/Initiative steht. 
        Mit diesem Index kann auf Elemente von theBasePositions und
        theInitialPositions zgegriffen werden.
        @return Zahl zwischen 0 und 10
    */
    int positionIndexFromState( int aGameState, int anInitiative ) const;


 public:
  //@}

  /**@name Anfragen und Setzen von simsteps */
  //@{
    /** Der simStep, an dem das letzte Catch-Kommando an den Server
        geschickt wurde. */
    int simStepOfLastCatch() const;
    
    /** Der simStep, an dem der Ball das letzte Mal gecatcht wurde. */
    int simStepWhenCatchedBallIsReleased() const;

    /** Der simStep, an dem das letzte Say-Kommando an den Server
        geschickt wurde. */
    int simStepOfLastSay() const;
    
    /** Der simStep, an dem die letzte Schiedsrichterentscheidung
        eingetroffen ist. */
    int simStepOfLastRefereeDecision() const;
    
    /** simStep der letzten Freistoss-Entscheidung. */
    int simStepOfLastFreekick() const;
    
    /** simStep der letzten SichtInformation. */
    int simStepOfLastSee() const;


 public:
  //@}

  /**@name Anfragen und Setzen den Goalie betreffend. */
  //@{
    /** Der GoalieMove.
        @return YES, wenn der Goalie einen Movebefehl geschickt hat. */
    bool goalieCanMove() const;
    
    /** Setzt den Wert, der aussagt, ob der Goalie schon  einen Movebefehl
	abgesendet hat. */
    void setCanGoalieMove( bool hasMoved );    
  //@}

  /** @name Einarbeitung neuer Informationen */
  //@{
    /** Integriert die Sensorinformationen in das Weltmodell.
        Bei einer neuen BodyInformation wird die Methode
        \Ref{processBodyInfo} ausgefuehrt.\\
        Bei einer neuen Schiedsrichterentscheidung wird die Methode
        \Ref{processRefereeInfo} ausgefuehrt.\\
        Bei einer neuen Hoerinformation, die nicht vom Schiedsrichter
        ist, wird die Methode \Ref{processHearInfo} ausgefuehrt und
        die dadurch erzeugte Situation mit der aktuellen Situation
        mittels der Methode \Ref{mergeSituations} fusioniert. \\
        Bei einer neuen Sichtinformation wird die Methode
        \Ref{processSeeInfo} ausgefuehrt und die dadurch erzeugte
        Situation mit der aktuellen mittels der Methode 
        \Ref{mergeSituations}fusioniert. Danach wird die
        fusionierte Situation als neue aktuelle Situation angenommen.\\
        Es koennen mehrere dieser Schritte hintereinander durchgefuehrt
        werden, falls mehrere neue Informationen vorliegen.\\
        Am Ende wird mit der Methode \Ref{improveConsistency} ein
        Konsistenzcheck auf dem Weltmodell und der aktuellen Situation
        durchgefuehrt und die Konsistenz wieder hergestellt, falls
        dies noetig sein sollte. \\
        @see processBodyInfo
        @see processRefereeInfo
        @see processHearInfo
        @see processSeeInfo
        @see mergeSituations
        @see improveConsistency */
    void integrateSensorInfo();

    /** Integriert die korrekten Sensorinformationen in einen separaten Bereich
        des Weltmodells. */
    void integrateCorrectSensorInfo();

    /** Registriert, da\3 eine Aktion abgesendet wurde.
        Die abgesendete Aktion wird gepuffert. Eine Simulation wird
        erst durchgefuehrt, wenn eine Bestaetigung der Ausfuehrung
        durch den SoccerServer vorliegt (durch BodyInformation).
        @param anAction Die abgeschickte Aktion */
    void registerExecution( const Action& anAction );

 public:
  //@}

 public:
  //@}

  /** @name Abfragen zur Vorzugsrichtung */
  //@{
    /** Liefert die bevorzugte Kickrichtung.
        Diese "Vorzugsrichtung" ist abh\"angig von der Position des Spielers.
        Durch sie werden wichtige Eigenschaften des Mannschaftsspiels
        umgesetzt, z.b. Flanken und Angriffe \"uber die Fl\"ugel.
        @param pos Position, f\"ur die die Vorzugsrichtung berechnet werden
        soll.
        @return Ein Richtungsvektor. Die Richtungskomponente ( \Ref{length()})
        direkt der kickBall-Skill \"ubergeben werden. Die Laenge gibt die
        optimale Ballgeschwindigkeit in m/simStep an. Die Division der Laenge
        durch globalConf->kickPowerRate ergibt eine Kraft, die kickBall
        \"ubergeben werden kann.*/
    Vector preferredKickDirection( const Vector& pos ) const;

    /** Liefert die bezorzugte R\"uckpassrichtung.
        @param pos Position, f\"ur die die Vorzugsrichtung berechnet werden
        soll.
        @return Ein Richtungsvektor. Die Richtungskomponente ( \Ref{length()})
        direkt der kickBall-Skill \"ubergeben werden. Die Laenge gibt die
        optimale Ballgeschwindigkeit in m/simStep an. Die Division der Laenge
        durch globalConf->kickPowerRate ergibt eine Kraft, die kickBall
        \"ubergeben werden kann.*/
    Vector backKickDirection( const Vector& pos ) const;

    /** Liefert die Region, in der sich eine Position befindet.
        Ja nach Region kann sich ein Spieler anders verhalten. Auch die
        Definition der Vorzugsrichtungsfunktion basiert auf der Region, in der
        sich eine Position befindet. 
        @param pos Position, f\"ur die die Region ermittelt werden soll.
        @return Die Nummer der Region, ein Enum vom Type RegionOnField.
        @see RegionOnField
    */
    RegionOnField region( const Vector& pos ) const;   
  //@}

  /** @name Methoden zum Debugging */
  //@{
    /** Liefert einen String fuer die Klassenbezeichnung zurueck. */
    string idString() const;

    /** Ausgabe auf einen String-Stream */
    friend ostream& operator<< (ostream&, const WorldModelInterface& );

    /** kompakte Ausgabe auf einen Stream fuer constant Log */
    void toStampStream( ostream& ) const;
  //@}

  /** @name Methoden fuer die Kommunikation */
  //@{
    /** Erzeugt eine Nachricht fuer den Ruf des Weltmodells.
        @return die Nachricht */
    Ath98String saySituation( const Situation& aSituation);

  //@}
  
   int simStepOfLastMichaelsError();

   int getAgentTimeOfLastSeeInfo();
                       

 // Variablen und Typen
 private:
  /** @name Klassenvariablen */
  //@{
    /** \"Ausseres Limit.
        y-Koordinate einer waagrechten Linie 6m vom Feldrand. */
    static const double OUTER_LIMIT;
    /** Inneres Limit.
        y-Koordinate einer waagrechten Linie 14m vom Feldrand. */
    static const double INNER_LIMIT;


  /** @name Instanzvariablen */
  //@{


  /** @name Excpetions (interne) */
  //@{
    /** Interne Fehlerklasse. Diese Exception wird gerufen wenn eine aeltere
        Situation fehlerhafte Werte aufweisst. */
    class IncorrectOldSituationException: public Exception {
    public:
      IncorrectOldSituationException()
        : Exception() {}
    };
    /** Interne Exception, die aufgerufen wird, wenn keine korrespondierenden
        Objekte gefunden werden k\"onnen */
    class NoCorrespondingPlayerException: public Exception {
    public:
      NoCorrespondingPlayerException()
        : Exception() {}
    };
  //@}

};



/** Global verfuegbares Weltmodell.
    Enthaelt die Annahmen des Spielers ueber den realen Zustand  seiner Umwelt.
*/
extern WorldModelInterface *globalWorldModel;


#endif

